﻿:Class HTTPRequest

    CR LF←NL←⎕UCS 13 10

⍝ Common Status Codes
    SC←(200 'OK')(301 'Moved Permanently')(401 'Unauthorized')(404 'Not Found')
    SC,←(500 'Internal Server Error')(503 'Service Unavailable')
    SC←↑SC

⍝ Common Content Types
    CT←('txt' 'text/plain')('htm' 'text/html') ('css' 'text/css')('xml' 'text/xml')('svg' 'image/svg+xml')
    CT,←('zip' 'application/x-zip-compressed')('pdf' 'application/pdf')
    CT←↑CT

⍝ Fields related to the Request

    :Field Public Instance Input
    :Field Public Instance Headers
    :Field Public Instance Command
    :Field Public Instance Page
    :Field Public Instance Arguments
    :Field Public Instance PeerAddr
    :Field Public Instance PeerCert
    :Field Public Instance Data
    :Field Public Instance Cookies
    :Field Public Instance Session←''
    :Field Public Instance Server←⎕NS ''

⍝ Fields related to the Response

    :Field Public Instance Response

    GetFromTable←{(⍵[;1]⍳⊂#.Strings.lc ⍺)⊃⍵[;2],⊂''}
    ine←{⍺←'' ⋄ 0∊⍴⍵:'' ⋄ ⍺,⍵} ⍝ if not empty
    inf←{∨/⍵⍷⍺:'' ⋄ ⍵} ⍝ if not found

    ∇ Make(cmd data);split;buf;input;args;req;hdrs;i;z;pars
      :Access Public Instance
      :Implements Constructor
⍝ Decode an HTTP command line: get /page&arg1=x&arg2=y
⍝ Return namespace containing:
⍝ Command: HTTP Command ('get' or 'post')
⍝ Headers: HTTP Headers as 2 column matrix or name/value pairs
⍝ Page:    Requested page
⍝ Arguments: Arguments to the command (cmd?arg1=value1&arg2=value2) as 2 column matrix of name/value pairs
      
      Response←⎕NS ''
      Response.(Status StatusText Headers File HTML HTMLHead PeerAddr)←200 'OK'(0 2⍴⊂'')0 '' '' ''
      Response.Request←cmd,data
      PeerCert←0 0⍴⊂'' ⋄ Data←0 2⍴⊂''
      PeerAddr←''
     
      input←1⊃,req←2⊃DecodeHeader cmd
      Input←input
      hdrs←{(0≠⊃∘⍴¨⍵[;1])⌿⍵}1 0↓req
      Headers←hdrs
     
      split←{p←(⍺⍷⍵)⍳1 ⋄ ((p-1)↑⍵)(p↓⍵)} ⍝ Split ⍵ on first occurrence of ⍺
     
      Command buf←' 'split input
      Command←#.Strings.lc Command
     
      buf z←'HTTP/'split buf
      Page args←'?'split ¯1↓buf
      Page←ArgXLT Page
     
      Arguments←URLDecodeArgs args
     
⍝ PeerCert←conns.PeerCert
     
      :If (1↑⍴hdrs)≥i←hdrs[;1]⍳⊂'content-type'
      :AndIf 'multipart/form-data'{⍺≡(⍴⍺)↑⍵}z←{(+/∧\⍵=' ')↓⍵}⊃hdrs[i;2]
          z←'--',(8+('boundary='⍷z)⍳1)↓z ⍝ boundary string
          Data←↑DecodeMultiPart¨¯1↓z{(⍴⍺)↓¨(⍺⍷⍵)⊂⍵}data ⍝ ¯1↓ because last boundary has '--' appended
      :ElseIf 'application/x-www-form-urlencoded'{⍺≡(⍴⍺)↑⍵}z
          Data←URLDecodeArgs data
      :Else
          Data←0 2⍴⊂''
      :EndIf
      :If (1↑⍴hdrs)≥i←hdrs[;1]⍳⊂'cookie'
          Cookies←↑{((#.Strings.lc(1⊃⍵)~' '))(DeCode 2⊃⍵)}¨';='Split⊃hdrs[i;2]
      :Else
          Cookies←0 2⍴⊂''
      :EndIf
     
      :If 9=⎕NC'SessionHandler' ⍝ Was a SessionHandler assigned?
          SessionHandler.HandleRequest ⎕THIS ⍝ If so, let it do its stuff
      :EndIf
    ∇

    ∇ r←URLDecodeArgs args
      r←(args∨.≠' ')⌿↑'='∘split¨{1↓¨(⍵='&')⊂⍵}'&',args ⍝ Cut on '&'
      r[;2]←ArgXLT¨r[;2]
    ∇

    ∇ r←URLEncode data;⎕IO;z;ok;nul;m;enlist
      :Access Public Shared
     
      nul←⎕UCS ⎕IO←0
      enlist←{⎕ML←3 ⋄ ∊⍵}
      ok←nul,enlist ⎕UCS¨(⎕UCS'aA0')+⍳¨26 26 10
     
      z←⎕UCS'UTF-8'⎕UCS enlist nul,¨,data
      :If ∨/m←~z∊ok
          (m/z)←↓'%',(⎕D,⎕A)[⍉16 16⊤⎕UCS m/z]
          data←(⍴data)⍴1↓¨{(⍵=nul)⊂⍵}enlist z
      :EndIf
     
      r←¯1↓enlist data,¨(⍴data)⍴'=&'
    ∇

    ∇ r←GetCookie name
      :Access Public Instance
     
      r←name GetFromTable Cookies
    ∇

    ∇ r←GetArgument name
      :Access Public Instance
      r←name GetFromTable Arguments
    ∇

    ∇ r←GetHeader name
      :Access Public Instance
      r←(#.Strings.lc name)GetFromTable Headers
    ∇

    ∇ SetCookie ctl;name;value;path;date;z;keep
      :Access Public Instance ⍝ Set a Cookie
     
      name value path keep←ctl,(⍴ctl)↓'CookieName' 'CookieValue' '/' 30
      date←#.Dates.IDNToDate keep+#.Dates.DateToIDN 3↑⎕TS ⍝ keep is # of days cookie should be valid
      z←#.Dates.CookieFmt(3↑date),23 59 59
      z←name,'=',value,'; path=',path,'; expires="',z,'"'
      Response.Headers⍪←'set-cookie'z
    ∇

    ∇ DelCookie ctl;name;path
      :Access Public Instance ⍝ Delete a cookie
     
      :If 1=≡ctl ⋄ ctl←,⊂ctl ⋄ :EndIf
      name path←ctl,(⍴ctl)↓'CookieName' '/'
      SetCookie name''path ¯30 ⍝ Expiry 30 days ago
    ∇

    ∇ Title x
      :Access Public Instance
      Response.HTMLHead,←'<title>',x,'</title>',NL
    ∇

    ∇ {tags}Script x
      :Access Public Instance
      :If 0=⎕NC'tags' ⋄ tags←'type="text/javascript"'
      :ElseIf ~∨/'type='⍷tags ⋄ tags,←' type="text/javascript"'
      :EndIf
      Response.HTMLHead,←Response.HTMLHead inf'<script ',tags,'>',(NL ine x),'</script>',NL
    ∇

    ∇ r←JSPlugIn file;root ⍝ Retrieve a JavaScript PlugIn
      :Access Public Instance
      root←{(1-(⌽⍵)⍳'\')↓⍵}⎕WSID
      r←#.Files.GetText root,'PlugIns\',file
    ∇

    ∇ Style file
      :Access Public Instance
      Response.HTMLHead,←Response.HTMLHead inf'<link href="',file,'" rel="stylesheet" type="text/css">',NL
    ∇

    ∇ {hdrs}Return x;type
      :Access Public Instance
     
      :If 2=⍴⍴x ⋄ :AndIf 2=¯1↑⍴x ⋄ x←URLEncode x ⋄ type←'text/plain' ⍝ 2-column matrix is name/value pairs
      :Else ⋄ type←'text/html'
      :EndIf
      Response.HTML←x
      :If 2=⎕NC'hdrs'
          Response.Headers⍪←hdrs
      :Else
          Response.Headers⍪←'content-type'(type,'; charset=utf-8')
      :EndIf
    ∇

    ∇ Fail x;i;f;root
      :Access Public Instance     
      root←{(1-(⌽⍵)⍳'\')↓⍵}⎕WSID
      :If 3=10|⎕DR x ⍝ Just a status code
          Response.Status←x
          :If (1↑⍴SC[;1])≥i←SC[;1]⍳x
              Response.StatusText←⊃SC[i;2] ⍝ Default text
          :EndIf
      :Else
          Response.(Status StatusText)←x
      :EndIf
      :If #.Files.Exists f←root,'ErrorPages\',(⍕x),'.htm'
          Response.HTML←(#.Files.GetText f),'<br/>',Response.HTML
      :EndIf
    ∇

    ∇ {hdrs}ReturnFile x;ct
      :Access Public Instance
     
      Response.(HTML File)←x 1
      :If 2=⎕NC'hdrs'
          Response.Headers⍪←hdrs
      :Else
          ct←(CT[;1]⍳⊂#.Strings.lc 3↑{(1-(⌽⍵)⍳'.')↑⍵}x)⊃CT[;2],⊂'application/binary'
          Response.Headers⍪←'content-type'ct
      :EndIf
    ∇

    ∇ r←DecodeHeader buf;len;d;i
 ⍝ Decode HTML Header
      :Access public shared
      len←(¯1+⍴NL,NL)+⊃{((NL,NL)⍷⍵)/⍳⍴⍵}buf
      :If len>0
          d←(⍴NL)↓¨{(NL⍷⍵)⊂⍵}NL,len↑buf
          d←↑{((p-1)↑⍵)({(' '=1↑⍵)↓⍵}(p←⍵⍳':')↓⍵)}¨d
          d[i;1]←#.Strings.lc¨d[i←1↓⍳1↑⍴d;1]
      :Else
          d←⍬
      :EndIf
      r←len d
    ∇

    ∇ r←DecodeMultiPart data;d;t;filename;name;i
      d←'Content-Disposition: 'GetParam data
      :If (⍴d)≥i←5+('name="'⍷d)⍳1
          name←(¯1+name⍳'"')↑name←i↓d
      :Else ⋄ name←''
      :EndIf
     
      t←'Content-Type: 'GetParam data
     
      data←(3+((NL,NL)⍷data)⍳1)↓data ⍝ Drop up to 1st doubleCR
      data←(¯1+¯1↑(NL⍷data)/⍳⍴data)↑data ⍝ Drop from last CR
     
      :Select t ⍝ Content type
      :CaseList 'plain/text' 'text/plain' 'text/html' '' ⍝ These are already processed
          ⍝ :Trap 92 ⋄ data←'UTF-8'⎕UCS ⎕UCS data ⋄ :EndTrap ⍝ From UTF-8 (will probably fail in Classic)
      :CaseList 'text/xml' 'application/msword' 'application/pdf' 'application/octet-stream' 'image/gif' 'image/pjpeg' 'image/bmp' 'application/x-zip-compressed' 'application/vnd.ms-excel'
          :If (⍴d)≥i←9+('filename="'⍷d)⍳1
              filename←(¯1+filename⍳'"')↑filename←i↓d
          :Else ⋄ filename←''
          :EndIf
          data←filename data
      :Else
          ∘
      :EndSelect
      r←name data
    ∇

    ∇ r←header GetValue(name type);i;h
      ⍝ Extract value from HTTP Header structure returned by DecodeHeader
     
      :If (1↑⍴header)<i←header[;1]⍳⊂lc name
          r←⍬ ⍝ Not found
      :Else
          r←⊃header[i;2]
          :If 'Numeric'≡type
              r←1⊃2⊃⎕VFI r
          :EndIf
      :EndIf
    ∇

    ∇ r←name GetParam HTTP;i;z;find
      find←{⊃(⍺⍷⍵)/⍳⍴⍵}
      r←''
      :If 0<i←name find HTTP
      :AndIf 0<i←NL find z←(¯1+i+⍴name)↓HTTP
          r←(i-1)↑z
      :EndIf
    ∇

      Split←{⎕IO ⎕ML←0 3         ⍝ Split Http syntax.
          ⍺←'?&=' ⋄ (,⍺){        ⍝ Default separator chars.
              0=⍴⍺:⍵             ⍝ No separator chars left: finished.
              sepr←↑⍺            ⍝ First char is separator char.
              (1↓⍺)∘∇∘{          ⍝ Recursively split at each sub level.
                  ⍵~sepr         ⍝ Separator removed.
              }¨(1+⍵=sepr)⊂⍵     ⍝ String partitioned at separator char.
          }⍵                     ⍝ Http string.
      }

      DeCode←{⎕IO ⎕ML←0 3            ⍝ Decode Special chars in HTML string.
          hex←'0123456789ABCDEF'     ⍝ Hex chars.
          {                          ⍝ Convert numbers.
              v f←⎕VFI ⍵             ⍝ Check for numbers.
              ~∧/v:⍵                 ⍝ Not all numbers: char vec.
              1=⍴f:↑f ⋄ f            ⍝ Numeric scalar or vector.
          }∊{                        ⍝ Enlist of segments.
              '%'≠↑⍵:⍵               ⍝ 1st seg may not contain special char.
              (⎕UCS 16⊥hex⍳1↓3↑⍵),3↓⍵  ⍝ Hex code replaced with corresp. ⎕AV char.
          }¨(1+⍵='%')⊂,⍵             ⍝ Segments split at '%' char.
      }


    ∇ r←ArgXLT r;i;j;m;z
    ⍝ Translate HTTP command line arguments
     
      ((r='+')/r)←' '
      :If 0≠⍴i←(r='%')/⍳⍴r
      :AndIf 0≠⍴i←(i<¯1+⍴r)/i
          z←r[j←i∘.+1 2]
          r[i]←⎕UCS 16⊥⍉¯1+('0123456789ABCDEF'⍳z)⌊'0123456789abcdef'⍳z
          m←(⍴r)⍴1 ⋄ m[j]←0
          r←m/r
      :EndIf
    ∇

:EndClass