﻿:Class Shop: MildPage
    :Include #.HTMLInput

    :Field Public CartAction← ''
    :Field Public Text← ''
    :Field Private ProductNumbers ← 101 103 102
    :Field Private Products←3 3⍴101 'Apple' 1 102 'Ball' 3 103 'Cactus' 4
    :Field Public Cart←0 2⍴''

    ∇ Render req;html;shop
      ⍝ This MiPage demonstrates the use of session variables and cookies.      

      :Access Public
      
      DoAction    ⍝ When a button is pressed, do something about it.    
      html←'div class="leftalign"'Enclose'MiShop'(Products Shop)ProductNumbers   ⍝ Insert the 'shop window' and align it to the left with CSS
      html,←'div class="rightalign"'Enclose 'MiCart' (Products YourCart) Cart      ⍝ Insert the 'shopping cart' and align it to the right with CSS   
      html,←'div class="clear"'Enclose''    ⍝ Insert <div> tag with the clear:both; CSS property. This is used whenever
     
      req.Return html
    ∇

    ∇ DoAction;x;data

      ⍝ Session Variable/Cookie data gathering
      ⍝ If the session variable 'SessCart' exists, and if it has an expected shape
      ⍝ save it to the Cart field.
      ⍝ If it does not exist, and there is a 'Cart' cookie,
      ⍝ Set the session variable 'Cart' and the Field 'Cart' to the value of the data in the cookie.

      :If 0≠req.Session.State.⎕NC'SessCart'
      :AndIf 2=⍴⍴req.Session.State.SessCart
          Cart←req.Session.State.SessCart
      :ElseIf 0≠⍴cookie←req.GetCookie'CookCart'
          Cart←req.Session.State.SessCart←CookieToData cookie
      :EndIf
     
      ⍝ If a button was pressed, handle it.
      ⍝ If 'Add To Shopping Cart' was pushed, call the AddCartValue method.
      ⍝ If 'Clear Cart', clear the Cart field
      :Select CartAction
      :Case 'Add To Shopping Cart'
          Cart←Cart AddCartValue GetData req.Data~¨' '
      :Case 'Clear Cart'
          Cart←0 2⍴''
      :EndSelect

      ⍝ If the Cart field is an expected shape, set the SessCart to its value.
      ⍝ If the Cart field is not empty, set CookCart to its value. Else, delete the cookie.
      :If 2=2⊃⍴Cart
          req.Session.State.SessCart←Cart
      :AndIf 0≠⊃⍴Cart
         req.SetCookie'CookCart'(DataToCookie Cart)'/' 2
      :Else
         req.DelCookie'CookCart'
      :EndIf 
    ∇

    ∇ html←prods shelf num;in
      ⍝ Generate a section of the shop.
      in←prods[;1]⍳num
      html←'A(n) ',(⍕prods[in;2]),' costs '
      html,←(⍕prods[in;3]),' dollar(s). Buy how many?'
      html←1 2⍴(html)((⍕num)Edit (,'0') 2)
    ∇

    ∇ html←{id}(prods Shop)pnums;in;shop
      ⍝ This widget inserts a shop made up of 'shelves'
      ⍝ id - the id for the container widget
      ⍝ prods - a n×3 matrix (productnumber productname priceperunit) product catalog.
      ⍝ pnums - Product numbers that will be displayed in the shop.
 
      :If 2=⎕NC'id'
          id←'shop'
      :EndIf
      html←BRA'h1'Enclose'My Store'
      shop←prods∘shelf¨pnums
      shop←⍪/shop
      shop←Table shop
      html,←shop
      html,←('CartAction'Submit'Add To Shopping Cart')
      ⍝html,←'CartAction'Submit'Remove From Shopping Cart'
      html←('id'id)#.HTML.div req('post'Form)html
    ∇

    ∇ html←{id} (prods YourCart) cart;carttable
      ⍝ This widget inserts a shopping cart that can be erased.
      :If 2=⎕NC'id'
          id←'cart'
      :EndIf
      html←BRA'h1'Enclose'Your Cart'
      in←prods[;1]⍳cart[;1]
      carttable←((prods[in;2]),[1.1]cart[;2])
      carttable,←prods[in;3]
      carttable,←prods[in;3]×cart[;2]
      carttable←carttable⍪'' '' 'Total:'(+/carttable[;4])
      header←1 4⍴'Item:' 'Quantity:' 'Unit Price:' 'Total Price:'
      html,←BRA Table header⍪carttable
      html,←('CartAction'Submit'Clear Cart')
      html←('id'id)#.HTML.div req('post'Form)html
    ∇

    ∇ r←cartdata SubtractCartValue shopping;in
      shopping[;2]×←¯1
      r←(⊃⍴cartdata)≥in←cartdata[;1]⍳shopping[;1]
      cartdata[(r/in);2]+←r/shopping[;2]
      r←{(0<⊃¨⍵[;2])⌿⍵}cartdata
    ∇

    ∇ r←cartdata AddCartValue shopping;in
      r←(⊃⍴cartdata)≥in←cartdata[;1]⍳shopping[;1]
      cartdata[(r/in);2]+←r/shopping[;2]
      cartdata←cartdata⍪(~r)⌿shopping
      r←{(0≠⊃¨⍵[;2])⌿⍵}cartdata
    ∇

    CookieToData←{GetData{((.5×⍴⍵),2)⍴⍵}{1↓¨(','=⍵)⊂⍵}⍵}
    DataToCookie←{⊃,/{',',⍵}¨⊃,/↓⍕¨⍵}
    ∇ r←GetData data
      r←⊃¨2∘⊃¨⎕VFI¨data~¨' '
      r←(0≠r[;2])⌿r
    ∇

:EndClass