﻿:Namespace mkscr

    Version←4.03

   ⍝ Supported types (PROG is take over by other program - not used yet):
    Types←    'FILL'  'PROG'  'FCTL'    'TEXT'     'BOOL'    'CHAR'     'CLIST'   'RANGE'    'NLIST'     'DATE'     'REGEX'
   ⍝ Corresponding .Net types:
    DNetTypes←'Label' 'PROG' 'TextBox' 'TextBox' 'CheckBox' 'TextBox' 'ComboBox' 'TextBox' 'ComboBox' 'DatePicker' 'TextBox'

    ∇ r←List
      :Access Shared Public
      r←⎕NS¨2⍴⊂⍬
    ⍝ Name, group, short description and parsing rules
      r.Name←'gsis' 'builddd'
      r.Group←⊂'XAML'
      r.Desc←'General Screen Input System' 'Build a new or edit a previous Data Definition'
      r.Parse←'2S -APP -SA  -XAML  -STYLE∊←↑⎕⌹ -WIN= -BGCOL= -TITLE= -SPACING=' ''
    ∇

    ∇ r←Run(Cmd Args);old
      :Access Shared Public
      :Select Cmd
      :Case 'gsis'
          Args.DEBUG←×⎕SE.SALTUtils.DEBUG
          r←MakeForm Args
      :Case 'builddd'
          old←##.THIS⍎Args,(Args∧.=' ')/'{⍵}'
          r←old BuildDD 10 20
      :EndSelect
    ∇

    ∇ r←level Help Cmd;⎕ML;CR
      :Access Shared Public
      ⎕ML←1 ⋄ CR←⎕UCS 13
      :Select Cmd
      :Case 'gsis'
          :If level>0 ⋄ r←Info
          :Else
              r←'Create an input function from data definition in free or XML format.' ''
              r,←⊂'Switches:'
              r,←⊂'-TITLE=   the window title'
              r,←⊂'-BGCOL=   the background color (e.g. Orange)'
              r,←⊂'-WIN=     the window minimum size'
              r,←⊂'-SPACING= the space between elements'
              r,←⊂'-SA       stand alone (no external calls)'
              r,←⊂'-STYLE=   ↑ : labels on top'
              r,←⊂'          ← : labels to the left'
              r,←⊂'          ⎕ or ⌹ : table (grid) format (⌹ for proportional column headings)'
              r,←⊂'-XAML     produce XAML,  not a function'
              r,←⊂'-APP      use Application instead of ShowDialog'
              r←1↓↑,/CR,¨r
          :EndIf
      :Case 'builddd'
          r←⊂'Create or modify an existing Data Definition in XML form'
          r,←⊂'Argument: the name of a variable holding the definition.'
          r,←⊂'Result: new XML definition'
          r←1↓↑,/CR,¨r
      :EndSelect
    ∇

    L←⎕av[3+⎕io]  ⋄  ⎕USING←,⊂''

    _←⍬
    _,←'This command creates some WPF examples from some simple data definitions.'
    _,←L,''
    _,←L,'For example,  doing',L
    _,←L,'      ]gsis Example1 fn',L
    _,←L,'creates function <fn> using the data definition in variable Example1.'
    _,←L,'<fn> can be run to enter data as specified by ''Example1'' which is a simple'
    _,←L,'free form character matrix of up to 6 "columns":'
    _,←L,''
    _,←L,'1. ID. Unique. '
    _,←L,'2. Description. Used as field label     '
    _,←L,'3. Compulsory: boolean (Y/N)'
    _,←L,'4. Type: one of '
    _,←L,' - CHAR : any Unicode char. Allowed set may be specified. '
    _,←L,' - NLIST: number. Allowed set must be specified. '
    _,←L,' - RANGE: number. # decimals, default value and limits must be specified.'
    _,←L,' - BOOL'
    _,←L,' - CLIST: character. Allowed list of choices must be specified. '
    _,←L,' - DATE'
    _,←L,' - REGEX: the input must match a specific pattern.'
    _,←L,' - FCTL : anything else:  Function name to handle case must be specified,'
    _,←L,'5. Length. The size to use in pixels on the screen.'
    _,←L,'   If a size of 0 is specified the fn will calculate the size needed.'
    _,←L,'6. Any parameter to the type used.'
    _,←L
    _,←L,'The input area depens on the label size AND the input size which'
    _,←L,'varies. For a number it is the size of the largest number allowed,'
    _,←L,'for a list it is the widest word in the list.'
    _,←L
    _,←L,'      The command accepts a few switches. See ]?gsis for details. This version'
    _,←L,'does not take the validation parameters into account, it merely produces '
    _,←L,'a fn to perform the input. '
    _,←L,L
    _,←L,'The variable holding the data definition can also be XML. In that case other'
    _,←L,'items like tip and help can be specified.'
    _,←L
    Info←_

    ⎕ex¨ 'L' '_'

    ⎕IO ⎕ML ⎕WX←1 1 3  ⋄  abc←'abcdefghijklmnopqrstuvwxyz'

    If←/⍨ ⋄ Where←{⍵/⍳⍴⍵}
    APLV←'.'⎕wg 'aplversion'
    ⎕fx'r←RSI n'((1+13>⍎2⍴2⊃APLV)⊃'r←(n+1)⊃⎕RSI,#' 'r←⍎(n+1)⊃⎕NSI,''#''')

    ∇ xml←{mat}BuildDD rows;S;n;OK;t;BR;new
     ⍝ Build a new Data Definition from an old one
      :If 0=⎕NC'mat' ⋄ mat←'' ⋄ :EndIf ⍝ default no data
      :If 0=##.⎕NC'THIS' ⋄ S←RSI 2 ⋄ :Else ⋄ S←##.THIS ⋄ :EndIf
      :If 0∊⍴mat ⋄ new←0 12⍴0
      :Else
          mat←S NormalizeDDEF mat
          new←↑mat.(id caption required type default0 decimals0 0 0 parm0 width0 tip0 help)
      :EndIf
      n←1⍴⍴BR←{0}¨new
      :If 0<⍴t←Where 8=Types⍳new[;4] ⍝ RANGE
          new[t;7 8]←↑2⊃¨⎕VFI¨new[t;9] ⋄ new[t;9]←⊂''
      :EndIf
      :Repeat
          :If ~OK←0∊n←1⍴⍴new←new EditDD(rows⌈n×1 2),⊂BR
            ⍝ Validation
              BR←{0}¨new
              new[;1]←new[;1]{' '∧.=v←⍺~' ':'Field',⍕⍵ ⋄ v}¨⍳n
              t←~∧/¨new[;1]∊¨⊂⎕A,abc,⎕D ⋄ BR[Where t;1]←1
              BR[Where 0∊¨⍴¨new[;2];2]←BR[Where 0∊¨⍴¨new[;4];4]←1 ⍝ compulsory fields
              t←(Types⍳new[;4])∊2 3 7 9 11 ⋄ BR[Where t∧0∊¨⍴¨new[;9];9]←1
              t←≠∘⌊⍨new[;6] ⋄ BR[Where t;6]←1
              t←≠∘⌊⍨new[;10] ⋄ BR[Where t;10]←1
              OK←~1∊BR
          :EndIf
      :Until OK
      new[;10]~←0 ⋄ new←⍕¨new ⍝ remove 0 width and turn character
      :If n=0 ⋄ xml←''
      :Else
      ⍝ Recreate the XML
          xml←,/⍉↑'cap' 'width' 'tip' 'help'Tag¨⊂[1]new[;2 10 11 12]
          xml←xml,¨'req'∘Tag¨'NY'[1+'1'∊¨new[;3]]
          n←4/8 9∊⍨t←Types⍳new[;,4] ⍝ RANGE
          new[;5 6 7 8]←n/¨new[;5 6 7 8]
          t←Where,t∊8 ⍝ RANGE
          new[t;9]←{⍺,' ',⍵}/new[t;7 8]
          n←,/⍉↑'default' 'decimals' 'parm'Tag¨⊂[1]new[;5 6 9]
          t←'type id="'∘,¨new[;4],¨'"' ⋄ xml←xml,¨t Tag¨n
          t←'item id="'∘,¨new[;1],¨'"' ⋄ xml←t Tag¨xml
          xml←¯1↓↑,/xml,¨⎕UCS 13
      :EndIf
    ∇

    Tag←{1<≡⍵:⍺∘∇¨⍵ ⋄ ('='∊⍺)<' ' ∧.=⍵:'' ⋄ '<',⍺,'>',(⍕⍵),'</',⍺[⍳¯1+⍺⍳' '],'>'}

    ∇ Data←mat EditDD rows;W;List;app;type;cols;nr;nc;sw;wg;⎕USING;⎕IO;t;i;b;e;max;d;ps;DefRec;Red;red;MW
    ⍝ rows is the number of rows to show at a time (default mat's rows) followed by the total # records.
      ⎕USING←'System.Windows,WPF/PresentationFramework.dll' 'System.Windows.Controls' 'System.Windows,WPF/PresentationCore.dll' 'System.Windows.Media'
⍝ Use a Grid in a ScrollViewer. Enable grid lines for a better view.
      ⎕IO←1
      :If 3>⍴rows ⋄ Red←{0}¨mat
      :Else ⋄ Red←3⊃rows
      :EndIf
      W←⎕NEW Window
      W.Content←sw←⎕NEW ScrollViewer ⋄ sw.Content←wg←⎕NEW Grid ⋄ wg.ShowGridLines←1
      W.Title←'Data Definition builder'
      (nr nc)←0 1+⍴mat  ⍝ nr is the # of records, max is the maximum desired
      (ps max)←1+⌈\nr⌈(2⍴rows)+2↑12×0∊nr+1↑rows ⍝ get a minimum of rows
      W.Height←27×ps ⍝ page size includes header
      W.Width←1088
      wg.RowDefinitions.Add¨⎕NEW¨max⍴RowDefinition
      wg.RowDefinitions.Height←⎕NEW GridLength 25
      wg.ColumnDefinitions.Add¨⎕NEW¨nc⍴ColumnDefinition
      type←TextBox TextBox CheckBox ComboBox TextBox TextBox TextBox TextBox TextBox TextBox TextBox TextBox CheckBox
      cols←'ID' 'Caption' 'Required?' 'Validation type' 'Default value' '# decimals' 'From' 'To' 'Parameter' 'Field width' 'Tip string' 'Help string' 'Delete' ⍝ add a "delete row" box
     
⍝ The header
      {}wg.Children.Add¨t←⎕NEW¨nc⍴Label ⋄ t.Content←cols
      Grid.SetColumn¨t,¨cols←¯1+⍳⍴t ⍝ column headers
      t.Background←⎕NEW SolidColorBrush,Colors.LightGray
      t[2 4].Background←⎕NEW SolidColorBrush,Colors.Orange
      t[⍳12].MinWidth←MW←65 65 49 69 65 65 65 65 65 65 123 65
      t[1].ToolTip←'Unique identifier for this field'
      t[2].ToolTip←'Label of the field'
      t[3].ToolTip←'Is this field compulsory?'
      t[4].ToolTip←'Field data Type'
      t[5].ToolTip←'Value used when none entered'
      t[6].ToolTip←'Max number of decimals accepted'
      t[7].ToolTip←'Range lowest value'
      t[8].ToolTip←'Range highest value'
      t[9].ToolTip←'Type parameter'
      t[10].ToolTip←'Width of the field (integer)'
      t[11].ToolTip←'Field Tip string'
      t[12].ToolTip←'Long description (UNUSED)'
     
⍝ Fill the grid
      red←⎕NEW SolidColorBrush,Colors.Red
      :For i :In ⍳max-1
          {}wg.Children.Add¨t←⎕NEW¨type
          :If d←i≤nr ⍝ do we have Data?
              :If ∨/b←Red[i;],0 ⋄ (b/t).Background←red ⋄ :EndIf
              b←mat[i;3],0 ⋄ e←mat[i;1 2 9 11 12]
          :Else
              b←0 ⋄ e←⊂''
          :EndIf
          t[3 13].IsChecked←b ⍝ booleans
          t[1 2 9 11 12].Text←e
          Grid.SetRow¨t,¨i ⋄ Grid.SetColumn¨t,¨cols
          {}(t[4]).Items.Add¨⊂¨List←Types∪e←mat[d/⊂i 4]
          t[4].SelectedIndex←1↑¯1,⍨¯1+List⍳e
          t[b].Text←⍕¨mat[d/i,¨b←5 6 7 8 9 10],d↓⊂''
          t[3 13].VerticalAlignment←VerticalAlignment.Center ⋄ t[3 13].HorizontalAlignment←HorizontalAlignment.Center
      :EndFor
     
      DefRec←¯1↓t ⍝ the last addition is also a default record if lines were added
     
    ⍝ Specify proportional width of each column
      (⌷wg.ColumnDefinitions).Width←{⎕NEW GridLength(⍵ GridUnitType.Star)}¨MW,55
     
      W.Hide
      {}W.ShowDialog  ⍝ no buttons, upper right X will close the form
      →0/⍨0∊⍴Data←{0}¨t←0 ¯1↓(b←0=t[;nc].IsChecked)⌿t←1 0↓max nc⍴⌷wg.Children
      t⍪←DefRec ⍝ add to the input to find properties
      t[;i]←t[;i←3~⍨⍳12].Text
      t[;3]←t[;3].IsChecked
      t[;i]←{+/2⊃⎕VFI ⍵}¨t[;i←5 6 7 8 10]
      Data←((0,⍨nr≥b/⍳max-1)∨t∨.≢t[⍬⍴⍴t;])⌿t ⍝ remove training default records
    ∇

    ∇ CloseAppDomain
      1 ⎕NQ'⎕se.mb.file.closead' 'select'
    ∇

    ∇ r←{obj}FindTextSize text;coord
     ⍝ Find width of each text vector
      text←⊂⍣(326≠⎕DR text)+text
      :If 0=⎕NC'obj' ⋄ obj←# ⋄ :EndIf
      coord←obj.Coord ⋄ obj.Coord←'Pixel'
      r←obj.GetTextSize¨text
      obj.Coord←coord
    ∇

    ∇ r←FindTextWidth text
     ⍝ Find width of each text vector
      r←2⊃¨FindTextSize text
    ∇

    ∇ xml←FreeToXML dd;cut;section;tag;n;ityp;parms;type;b;i;rtb;rlb
     ⍝ Transform a free form data definition into XML.
      ⎕ML←1
      cut←{b∧6≥+\b←⍵<1,¯1↓⍵}' '∧.=dd←{(⍵∨.≠' ')⌿⍵}⎕FMT dd ⍝ cut into at most 6 parts
      rtb←{(-+/∧\' '=⌽⍵)↓⍵} ⋄ rlb←{(+/∧\' '=⍵)↓⍵}
      section←cut⊂dd ⍝ ID, description, compulsory, type, length and params
      tag←{0∊⍴⍵:'' ⋄ 1<≡⍵:⍺∘∇¨⍵ ⋄ b a←'<',¨'' '/',¨⊂⍺,'>' ⋄ b,(rtb ⍵),a}
      xml←,⌿↑'cap' 'req' 'width'tag¨↓¨section[2 3 5]
      n←⊃⍴ityp←Types⍳type←~∘' '¨↓Ucase 4⊃section
      parms←n 3⍴⊂''
      :If 0<⍴i←Where(ityp∊6 7)∨b←ityp∊8 9 ⍝ those with parameters
      :AndIf 6=+/cut ⍝ parameters must be there
          parms[i;]←↑b[i]{⍺≤' '∧.=⍵:'' ''(rlb ⍵)
              rtb¨{⍵⊂⍨b∧3≥+\b←b<1,¯1↓b←⍵=' '}rlb ⍵}¨↓(6⊃section)[i;]
      :EndIf
      parms←,⌿↑'decimals' 'default' 'parm'tag¨↓⍉parms
      tag←{'<',⍺⍺,' id="',(⍺~' '),'">',⍵,'</',⍺⍺,'>'}
      xml←xml,¨type('type'tag)¨parms
      xml←↑(↓1⊃section)('item'tag)¨xml
    ∇

      Lcase←{(s u)←(⍴⍵)(,⍵) ⋄ (U L)←⎕A'abcdefghijklmnopqrstuvwxyz'
          ~∨/b←u∊U:⍵ ⋄ (b/u)←L[U⍳b/u] ⋄ s⍴u}

    ∇ code←parms MAKEGRID DD;⎕IO;code;Typ;Q;i;t;fmt;fnname;⎕ML;b;NX;w;NoN;widthpos;def;SD
     ⍝ Make a screen from Data Definition. In  GRID  form.
     ⍝ parms provides other arguments. See <MakeForm> for details.
     
      ⎕ML←⎕IO←1
      (NX SD)←~parms.(XAML APP) ⍝ non XAML generation mode and ShowDialog mode
     
      fnname←2⊃parms.Arguments ⍝ function name
      code←⊂'Data←{mat}',fnname,' rows;W;List;app;type;cols;nr;nc;sw;wg;⎕using;⎕io;t;i;b;e;max;d;ps',NX/';DefRec'
      code,←⊂'⍝ Built automatically on ',⍕100⊥⍉2 3⍴⎕TS
      code,←⊂'⍝ rows is the number of rows to show at a time (default mat''s rows) followed by the total # records.'
      code,←⊂':if 0=⎕nc''mat'' ⋄ mat←0 ',(t←⍕⍴DD),'⍴0 ⋄ :endif ⍝ default no data'
      code,←⊂'⎕signal 5/⍨',t,'≠¯1↑⍴mat ⍝ check # cols corresponds to argument'
     
     ⍝ The DLLs needed
      t←' ''System.Windows,WPF/PresentationCore.dll'' ''System.Windows.Media''',parms.XAML/' ''System.Windows.Markup'''
      code,←⊂'⎕USING←''System.Windows,WPF/PresentationFramework.dll'' ''System.Windows.Controls''',t
      code,←(NX>SD)↑parms.SA↓⎕NR'CloseAppDomain'
      code,←⊂'⍝ Use a Grid in a ScrollViewer. Enable grid lines for a better view.'
      code,←'⎕io←1' 'W←⎕new Window' 'W.Content←sw←⎕new ScrollViewer ⋄ sw.Content←wg←⎕NEW Grid ⋄ wg.ShowGridLines←1',parms.DEBUG/⊂'W.Show'
      :If 0≢t←parms.BGCOL
          code,←⊂'wg.Background←⎕new SolidColorBrush,Colors.',(Ucase 1↑t),1↓t
      :EndIf
      code,←⊂1⌽'''W.Title←''',parms.{⍵ Switch'TITLE'}fnname
     
     ⍝ Grid format
      code,←'' '(nr nc)←0 1+⍴mat  ⍝ nr is the # of records, max is the maximum desired'
     
     ⍝ Find width of each column. These numbers represent the PROPORTIONAL value of each column.
     ⍝ ⎕NC does not report 0 for fields without a value so the class sets unknown values to ⎕NULL
      b←0,⍨0=w←{b←⎕NULL≢¨v←0,⍵ ⋄ 1↓b\b/v}DD.width ⍝ take specified width into consideration
      Typ←5,⍨Types⍳DD.type ⍝  add BOOL at the end
      w[Where b∧Typ=10]←FindTextWidth'2011-08-08 ⎕⎕⎕' ⍝ default width for dates
      t←⍕w←w⌈FindTextWidth DD.caption   ⍝ record width of headers
      code,←'(ps max)←1+⌈\nr⌈rows+2↑12×0∊nr+1↑rows ⍝ get a minimum of rows' 'W.Height←27×ps ⍝ page size includes header't
      widthpos←⍴code ⍝ note where we set the width
      code,←⊂'wg.RowDefinitions.Add¨⎕NEW¨max⍴RowDefinition'
      code,←⊂'wg.RowDefinitions.Height←⎕NEW GridLength 25'
      code,←⊂'wg.ColumnDefinitions.Add¨⎕NEW¨nc⍴ColumnDefinition'
     
      Q←{q,((1+q=⍵)/⍵),q←''''}
      code,←⊂'type←',⍕DNetTypes[Typ]
      code,←⊂'cols←',(⍕Q¨DD.caption),' ''Delete'' ⍝ add a "delete row" box'
     
     ⍝ Column headers will be made up from DD ids.
      code,←'' '⍝ The header'
      code,←'{}wg.Children.Add¨t←⎕new¨nc⍴Label ⋄ t.Content←cols' 'Grid.SetColumn¨t,¨cols←¯1+⍳⍴t ⍝ column headers'
      code,←⊂'t.Background←⎕new SolidColorBrush,Colors.LightGray'
      i←⍕Where b←DD.required
      code,←(∨/b)/⊂'t[',i,'].Background←⎕new SolidColorBrush,Colors.Orange'
      :If 0<⍴i←⍕Where t←w>0 ⋄ code,←⊂'t[',i,'].MinWidth←',⍕t/w ⋄ :EndIf
     
     ⍝ Tooltips
      :If 0<⍴i←Where' '∨.≠¨DD.tip
          code,←i{'t[',(⍕⍺),'].ToolTip←',⍕Q ⍵}¨DD[i].tip
      :EndIf
     
     ⍝ The contents
      code,←'' '⍝ Fill the grid' ':for i :in ⍳max-1' '{}wg.Children.Add¨t←⎕new¨type'
      code,←⊂'Grid.SetRow¨t,¨i ⋄ Grid.SetColumn¨t,¨cols'
      code,←NX/':if d←i≤nr ⍝ do we have Data?'('b←',((1<⍴i)/'mat[i;',(⍕¯1↓i←Where Typ=5),'],'),'0 ⋄ e←mat[i;',(t←⍕Where Typ∊3 4 6 10),']')
      code,←NX 1 NX/':else' 'b←0 ⋄ e←⊂''''' ':endif'
      code,←⊂'t[',(⍕i),'].IsChecked←b ⍝ booleans'
      code,←⊂'t[',t,'].Text←e'
     
     ⍝ Fill in the comboboxes
      :For i :In Where 7=Typ  ⍝ char lists
         ⍝ The parm var contains the strings
          code,←⊂'{}(t[',t,']).Items.Add¨⊂¨List←',(⍕Q¨b←DD[i].parm),NX/'∪e←mat[d/⊂i ',(t←⍕i),']'
          code,←NX/⊂'t[',t,'].SelectedIndex←1↑¯1,⍨¯1+List⍳e'
      :EndFor
      NoN←~∘⎕NULL
      :For i :In Where 8=Typ  ⍝ numbers
         ⍝ The parm var contains the range
          fmt←⍕NoN DD[i].decimals
          code,←NX/⊂'t[',t,'].Text←',fmt,'⍕d↓0,mat[d/⊂i ',(t←⍕i),']'
      :EndFor
      :For i :In Where Typ∊9  ⍝ num lists
          fmt←⍕NoN DD[i].decimals ⋄ def←NoN DD[i].default
          code,←⊂'{}(t[',t,']).Items.Add¨⊂¨',fmt,'⍕¨List←',(⍕def,DD[i].parm),NX/'∪e←mat[d/⊂i ',(t←⍕i),']'
         ⍝ If there is a value we add and show it, if none we display the default, if any
          code,←NX/⊂'t[',t,'].SelectedIndex←1↑',(4××⍴def)↓'¯1,⍨¯1+List⍳e'
      :EndFor
     
     ⍝ Center the CheckBoxes
      i←⍕Where 5=Typ
      t←'Vertical' 'Horizontal'
      code,←(0<⍴i)/{⍺,'⋄',⍵}/i∘{'t[',⍺,'].',⍵,'Alignment←',⍵,'Alignment.Center'}¨t
     
     ⍝ Close the loop, prepare for call
      code,←':endfor' ''
      code,←NX/'DefRec←¯1↓t ⍝ the last addition is also a default record if lines were added' ''
      code[widthpos]←⊂'W.Width←',⍕+/24+w
      :If '⌹'∊parms.STYLE
          code,←'⍝ Specify proportional width of each column'('(⌷wg.ColumnDefinitions).Width←{⎕NEW GridLength(⍵ GridUnitType.Star)}¨',⍕w,20)''
      :Else
          code,←'⍝ Specify automatic width of each column' '(⌷wg.ColumnDefinitions).Width←⎕NEW GridLength(1 GridUnitType.Auto)' ''
      :EndIf
     ⍝ Add XAML generation code if wanted
      :If parms.XAML
          code,←⊂'Data←XamlWriter.Save W⋄W.Close'
      :Else
          code,←runmethod parms.(DEBUG APP)
         ⍝ Create the code for retrieving the data
          code,←⊂'→0/⍨0∊⍴Data←{0}¨t←0 ¯1↓(b←0=t[;nc].IsChecked)⌿t←1 0↓max nc⍴⌷wg.Children'
          code,←⊂'t⍪←DefRec ⍝ add to the input to find properties'
         ⍝ There are 4 .Net types
          :If 0<⍴i←⍕Where t←¯1↓Typ∊3 4 6 7 8 9 10
              i←(~∧/t)/'[;',i,']'
              code,←⊂'t',i,'←t',i,'.Text'
          :EndIf
          :If 0<⍴i←⍕Where t←¯1↓Typ∊5
              i←(~∧/t)/'[;',i,']'
              code,←⊂'t',i,'←t',i,'.IsChecked'
          :EndIf
          :If 0<⍴i←⍕Where t←¯1↓Typ∊8 9
              code,←⊂'t',i,'←{+/2⊃⎕vfi ⍵}¨t',i←(~∧/t)/'[;',i,']'
          :EndIf
          code,←⊂'Data←((0,⍨nr≥b/⍳max-1)∨t∨.≢t[⍬⍴⍴t;])⌿t ⍝ remove training default records'
      :EndIf
    ∇

    ∇ code←runmethod(deb app)
     ⍝ Find code to run form
      :If ~app
          code←(~deb)↓'W.Hide' '{}W.ShowDialog  ⍝ no buttons, upper right X will close the form'
      :Else
          code←⊂'app←⎕new Application'
          code,←⊂'{}app.Run W ⍝ no buttons, upper right X will close the form'
      :EndIf
    ∇

    ∇ code←parms MAKESW DD;vert;W;⎕IO;code;sz;Typ;Q;i;t;width;minset;fmt;fnname;⎕ML;b;d;NoN;mp;mb
     ⍝ Make a screen from Data Definition. Sequential Version.
     ⍝ parms provides other arguments. See <MakeForm> for details.
     
      ⎕ML←⎕IO←1 ⋄ vert←∨/'↑='∊parms.STYLE
      fnname←2⊃parms.Arguments ⍝ fn name
     
     ⍝ Generate code - header
     ⍝ We do not generate code for utilities unless they do not exist in the ws or it's been requested
      (mp mb)←t←parms.SA∨3≠parms.Source.⎕NC↑'makePair' 'bjp'
      code←⊂'Data←',fnname,';W;app;ctls;min;label;sz;t;type;nr;wp;⎕using;⎕io',∊t/';makePair' ';bjp'
      code,←⊂'⍝ Built automatically on ',⍕100⊥⍉2 3⍴⎕TS
     ⍝ Set ⎕USING, close DLLs
      t←(0≢parms.BGCOL)/' ''System.Windows,WPF/PresentationCore.dll'' ''System.Windows.Media'''
      t,←parms.XAML/' ''System.Windows.Markup'''
      code,←⊂'⎕USING←''System.Windows,WPF/PresentationFramework.dll'' ''System.Windows.Controls''',t
      code,←parms.(XAML<APP)↑parms.SA↓⎕NR'CloseAppDomain'
      code,←'⎕io←1' 'W←⎕new Window ⋄ W.Content←wp←⎕NEW WrapPanel',parms.DEBUG/⊂'W.Show'
      :If 0≢t←parms.BGCOL
          code,←⊂'wp.Background←⎕new SolidColorBrush,Colors.',(Ucase 1↑t),1↓t
      :EndIf
      code,←⊂1⌽'''W.Title←''',parms.{⍵ Switch'TITLE'}fnname
      :If mp ⋄ code,←⎕NR'makePair' ⋄ :EndIf  ⍝ stand-alone
     
      t←',⊂'If 1∊⍴Typ←Types⍳DD.type
      Q←{q,((1+q=⍵)/⍵),q←''''}
      code,←⊂'type←',t,⍕Q¨DNetTypes[Typ]
      code,←⊂'label←',t,⍕Q¨DD.caption
      minset←⍴code←code,⊂'min←?'  ⍝ we'll replace that line later
     
     ⍝ Determine approximate size
     ⍝ SPACING specifies MINimum pixels between fields
      t←(⍕vert),' makePair¨↓label,',(⍕parms.SPACING),',',(⍕DD.required),',type,[1.1]min'
      code,←⊂'{}wp.Children.Add¨ctls←',t,' ⍝ add each field to the form'
      code,←⊂(~parms.DEBUG)/'W.(Top Left)←-999⋄W.Show ⍝ move out of the way to calculate sizes'
      t←parms.(DEBUG<XAML)/'⋄W.Hide'
      code,←⊂'W.UpdateLayout',t,' ⍝ ensure we get accurate numbers next'
      code,←mb/⎕NR'bjp'
      code,←⊂'sz←20+t,⍨',(~vert)↓'190×+/sz bjp⍨t←⌈/sz,⌈5×{(+/⍵)÷⍴⍵}sz←ctls.ActualWidth'
      code,←⊂'W.(Top Left)←333',⍨(~parms.DEBUG)/'W.Hide⋄'
      code,←⊂'W.(Height Width)←sz'
      :If 0≢t←parms.WIN ⋄ code,←⊂'W.(MinHeight MinWidth)←',t ⋄ :EndIf
     
      width←DD.width
      NoN←~∘⎕NULL
     
     ⍝ Find the max width for RANGEs
      :For i :In Where 8=Typ  ⍝ RANGEs
         ⍝ The parm var contains the numbers
          :If width[i]=0
              :If 0=⍴t←NoN DD[i].decimals ⋄ width[i]←2×FindTextWidth⍕¯1↑DD[i].parm
              :Else ⋄ width[i]←2×FindTextWidth t⍕¯1↑DD[i].parm
              :EndIf
          :EndIf
          :If 0<⍴t←NoN DD[i].default ⋄ code,←⊂'(2⊃⌷ctls[',(⍕i),'].Children).Text←⍕',⍕t ⋄ :EndIf
      :EndFor
     ⍝ Fill in the comboboxes
      :For i :In Where 7=Typ  ⍝ char lists
         ⍝ The parm var contains the strings
          :If width[i]=0 ⋄ width[i]←30+↑⌈/FindTextWidth¨DD[i].parm ⋄ :EndIf
          t←Q⍣parms.SA¨DD[i]⍎'parm',parms.SA↓'0'
          code,←⊂'{}(2⊃⌷ctls[',(⍕i),'].Children).Items.Add¨⊂¨',⍕t
      :EndFor
      :For i :In Where Typ∊9  ⍝ num  lists
          :If width[i]=0 ⋄ width[i]←30+↑⌈/FindTextWidth¨⍕¨DD[i].parm ⋄ :EndIf
          fmt←⍕NoN DD[i].decimals
          t←DD[i]⍎'parm',parms.SA↓'0'
          :If b←0<⍴d←NoN DD[i].default
              :If ~parms.SA ⋄ t←'(',t,'),',⍕d
              :Else ⋄ t←{⍵[⍋⍵]}t∪d ⋄ :EndIf
          :EndIf
          code,←⊂'{}(t←2⊃⌷ctls[',(⍕i),'].Children).Items.Add¨⊂¨',fmt,'⍕¨',⍕t
          code,←b/⊂'t.SelectedIndex←',⍕¯1+t⍳d
      :EndFor
     ⍝ Center the CheckBoxes
      :For i :In Where 5=Typ
          t←(1+vert)⊃'Vertical' 'Horizontal'
          code,←⊂'(2⊃⌷ctls[',(⍕i),'].Children).',t,'Alignment←',t,'Alignment.Center'
      :EndFor
     
      width+←60×(0=width)∧Typ∊4 6 ⍝ char fields
     
     ⍝ Replace the min size line
      code[minset]←⊂'min←',⍕width
     
     ⍝ Tooltips
      :If 0<⍴i←Where DD.tip≢¨⎕NULL
      :AndIf 0<⍴i←(' '∨.≠¨DD[i].tip)/i
          code,←i{'(2⊃⌷ctls[',(⍕⍺),'].Children).ToolTip←',⍕Q ⍵}¨DD[i].tip
      :EndIf
     
     ⍝ Add XAML generation code if wanted
      :If parms.XAML
          code,←⊂'Data←XamlWriter.Save W⋄W.Close'
      :Else
          code,←runmethod parms.(DEBUG APP)
     
         ⍝ Create the code for retrieving the data
          code,←⊂'Data←2⊃¨⌷¨ctls.Children'
         ⍝ There are 4 .Net types
          :If 0<⍴i←⍕Where t←Typ∊3 4 6 7 8 9 10 11
              i←(~∧/t)/'[',i,']'
              code,←⊂'Data',i,'←Data',i,'.Text'
          :EndIf
          :If 0<⍴i←⍕Where t←Typ∊5
              i←(~∧/t)/'[',i,']'
              code,←⊂'Data',i,'←Data',i,'.IsChecked'
          :EndIf
      :EndIf
     
     ⍝ End
    ∇

    ∇ r←MakeForm DD_FN;T;⎕IO;arg;FN;DDEF;S;Code;local;t;grid
     ⍝ DanB 2011 CREATE FULL SCREEN input function from data definition in variable (1st arg)
     ⍝ possibly followed by the fn name to create.
     ⍝ The fn produced will perform the input of N fields.
     ⍝ It won't require any user written fn to exist if so desired.
     
     ⍝ This fn can  be used directly from the namespace or as a user command
     
     ⍝ 'DD_FN' is the DD name followed by the FnName  and 0 or more of the following switches
     ⍝ TITLE   the window title
     ⍝ STYLE   specifies behaviour:
     ⍝      ↑: Put Description over box (Vertical)
     ⍝      ←: Put Description to left of box (DEFAULT)
     ⍝ ⌹ or ⎕: Use a grid to enter data (⌹ is proportional width headings)
     ⍝ WIN   the window minimum size desired
     ⍝ SPACING between items
     ⍝ BGCOL   the background color
     ⍝ SA      stand alone (no external names used)
     ⍝ XAML    produce XAML code instead
      ⎕IO←1
      :If 0<t←⍴⍴arg←DD_FN  ⍝ is this a pre parsed package?
          ⎕←'VERSION ',⍕Version
          :If (t=1)>'<'∊1↑arg
              S←'-BGCOL= -APP -SA -SPACING= -STYLE∊←↑⌹⎕ -TITLE= -WIN= -DEBUG -XAML'
              arg←(⎕NEW ⎕SE.Parser(S'upper nargs=2s')).Parse DD_FN
          :Else ⍝ must be rank 2
              arg←⎕NS'' ⋄ arg.(BGCOL APP SA SPACING STYLE TITLE WIN DEBUG XAML)←0
              arg.Arguments←((⍕⎕THIS),'.DD_FN')'ExampleFn'
              arg.Switch←{⍺←0 ⋄ ⍺}
          :EndIf
          S←RSI 1
      :Else ⋄ S←##.THIS
      :EndIf
      local←arg.XAML ⍝ tell MAKESW to generate code to make the form?
      'Too few arguments'⎕SIGNAL 11⍴⍨0∊⍴t←arg.Arguments
      (T FN)←arg.Arguments←2↑t,⊂'ExampleFn' ⋄ T←T~' '
      ('<',FN,'> is an invalid fn name')⎕SIGNAL 11 If⍱/0 3∊S.⎕NC FN
     ⍝ The data definition may come in several formats: XML string or as free format character matrix
      DDEF←S NormalizeDDEF S⍎T
     
      arg.Source←S ⍝ where we were called from
     
     ⍝ The window max size may be 2 #s (window centered)
      arg.(SPACING←12 Switch'SPACING')   ⍝    min spacing
      :If grid←∨/'⎕⌹'∊arg.STYLE ⋄ Code←arg MAKEGRID DDEF
      :Else ⋄ Code←arg MAKESW DDEF
      :EndIf
     ⍝ ⎕SHADOW local/FN
      r←S.⎕FX Code
      :If arg.XAML
          r←⎕XML⍣2+S⍎FN,grid/' 1' ⍝ grid fns take an arg
      :Else
          {3∊10|⎕DR ⍵:÷⎕←'Unable to create function' ⋄ z←⍵}r
      :EndIf
    ∇

    ∇ ndd←{source}NormalizeDDEF dd;cdd;source;char;datum
     ⍝ Transform a data definition into a more suitable format.
     ⍝ This fn accepts a few formats: free format & XML.
      :If 2≡≡dd
          dd←0 1↓⍕↑{b←'_'=v←⍵ ⋄ (b/v)←' ' ⋄ v⊂⍨b<¯1↓1,b←' '=⍵}¨dd ⍝ accept VTVs
      :EndIf
      :If char←isChar dd
      :AndIf '<'≠1⍴dd←⎕FMT dd ⍝ we have a simple definition (char matrix)
          dd←FreeToXML dd
      :EndIf
     ⍝ Now either XML or APL
      :If char
          dd←⎕XML,dd  ⍝ XML string
      :EndIf
      cdd←(0=dd[;1])⊂[1]dd ⋄ ndd←⍬
      :If 0=⎕NC'source' ⋄ source←RSI 2 ⋄ :EndIf
      :For datum :In cdd
          ndd,←⎕NEW DataField,⊂source datum
      :EndFor
    ∇

      Ucase←{(s u)←(⍴⍵)(,⍵)
          (U L)←⎕A'abcdefghijklmnopqrstuvwxyz'
          ~∨/b←u∊L:⍵ ⋄ (b/u)←U[L⍳b/u] ⋄ s⍴u}

    bjp←{(a↑1)⌹(∘.=⍨⍳a←⍴,⍵)-<⍀b∘.>⍺+¯1↓0,b←+\,⍵} ⍝ BlackJack Partionning

    isChar←{∨/0 2=10|⎕DR ⍵}

      makePair←{txt space color type min←⍵
        ⍝ arg is caption, spacing, color (1=warn, 2=bad), .Net type and min width
          colpath←(color>0)/'System.Windows,WPF/PresentationCore.dll' 'System.Windows.Media'
          ⎕USING←'System.Windows,WPF/PresentationFramework.dll' 'System.Windows.Controls',colpath
          g←⎕NEW StackPanel ⋄ g.Margin←⎕NEW Thickness space
          ⍺←0 ⍝ vertical is 1
          g.Orientation←Orientation.(Horizontal Vertical)[1+⍺]
          t←g.Children.Add¨lx←⎕NEW¨Label(⍎type)
          lx[1].Content←txt{⍺}lx[2].MinWidth←min
          g{⍺}color{⍺=0:0 ⋄ ⍵.Background←⎕NEW SolidColorBrush,Colors⍎⍺⊃'Orange' 'Red'}lx[2]
      }

    xCut←{⍺←1↑⍵ ⋄ 1↓¨(⍵∊⍺)⊂⍵}

    :Class DataField
⍝ Create a ns containing all info.
        ⎕io←1
⍝ First the always there info:

        :field public    id
        :field public    required
        :field public    caption
        :field public    type
        :field public    tip
        :field public    tip0
        :field public    help
        :field public    width
        :field public    width0
        :field public    chkStmt←''       ⍝ validation statement

⍝ then the possibly absent info
        :field public    default←⎕null
        :field public    default0←''
        :field public    decimals←⎕null   ⍝ number of decimals
        :field public    decimals0←''
        :field public    parm←⎕null       ⍝ the interpreted parameter
        :field public    parm0←''         ⍝ the original parameter string

        ntypes←'NLIST' 'RANGE' 'CHAR' 'CLIST' 'REGEX'
        get←{⍺←'' ⋄ ⊃(datum[;3]⍪⊂⍺)[datum[;2]⍳⊂⍵]}

        ∇ boa(source datum);n;b;val;extra;fix;T;s;rep;ndec;t
          :Access public
          :Implements constructor
          id←(1 4)(1 2)⊃datum
          required←∨/'yY'∊get'req'
          caption←get'cap'
          help←get'help'
          width←⍎'0',width0←get'width'
          n←datum[;2]⍳⊂'type'
          T←⊂type←(n 4)(1 2)⊃datum
          s←(T∊'CHAR' 'TEXT')/'Enter any character'
         ⍝ The data type
          :If 1∊b←∧\2=n↓datum[;1]   ⍝ any parameters?
              fix←source∘⍎⍣(T∊ntypes)
              parm0←get'parm'       ⍝ the original string
              chkStmt←(T≡⊂'FCTL')/parm0,' '
              val←fix¨b/n↓datum[;3]
              extra←b/n↓datum[;2]
              extra{⍎⍕⍺,'←⍵'}¨val
              :If 4=n←ntypes⍳T      ⍝ CLIST. No validation required.
                  :If width=0 ⋄ width←##.FindTextWidth(n⍳⌈/n←∊⍴¨parm)⊃parm ⋄ :EndIf
              :ElseIf 3=n           ⍝ CHAR
              :AndIf extra∊⍨⊂'parm'
                  s←1⌽'"Enter characters in the set "',parm
                  chk←parm∘{(0<⍴s)/'Invalid characters: ',s←∪⍵~⍺}
                  chkStmt←'{(0<⍴s)/''Invalid characters: '',s←∪⍵~',parm0,'}'
              :ElseIf 2=n           ⍝ RANGE
                  b←extra∊⊂'decimals' ⋄ ndec←(b⍳1)⊃val,0 ⋄ b←∨/b
                  decimals0←get'decimals'
                  rep←⍕¨(extra⍳⊂'parm')⊃val ⍝ 2 numbers
                  width{⍺=0:⍵ ⋄ ⍺}←##.FindTextWidth 2⊃rep
                  t←(1+b∧ndec=0)⊃'a number' 'an integer'
                  s←'Enter ',t,' from ',(1⊃rep),' to ',2⊃rep
                  chk←parm∘chkRange
                  t←'0∊⍴b:''''⋄((v<',(⍕parm[1]),')∨v>',(⍕parm[2]),'):''Number outside bounds''}'
                  chkStmt←'{⎕io←1⋄0∊1⊃b v←⎕VFI ⍵:''Invalid number''⋄1<⍴b:''Too many numbers''⋄',t
              :AndIf b∧ndec≠0
                  s,←' with max ',(⍕n←|ndec),' decimals'
                  chk←(parm,n)∘chkRange
                  chkStmt←(¯1↓chkStmt),'⋄''Too many decimals''/⍨',(⍕n),'<⍴(v⍳''.'')↓v←v/⍨~∨\''eEjJ''∊⍨v←⍵~'' ''}'
              :ElseIf n=1           ⍝ NLIST. No validation required.
                  :If width=0 ⋄ width←##.FindTextWidth⍕⌈/parm ⋄ :EndIf
              :ElseIf n=5           ⍝ REGEX
                  chk←parm∘{'Invalid string'/⍨(,0)≢⍺⎕S  0+⍵}
                  chkStmt←'{''Invalid string''/⍨(,0)≢',parm0,' ⎕S 0+⍵}'
              :EndIf
          :EndIf
          tip←s{0∊⍴⍵:⍺ ⋄ ⍵}tip0←get'tip'
        ∇

        ∇ r←Validate txt
          :Access public
          r←chk txt
        ∇
          chkRange←{⎕IO←1 ⍝  ⍺: from, to  [,ndec]
              0∊1⊃b v←⎕VFI ⍵:'Invalid number'
              1<⍴b:'Too many numbers'
              0∊⍴b:''
              ((v<⍺[1])∨v>⍺[2]):'Number outside bounds'
              3>⍴⍺:''
              v←v/⍨~∨\'eEjJ'∊⍨v←⍕v
              (⍺[3]<⍴(v⍳'.')↓v)/'Too many decimals'}

    :EndClass
:EndNamespace