﻿:namespace Monitor
⍝  Implement "Monitor", "APLMON", and "CPUTime" User Commands.  V2.01

    ⎕IO←1 ⋄ ⎕DIV←⎕ML←0 ⍝ Set those here to avoid inheriting them from outside

    ∇ r←lev Help Cmd
      :Access Shared Public
      :Select Cmd
      :Case 'Monitor'
          r←'monitor [expression]' '' 'switches:'
          r,←⊂'-on            switches monitoring on'
          r,←⊂'-off           switches monitoring off'
          r,←⊂'-report        displays monitor report'
          r,←⊂'-caption       report''s caption'
          r,←⊂'-fns=fn1,fn2   only for named fns'
          r,←⊂'-top=n         only report on top n fns'
          r,←⊂'-min=n         only report lines/fns consuming n% (default 1%)'
          r←↑r
     
      :Case 'APLMON'
          r←'aplmon [expression]' '' 'switches:'
          r,←⊂'-file=filename   (only required when turing aplmon on)'
          r,←⊂'-on            switches monitoring on'
          r,←⊂'-off           switches monitoring off'
          r←↑r
     
      :Case 'CPUTime'
          r←'cputime expression' '' 'Switches:'
          r,←⊂'-repeat=      repeat expression N times or until N Msec have gone by'
          r,←⊂'-details[=]   show detailed info'
          :If lev>0
              r,←'' '-repeat accepts the number of iterations or time to spend when followed'
              r,←⊂'by "s" (for seconds) or "m" (milisecs).'
              r,←⊂'It can also accept a combination of both iterations and period, for ex the'
              r,←⊂'maximum between 10 iterations and 1000 milisecs. If 1 second is not enough'
              r,←⊂'to run the expression 10 times it repeats until the expression has been'
              r,←⊂'executed 10 times. On the other hand if the expression ran 10 times in'
              r,←⊂'less than 1 sec it continues to run until 1 sec has gone by. It would'
              r,←⊂'be specified this way: -repeat=10 ⌈ 1s'
              r,←'' '-details or -details=all displays other info:'
              r,←⊂'- the ⎕AI and ⎕MONITOR CPU and elapsed numbers'
              r,←⊂'- the ⎕MONITOR results of running the test function'
              r,←⊂'  -details=none simply returns the numeric results (milisecs) as a matrix of Nx2'
              r,←⊂'  -details=ai returns the numeric results of ⎕MONITOR and ⎕AI as a matrix of Nx4'
          :EndIf
          r←↑r
     
      :EndSelect
    ∇

    ∇ r←List;nam ⍝ this function returns 1 or more namespaces (here 3)
      :Access Shared Public
      r←,⎕NS¨0⍴¨nam←'Monitor' 'APLMON' 'CPUTime' ⍝ create N commands
      r.Name←nam                     ⍝ their names
      r.Desc←'Set/Report ⎕MONITOR' 'Turn APLMON on or off' 'Time execution of an expression' ⍝ descriptions
      r.Group←⊂'SysMon'              ⍝ same group
      r[1].Parse←' -fns= -top= -caption= -min= -on -off -report'
      r[2].Parse←' -file= -on -off'
      r[3].Parse←' -repeat= -details[=]all none ai'
    ∇

    ∇ r←Run(Cmd Args);cs;fns;top;min;switch;expr;cap;file;ai;on;off;report;clear;repeat;n;path;exp;f;minrun;t;eex;runfor;fn;branch;cond;results;all;tell;detail;ran;mondata;reparg;showai;showresult;done;cnt
      :Access Shared Public
     
      cs←##.THIS ⍝ Calling Space
     
      :Select Cmd
      :Case 'Monitor'
          (on off report clear top)←Args.(on off report clear top)
          fns←{(⊂'')~⍨1↓¨(⍵∊' ,')⊂⍵}',',Args.fns~0
          :If 0∊⍴fns ⋄ fns←cs.⎕NL ¯3 ⋄ :EndIf
          min←0.01×1 Args.Switch'min' ⍝ Minimum 1% by default
          cap←''Args.Switch'caption'
     
          :If 0≠⍴Args.Arguments ⍝ Non-empty expression
              expr←⊃,/' ',¨Args.Arguments
              ⎕←1 MonSet cs fns
              ⎕←'Running: ',expr
              cs⍎expr
              r←MonRep cs fns top min expr
          :ElseIf on∨off
              r←on MonSet cs fns
          :ElseIf report
              r←MonRep cs fns top min cap
          :ElseIf clear
              ⎕EX'⎕SE.MonitorData'
          :Else ⋄ r←Help Cmd
          :EndIf
     
      :Case 'APLMON'
          :If 12>2⊃⎕VFI{(¯1+⍵⍳'.')↑⍵}2⊃'.'⎕WG'APLVersion'
              r←'APLMON requires version 12.0 or later'
          :Else
              on off←Args.Switch¨'on' 'off'
              :If 0≡file←Args.file
                  file←'\aplmon_',(,'ZI4,5ZI2'⎕FMT 1 6⍴⎕TS),'.csv'
                  file←({(-(⌽⍵)⍳'\')↓⍵}2 ⎕NQ'.' 'GetEnvironment' 'log_file'),file
              :EndIf
              file←file,(~'.'∊file)/'.csv'
     
              :If 0≠⍴Args.Arguments ⍝ Non-empty expression
                  expr←⊃,/' ',¨Args.Arguments
                  r←2 ⎕NQ'.' 'APLMON'file
                  cs⍎expr
                  r←'Written: ',2 ⎕NQ'.' 'APLMON' ''
              :Else
                  r←2 ⎕NQ'.' 'APLMON'(on/file)
                  r←'was: ',r,(0=⍴r)/'OFF'
              :EndIf
          :EndIf
     
      :Case 'CPUTime'
          ran←1⊃(repeat fn runfor)←splitRepeat⍕reparg←Args.repeat ⋄ (showai showresult)←∨\'ai' 'none'∊⊂Args.details
          results←0(2+2×showai)⍴all←∨/Args.details∘≡¨1 'all' ⋄ detail←{⍵:⍺}∘all ⋄ tell←{⍵:⍺}∘(⍱/Args.details∘≡¨'none' 'ai')
          :For exp :In Args.Arguments
              tell''
              tell⍕'* Benchmarking "',exp,'"',(reparg≢0)/', repeat=',reparg
     
            ⍝ Find how long it takes to do it once:
              {}⎕WA ⍝ max memory use
              r←~'←'∊exp ⋄ ai←⎕AI ⋄ {}{6::r∘←0 ⋄ cs⍎⍵}exp ⋄ ai←2↑1↓⎕AI-ai
              detail'Test run took ',⍕ai
              minrun←512 64 8 1[⎕IO++/1 10 100<1↑ai] ⋄ eex←{1↓⍵⍴⍨⍺×⍴⍵}∘('⋄',(r/'{}'),exp)
     
            ⍝ Have we already made the test?
              :If (runfor≤1↓ai)∧repeat≤1 ⋄ ai←ai∘.×1 1 0 ⍝ make ⎕MON values same
                  detail'** Expression was only run once' ⋄ mondata←⍬
     
              :ElseIf t←∧/cond←1<runfor,repeat ⍝ we have both # of time AND period
                  n←⍕⎕THIS ⋄ cond←1⌽cond,t
                  branch←'→2×⍳',↑,/cond/('(0<',n,'.cnt←',n,'.cnt-1)')('∧∨'[1+'⌈'∊fn])(n,'.n>⎕ai[3]')
                  (n minrun)←defineFn minrun branch
                  cnt←⌈repeat÷minrun
                  2 3 ⎕MONITOR'f' ⋄ n←runfor+3⊃ai←⎕AI ⋄ f ⋄ ai←2↑1↓⎕AI-ai
                  ran←minrun×1⍴r←2 3↑0 1↓mondata←⎕MONITOR'f' ⋄ ai←⍉ai⍪0 1↓r
     
              :ElseIf runfor>0  ⍝ run for a specified period
                ⍝ Define temp fn to run until time limit, set monitor on it, and run it
                  (n minrun)←defineFn minrun,⊂'→2×⎕ai[3]<',(⍕⎕THIS),'.n'
                  3 2 ⎕MONITOR'f' ⋄ n←runfor+3⊃ai←⎕AI ⋄ f ⋄ ai←2↑1↓⎕AI-ai
                  ran←minrun×1⍴r←2 3↑0 1↓mondata←⎕MONITOR'f' ⋄ ai←⍉ai⍪0 1↓r
     
              :Else  ⍝ repeat a specific # times
                ⍝ Create a fn that makes 2 sets of tests in blocks of minrun and remainder:
                  n←(⍕⎕THIS),'.n' ⋄ branch←'→2×⍳0<',n,'←',n,'-1'
                 ⍝ Build a fn that will be used to perform the test
                  ran←×/((n cnt)minrun)←defineFn repeat branch
                  4 2 3 ⎕MONITOR'f' ⋄ ai←⎕AI ⋄ f ⋄ ai←2↑1↓⎕AI-ai
                  r←+⌿2 2 2⍴4 2↑0 2↓mondata←⎕MONITOR'f' ⋄ ai←⍉ai⍪r
     
              :EndIf
            ⍝ Show ⎕MONITOR info plus ⎕AI's and branch line info if details wanted
              cap←'' '⎕AI' 'Exp' '[2]→line' ⋄ t←mondata≢⍬
              tell(1 0 1 0∨all∧t)/cap⍪'CPU (avg):' 'Elapsed:',(n←ai÷ran)
              detail'Effective repeat=',(⍕ran),t/', # expression on repeat line was ',⍕minrun
              detail'⎕MONITOR data:'mondata
              results⍪←,⍉n[;2,⍳showai]
          :EndFor
          r←showresult⌿results
      :EndSelect
    ∇

    ∇ (cnt minrun)←defineFn(minrun branch);done
      cnt←1 ⋄ done←0
      :Repeat
          :Trap 10 11 16
              done←÷'f'∊⎕FX'f' '⎕cs cs'(eex minrun),⊂branch
          :Else
              cnt×←2 ⋄ minrun←⌈minrun÷2
          :EndTrap
      :Until done
    ∇

    ∇ (times fn period)←splitRepeat string;n;b;K;M
     ⍝ Parse 'repeat' switch's value: # of times, period, min/max
     ⍝ The string may be a single number optionnaly followed by 'm' to denote miliseconds or 's' (secs)
     ⍝ or it may be 2 numbers, one of which followed by 'm' or 's', separated by either ⌈ or ⌊
      (times fn period)←1 '' 0 ⋄ K←1,1000*∨/b←string='s' ⋄ (b/string)←'m'        ⍝ 1 sec=1000 msecs
      (b n)←⎕VFI b\string/⍨b←~string∊'⌈⌊m' ⋄ b←(∧/b)∧(+/M←string='m')∊(⍴n)↓0 0 1 ⍝ accept only 1 m
      'Invalid repetition number'⎕SIGNAL 11/⍨b≤⍴string~⎕D,' m⌈⌊'
      :If 1=⍴n←⌈n
          (times period)←K×1 0⌈(∨/M)⌽n,0
      :Else
          fn←1↑'⌈⌊'∩string,'⌊' ⍝ ⌊ implied otherwise
          (times period)←K×n⌽⍨∨/((∨\M)/string)∊⎕D
      :EndIf
    ∇

    ∇ r←on MonSet(space fns);f
      :For f :In fns
          (on/0,⍳1↑⍴cs.⎕NR f)cs.⎕MONITOR f
      :EndFor
      r←'Monitoring switched ',((1+on)⊃'off' 'on'),' for ',(⍕1↑⍴fns),' functions'
    ∇

    ∇ r←MonRep(space fns top min cap);mondata;max;all;i;ix;z;m
     
      :If 0≠⍴fns
          mondata←cs.⎕MONITOR¨fns
          max←⌈⌿all←↑2↓¨4⍴¨mondata
      :AndIf 0∨.≠max
          i←{⍵/⍳⍴⍵}(all÷[2]max)∨.≥min
          top←(⍴i)⌊top+(0=top)×⍴i
          i←i[top↑⍒all[i;]]
     
          r←0 7⍴⊂''
          :For ix :In i
              z←ix⊃mondata
              z←(m←(z[;3 4]÷[2]max)∨.>min)⌿z
              z←z[;2 3 4],(100×z[;3]÷max[1]),z[;,1]
              r←r,[1](fns[ix]),z,↓m⌿cs.⎕CR ix⊃fns
          :EndFor
          r←AddTab cap r
      :Else
          r←'(no monitor data found)'
      :EndIf
    ∇

    ∇ {larg}MonitorDataHandler msg;sf;m;fn
     ⍝ Handle Callbacks on the Grid
     
      sf←⎕SE.MonitorData.TC.((Buttons.State=1)/SubForms)
      :Select larg
     
      :Case 'Close'
          ⎕SE.MonitorData.TC.((Buttons SubForms)←(0=Buttons.State)∘/¨Buttons SubForms) ⍝ Close active subform
      :Case 'Copy'
          :If 0≠⍴sf ⋄ (⊃sf).G.GridCopy ⍬ ⋄ :EndIf ⍝ Copy to clipboard
      :Case 'Edit'
          :If 0≠⍴sf ⋄
              m←0≠⊃∘⍴¨fn←(⊃sf).G.Values[;1]
              fn←(+\m)[7⊃msg]⊃m/fn
              :If 3=#.⎕NC fn
                  #.⎕ED fn
              :EndIf
          :EndIf
      :Case 'Finalize'
          ⎕EX'⎕SE.MonitorData'
      :EndSelect
     
    ∇

    ∇ TC←Make;form
      :If 9=⎕NC'⎕SE.MonitorData.TC'
      :AndIf 'Form'≡'⎕SE.MonitorData'⎕WG'Type'
          TC←⎕SE.MonitorData.TC
      :Else
          form←⎕SE.(MonitorData←⎕NEW'Form'(('Caption' ']monitor output')('Coord' 'Pixel')('Size'(400 800))))
          ⎕SE.Dyalog.⎕FX ⎕CR'MonitorDataHandler'
          form ⎕WS'Event'('Close' 'MonitorDataHandler' 'Finalize')
          form.(MB←⎕NEW'MenuBar'⍬)
          form.MB.(AC←⎕NEW'Menu'(⊂'Caption' '&Action'))
          form.MB.AC.(CL←⎕NEW'MenuItem'(('Caption' 'C&lose Active Tab')('Event'('Select' '⎕SE.Dyalog.MonitorDataHandler' 'Close'))))
          form.MB.AC.(CP←⎕NEW'MenuItem'(('Caption' '&Copy to Clipboard')('Event'('Select' '⎕SE.Dyalog.MonitorDataHandler' 'Copy'))))
     
          TC←⎕SE.MonitorData.(TC←⎕NEW'TabControl'⍬)
          TC.(Buttons←SubForms←⍬)
      :EndIf
    ∇

    ∇ Label←AddTab(Label Data);id;tb;form;props;newfn
     ⍝ Add a SubForm with a new Monitor Report
     
      TC←Make ⍝ ⎕SE.MonitorData.TC
      id←⍕{⊃(⍳1+⍴⍵)~⍵}2⊃⎕VFI⍕1↓¨'T'TC.⎕NL-9
      :If 0=⍴TC.Buttons ⋄ Label←Label,' [1]'
      :Else ⋄ Label←⊃({Label,' [',(⍕⍵),']'}¨⍳1+⍴TC.Buttons)~TC.Buttons.Caption
      :EndIf
     
      TC.Buttons,←tb←TC.⎕NEW'TabButton'(⊂'Caption'Label)
      TC.SubForms,←form←TC.⎕NEW'SubForm'(⊂'TabObj'tb)
     
      newfn←1,2≢/Data[;1] ⍝ Where do we have new fns?
      Data[(~newfn)/⍳⍴newfn;1]←⊂''
     
      props←('Values'Data)('Size'form.Size)('Attach'('Top' 'Left' 'Bottom' 'Right'))
      props,←('TitleWidth' 0)('ColTitles'('Function' 'Count' 'CPU' 'Elapsed' '%CPU' '' 'Code'))
      props,←('FormatString'('CI8' 'F6.1' 'P<[>Q<]>BI6' ''))('CellTypes'((⍴Data)⍴4 1 1 1 2 3 4))
      props,←('GridLineFCol'((255 255 255)(127 127 127)(0 0 0)))('RowLineTypes'(1+2×1⌽newfn))('ColLineTypes'((¯1↑⍴Data)⍴2 2 2 2 2 1 2))
      props,←('CellHeights'(17+10×newfn))('Event'('CellDblClick' '⎕SE.Dyalog.MonitorDataHandler' 'Edit'))
      form.G←form.⎕NEW'Grid'props
      {form.G.SetColSize ⍵ ¯3}¨⍳¯1↑⍴Data
     ⍝ Put tip on with timestamp
    ∇

:Endnamespace