﻿:Namespace XED           ⍝ WPF editor.      V1.0 - Original author: Pierre Gilbert.
                         ⍝ September 2010 - V1.01  modified by DanB 2011

⍝ This user command will allow XAML editing through a third party editor: Avalon
⍝ The Avalon DLL must be located in the same place as this script or the ws containing this script.
⍝ This goes for any reference to WSPath,  e.g. XAML {APL:WSPath} will refer to the path of the script
⍝ if is was fetched from file or the path of the ws if it resides in one.
⍝ Requirements: v12.1 Unicode and .Net 3.0

    ∇ r←List
      r←⎕NS¨1⍴⊂⍬
    ⍝ Name, group, short description and parsing rules
      r.Name←⊂'XAMLeditor'
      r.Group←⊂'XAML'
      r[1].Desc←'XAML editor'
      r.Parse←⊂'1S'
    ∇

    ∇ r←Run(Cmd Args)
      :Select Cmd
      :Case 'XAMLeditor'
          RunEd↑Args.Arguments
          r←0 0⍴''
      :EndSelect
    ∇

    ∇ r←level Help Cmd;⎕ML
      ⎕ML←1
      :Select Cmd
      :Case 'XAMLeditor'
          r←⊂'This editor will allow you to edit XAML text and even run it to test it.'
          r,←⊂'There are a few examples included with the editor.'
          :If level>0
              r,←'' 'The editor will allow you to test your code with the leftmost button (green arrow).'
              r,←⊂'You can also load and save code from/to variables in the workspace of from/to file.'
              r,←'You can previre and print your code and change the appearance of the window (e.g. font)' 'through the buttons on the toolbar.'
              r,←'' 'Original author: Pierre Gilbert'
          :EndIf
          r←↑r
      :EndSelect
    ∇

    ∇ r←sourcePath file;source;s
     ⍝ Find full path of source file
      source←⎕WSID                ⍝ use WS by default
      :If 2=⎕NC s←'##.SourceFile' ⍝ is this a UCMD?
      :OrIf 2=⎕NC s←'SALT_Data.SourceFile'
          source←⍎s
      :EndIf
      r←source{⍵,⍨⍺↓⍨-⊥⍨~'/\'∊⍨⍺}file
    ∇

    WSPath←¯1↓sourcePath''

    ⎕USING←'System,mscorlib.dll'
    _←Environment.GetFolderPath Environment.SpecialFolder.ProgramFiles
    _NetPath←_,'\Reference Assemblies\Microsoft\Framework\.NETFramework\v4.0\Profile\Client\'

    ⎕USING←0⍴⊂'' ⋄  CR←⎕ucs 13  ⋄  ⎕IO ⎕ML ⎕WX←1 3 3

    fmXED_obj←⎕null ⍝ these must be there to start with

    APPmode←0 ⍝ run in Application mode?

    If←/⍨ ⋄ Where←{⍵/⍳⍴⍵} ⋄ isChar←{0  2∨.=10|⎕dr,⍵}

    ∇ fmXED_Closing(sender event)
    ⍝ Handler for the Closing Event
      :If sender.IsVisible
          sender.Hide    ⍝ if the Window is visible just hide it
          event.Cancel←1 ⍝ and cancel closing the form
      :End
    ∇

    ∇ Close      ⍝ Function to Close XED manually
      :If 'System.Windows.Window'≡⍕fmXED_obj
          fmXED_obj.Hide
          fmXED_obj.Close
      :End
      tmfoldingUpdate_obj.Stop
      fmXED_obj←FoldingManager_obj←FoldingStrategy_obj←textEditor_obj←tmfoldingUpdate_obj←0
      fmXED_toi[;3]←0
    ∇

    ∇ CloseAppDomain
      1 ⎕NQ'⎕SE.mb.file.closead' 'Select'
    ∇

    ∇ {fmOwner}ExecXaml XamlString;root_object;sink;win;⎕USING
     ⍝ 'Compile' the XAML and Show the Result
     ⍝
     ⍝ XamlString = XAML String to Execute ('compile')
     ⍝ fmOwner    = Optional form that will Own the error message box in case of error.
     
      ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
     
      →0 If 0=⍴XamlString
     
     ⍝ Obtain and validate the Root Object of the XAML String
      →0 If ⎕NULL≡root_object←fmOwner GetObjectFromString XamlString
     
     ⍝ Show the root_object
      :If 'System.Windows.Window'≡⍕root_object.GetType
      :OrIf 'System.Windows.Navigation.NavigationWindow'≡⍕root_object.GetType
          :Trap DEBUG↓0
              :If 0≠⎕NC'fmOwner' ⍝ used ?
                 ⍝ Set the Owner of the Dialog form so it will be modal.
                  root_object.WindowStartupLocation←WindowStartupLocation.CenterOwner
                  root_object.Owner←fmOwner
     
                 ⍝ Set the Opacity of the Owner form
                  fmOwner.Opacity←0.6
              :End
     
             ⍝ Show and Wait while the Form is Visible
              root_object.Hide ⍝ otherwise <ShowDialog> will fail
              {}root_object.ShowDialog
              root_object.Close
     
              :If 0≠⎕NC'fmOwner' ⍝ used ?
                 ⍝ Set the Opacity of the Owner form
                  fmOwner.Opacity←1
              :End
     
          :Else
              ShowLastError
              root_object.Close
     
              :If 0≠⎕NC'fmOwner' ⍝ used ?
                 ⍝ Set the Opacity of the Owner form
                  fmOwner.Opacity←1
              :End
     
          :EndTrap
     
      :ElseIf 'System.Windows.Application'≡⍕root_object.GetType
     
      :Else
         ⍝ Root Object is Not a Window. Try to Make it the Content of a Window.
          :Trap DEBUG↓0
              win←⎕NEW Window
              win.Content←root_object
     
              :If 0≠⎕NC'fmOwner' ⍝ used ?
                 ⍝ Set the Owner of the Dialog form so it will be modal.
                  win.WindowStartupLocation←WindowStartupLocation.CenterOwner
                  win.Owner←fmOwner
     
                 ⍝ Set the Opacity of the Owner form
                  fmOwner.Opacity←0.6
              :End
     
                 ⍝ Show and Wait while the Form is Visible
              sink←win.ShowDialog
              win.Close
     
              :If 0≠⎕NC'fmOwner' ⍝ used ?
                 ⍝ Set the Opacity of the Owner form
                  fmOwner.Opacity←1
              :End
     
          :Else
              ShowLastError
              win.Close
              :If 0≠⎕NC'fmOwner' ⍝ used ?
                 ⍝ Set the Opacity of the Owner form
                  fmOwner.Opacity←1
              :End
          :EndTrap
      :End
    ∇

    ∇ r←RootObject FixImages Toi;bmpfrm;bmpimg;cur;fileStream;i;imgname;imgtype;MS;obj;objname;objtype;size;⎕USING
     ⍝ Automatically Set the Images, Icons and Cursors that are matching the names in the first column of the Table of Images..
     ⍝ The Toi file is made from the function: MakeToi
     ⍝
     ⍝ Toi = Table of Images Matrix,[;1] Name of Image in Root Object
     ⍝                              [;2] File Stream of Image
     ⍝                              [;3] Image Object Created
     ⍝ RootObject = Root Object where to fix the Image Object
     
      ⎕USING←('System.Windows,',_NetPath,'PresentationCore.dll')'System.IO,mscorlib.dll'
     
     ⍝ Populate the Table with the Image Objects
      :For i :In ⍳1⍴⍴Toi
          objname←⊃Toi[i;1]
          fileStream←⊃Toi[i;2]
     
          :If 'BMP'≡3↑fileStream
             ⍝ Create a BitmapImage Object from a File Stream
              MS←⎕NEW MemoryStream(⊂⎕UCS 3↓fileStream)
              bmpimg←⎕NEW Media.Imaging.BitmapImage  ⍝ System.Windows.Media.Imaging.BitmapImage
     
              bmpimg.BeginInit
              bmpimg.StreamSource←MS
              bmpimg.EndInit
     
              Toi[i;3]←bmpimg
     
          :ElseIf 'ICO'≡3↑fileStream
             ⍝ Create a BitmapFrame Object from a File Stream
              MS←⎕NEW MemoryStream(⊂⎕UCS 3↓fileStream)
              bmpfrm←Media.Imaging.BitmapFrame  ⍝ System.Windows.Media.Imaging.BitmapFrame
              bmpfrm←bmpfrm.Create MS
     
              Toi[i;3]←bmpfrm
     
          :ElseIf 'CUR'≡3↑fileStream
             ⍝ Create a Cursor Object from a File Stream
              MS←⎕NEW MemoryStream(⊂⎕UCS 3↓fileStream)
              cur←⎕NEW Input.Cursor MS
     
              Toi[i;3]←cur
     
          :ElseIf 'XML'≡3↑fileStream
              ⍝ Create an Image from a XAML File Stream
              ⍝ To Do.
          :Else
              'Should not be here: ',3↑fileStream
          :End
      :End
     
      r←Toi
     
      ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
     
     ⍝ Set the Images
      :For i :In ⍳↑⍴Toi
     
          imgname←⊃Toi[i;1]
          objname←(¯1+imgname⍳'.')↑imgname
     
          :Trap DEBUG↓0
             ⍝ Obtain and Validate Object Name
              obj←⍎'RootObject.',objname
          :Else
             ⍝ Failure. Probably for a value error.
              'WPF - Object ',objname,' Does not Exist and Property ',imgname,' is not Fixed'
              :Continue
          :EndTrap
     
          :Select objtype←obj.GetType
          :Case Window           ⍝ System.Windows.Window
              imgtype←'icon'
     
          :Case Controls.Image   ⍝ System.Windows.Controls.Image
              imgtype←'bitmap'
     
          :Case Input.Cursor     ⍝ System.Windows.Input.Cursor
              imgtype←'cursor'
          :Else
              'FixClass ERROR: Object: ',objtype,' Not Defined'
              :Continue
          :EndSelect
     
          :Select imgtype
          :Case 'bitmap'
             ⍝ Set the Image Source with the BitmapImage Object
              obj.Source←Toi[i;3]
              obj.Stretch←obj.Stretch.Uniform
     
              :If ∨/objname∊⎕D
              :AndIf 0≠size←⍬⍴2⊃⎕VFI(objname∊⎕D)/objname
                  obj.Height←size
                  obj.Width←size
              :EndIf
     
          :Case 'icon'
             ⍝ Set the Icon Property with the BitmapFrame Object
              obj.Icon←Toi[i;3]
     
              :If ∨/objname∊⎕D
              :AndIf 0≠size←⍬⍴2⊃⎕VFI(objname∊⎕D)/objname
                  obj.Height←size
                  obj.Width←size
              :EndIf
     
          :Case 'cursor'
             ⍝ Set the Cursor Property with the Cursor Object
              obj.Cursor←Toi[i;3]
     
          :Else
              'SHOULD NOT BE HERE'
          :EndSelect
      :EndFor
    ∇

    ∇ FoldingUpdate_Tick
     ⍝ Timer to Update the Folding of the Editor
      :If 'System.Windows.Window'≢⍕fmXED_obj
      :AndIf 'System.Windows.Threading.DispatcherTimer'≡⍕tmfoldingUpdate_obj
         ⍝ Stop the timer if the main Window is Closed
          ⍎DEBUG/'''Timer stopped'''
          tmfoldingUpdate_obj.Stop
          →0
      :EndIf
     
     ⍝ Main Window is Visible and there is text in the Editor.
      :Trap DEBUG↓0
          :If 0<textEditor_obj.Document.TextLength
             ⍝ Revise Folding Strategy
              FoldingStrategy_obj.UpdateFoldings FoldingManager_obj textEditor_obj.Document
          :End
      :EndTrap
    ∇

    ∇ r←GetLastError
      ⍝ Returns the Last Error
      :If 90=⎕EN
          r←'EXCEPTION: ',⎕EXCEPTION.Message
      :Else
          r←(1⊃⎕DM),': ',{(' '=1↑⍵)↓((1↓a,0)∨a←' '≠⍵)/⍵}(2⊃⎕DM)
      :EndIf
    ∇

    ∇ r←InsertPath r;rep
    ⍝ Replaces {WSpath} with path of workspace
      r←,ra
      rep←{⎕IO←0 ⋄ f by←⍵ ⋄ sh←⍴f←'{',f ⋄ (⍴⍺)=p←1⍳⍨b←f⍷⍺:⍺
          new←(≤\(∨\b)⍲⍺='}')/⍺ ⍝ remove following '}'
          ⍵ ∇⍨new[⍳p],by,(p+sh)↓new}
      r rep←'APL:WSPath'WSPath
    ∇

    ∇ RootObj←{fmOwner}GetObjectFromString XamlString;attribs;bitmap;caller;data;err;event;events;expr;icon;inter;lineno;MsgBox_obj;name;names;ref;sink;⎕USING;new;n;cbm;k;i;ind;j;m;commands;cmd;exe;can;cb
     ⍝ Creates a WPF Object From an APL Character Vector Representing a XAML File.
     ⍝ Parts of the code is taken from Dyalog.
     ⍝ Globals Used:  ShowLastError and fmMsgBox_xaml
     ⍝
     ⍝ XamlString = Vector of Characters representing a XAML string
     ⍝ fmOwner    = Form that will own the Message Box (Used in case of error).
     ⍝ RootObj    = .Net WPF Root Object
     ⍝
     ⍝ If a Control has an event attribute that begins with '∇' and a Name, it will be fixed.
     ⍝ If a Control has an event attribute that begins with '∇' and no Name, it will not be fixed and make no error.
     ⍝ If a Control has an event attribute that does not begins with '∇', it will make an error.
     
      RootObj←⎕NULL       ⍝ Default value in case of error
      →0 If 0=⍴XamlString
     
     ⍝ Check if Valid XML
      :Trap DEBUG↓0
         ⍝ The following line will bug if there is a syntax error in XamlString
          ⎕USING←'System.Xml,System.Xml.dll'
          (⎕NEW XmlDocument).LoadXml(⊂XamlString)
     
         ⍝ The string is valid XML. Clean-it before 'compiling'.
          data←⎕XML XamlString
          commands←0 4⍴⊂'' ⍝ Name Command Executed CanExecute
          cbm←data[;2]∊⊂'CommandBinding'
          :For i :In {⍵/⍳⍴⍵}({1↓(-+/∨\⍵='.')↑⍵}¨data[;2])∊⊂'CommandBindings'
              n←+/∨\⌽(data[i;1]-1)=data[⍳i;1]
              :If ∨/m←(⊃data[n;4])[;1]∊'Name' 'x:Name' ⋄ name←,⊃m⌿(⊃data[n;4])[;2]
              :Else ⋄ name←'CmdCtl',⍕i ⋄ data[n;4]←data[n;4],[1]¨⊂'Name'name
              :EndIf
              ind←i+{⍵/⍳⍴⍵}(∧\i↓data[i;1]<data[;1])∧i↓cbm
              :For j :In ind
                  k←(⊃data[j;4])[;1]⍳'Command' 'Executed' 'CanExecute'
                  commands,[1]←(⊂name),((⊃data[j;4])[;2],⊂'')[k]
              :EndFor
              data[i+⍳n;1]←¯1
          :EndFor
          data←(data[;1]≠¯1)⌿data
          data[;4]←{a←⍵[;1] ⋄ (~(((a⍳¨':')↓¨a)∊C)∨a∊C←⊂'Class')⌿⍵}¨data[;4] ⍝ remove CLASS elements
          attribs←data[;4]                                                  ⍝ Attributes of all the nodes
          names←{1⊃((⍵[;1]∊'x:Name' 'Name')/⍵[;2]),⊂''}¨attribs             ⍝ Find all 'x:Name' and 'Name' attributes
          :If 0≠⍴n←{⍵/⍳⍴⍵}names{⍬⍴(0=⍴⍺)∧∨/(↑¨⍵[;2])∊'∇'}¨attribs
              names[n]←(⊂'Ctl'),¨⍕¨n
              attribs[n]←names[n]{(,¨'Name'⍺),[1]⍵}¨attribs[n]
          :EndIf
          events←⊃⍪/names{(⊂⍺),('∇'=↑¨⊃¨⍵[;2])⌿⍵}¨attribs                   ⍝ Events are all attributes with values beginning with ∇
          {BuildCommand¨(⍵[;1]∊⊂'Command')⌿⍵[;2]}¨attribs                   ⍝ Build command bindings and functions
          data[;4]←{(~'∇'=↑¨⊃¨⍵[;2])⌿⍵}¨attribs                             ⍝ Remove the events: attribute definition beginning with '∇'
          (data events new)←HandleTags data events
          XamlString←InsertPath ⎕XML data
     
      :Else
         ⍝ There is an Error. Show it and exit.
          ShowLastError ⋄ →0
     
      :EndTrap
     
     ⍝ Obtain the Root Object of the XAML String
     ⍝ You can get an error here if, for example, there is an event without a '∇'
     ⍝ Since XamlString may have changed because the events are removed, and the line numbers are not the
     ⍝ same has showned in the editor, GetObjectFromString must now show the errors with a custom message box.
     
      :Trap DEBUG↓0
         ⍝ 'Compile' the Xaml String to Obtain the Root Object
          ⎕USING←'System.IO,mscorlib.dll' 'System.Xml,System.Xml.dll'
          ⎕USING,←⊂('System.Windows.Markup,',_NetPath,'PresentationFramework.dll')
          RootObj←XamlReader.Load XmlReader.Create(⎕NEW StringReader(⊂XamlString))
     
      :Else
         ⍝ There is an error. Show it in a WPF Custom Message Box.
     
          :If 0≠⎕NC'fmOwner' ⍝ used ?
             ⍝ Set the Wait Cursor on fmOwner form
              ⎕USING←'System.Windows,',_NetPath,'PresentationCore.dll'
              fmOwner.Cursor←Input.Cursors.Wait
          :End
     
         ⍝ Add line numbers to the XamlString and build the error message
          XamlString←⎕FMT XamlString
          lineno←('<[>,Q<]>LI',⍕2+⌊10⍟n)⎕FMT⍳n←1↑⍴XamlString
          err←'EXCEPTION: ',(⎕EXCEPTION.Message),(2⍴CR),,lineno,' ',XamlString,CR
     
         ⍝ Prepare a Custom WPF Message Box
          ⎕USING←'System.IO,mscorlib.dll' 'System.Xml,System.Xml.dll'
          ⎕USING,←⊂'System.Windows.Markup,',_NetPath,'PresentationFramework.dll'
          MsgBox_obj←XamlReader.Load XmlReader.Create(⎕NEW StringReader(⊂fmMsgBox_xaml))
     
         ⍝ Set the Error Text
          (MsgBox_obj.FindName⊂'tbError').Text←err
     
         ⍝ Set the Error Icon of the message box
          ⎕USING←'System.Drawing,System.Drawing.dll'('System.Windows,',_NetPath,'PresentationCore.dll')
          ⎕USING,←⊂('System.Windows,',_NetPath,'WindowsBase.dll')
          icon←SystemIcons.Error
          bitmap←Interop.Imaging.CreateBitmapSourceFromHIcon(icon.Handle)(Int32Rect.Empty)(Media.Imaging.BitmapSizeOptions.FromEmptyOptions)
          (MsgBox_obj.FindName⊂'imgIcon').Source←bitmap
     
          :If 0≠⎕NC'fmOwner' ⍝ used ?
             ⍝ Set the Message Box Title Bar Icon
              MsgBox_obj.Icon←fmOwner.Icon
     
             ⍝ Set the Owner and Opacity
              MsgBox_obj.Owner←fmOwner
              fmOwner.Opacity←0.6
          :End
     
         ⍝ Wait on the form
          :Trap DEBUG↓0
              {}MsgBox_obj.ShowDialog
          :EndTrap
     
          :If 0≠⎕NC'fmOwner' ⍝ used ?
             ⍝ Finished waiting. Set the normal opacity
              fmOwner.Opacity←1
     
             ⍝ Set Normal Cursor
              ⎕USING←'System.Windows,',_NetPath,'PresentationCore.dll'
              fmOwner.Cursor←Input.Cursors.Arrow
          :End
     
         ⍝ Exit
          →0
     
      :EndTrap
     
     ⍝ Create refs to all named child objects
      names←(,⊃0≠⍴¨names)/names            ⍝ Keep only the not empty names
      names←(0=RootObj.⎕NC↑names)/names    ⍝ Only names which do not already have a meaning
     
     ⍝ FindName will not Find the names on certain object (like template), so we need to iterate
     ⍝ and Trap the error when the name is not found. All the names references are saved in '_names'
      :For name :In names
          RootObj._names←''
          :Trap DEBUG↓0
              ⍎'RootObj.',name,'←RootObj.FindName ⊂name'
              RootObj._names,←⊂name
          :Else
              ⎕←'GetObjectFromString Warning - Name Not found: ',name
          :EndTrap
      :EndFor
     
     ⍝ Show Warning for Events that will not be fixed because they are nameless
      :If 0≠↑⍴inter←(~events[;1]∊names)⌿events
          ⎕←'<GetObjectFromString> Warning: Following Nameless Event(s) Are Not Fixed:'
          ⎕←inter
      :End
     
     ⍝ Wire up events
      events←(events[;1]∊names,new)⌿events ⍝ Named objects only
      caller←(1⊃⎕NSI),'.'
     
      :For (ref event expr) :In ↓events
          :Trap DEBUG↓0
              expr←('∇'=1↑expr)↓expr
              expr←((~'#'∊expr)/caller),expr
              ⍎'RootObj.',ref,'.on',event,'←''',expr,''''
          :Else
              ShowLastError ⋄ →0
          :EndTrap
      :EndFor
     
      :For i :In ⍳1↑⍴commands
          cmd←GetCommand commands[i;2]
          exe←⎕OR((~'#'∊expr)/caller),expr←⊃commands[i;3]
          can←⎕OR((~'#'∊expr)/caller),expr←⊃commands[i;4]
          ⎕USING←'System.Windows.Input,WPF/PresentationCore.dll'
          cb←⎕NEW CommandBinding(cmd exe can)                    ⍝ Prepare Save Command to have User actions
          (RootObj.FindName commands[i;1]).CommandBindings.Add cb    ⍝ Bind the command at the Window level
      :EndFor
     
    ∇

    ∇ r←GetCommand ra;a;⎕USING;all;C;nc;i
    ⍝ Get a reference to the command(s) in ra
      ⎕USING←'System.Windows.Input,WPF/PresentationCore.dll' 'System.Windows,WPF/PresentationFramework.dll' 'System.Windows.Documents'
      C←⍎⍕all←'ApplicationCommands' 'ComponentCommands' 'MediaCommands' 'NavigationCommands' 'EditingCommands'
      nc←↑,/⍴¨a←C.⎕NL-2
      :If 1≥≡ra ⋄ ra←⊂,ra ⋄ :EndIf
      i←(⊃,/a)⍳ra
      r←C[(nc/⍳⍴nc)[i]]⍎¨ra
    ∇

    ∇ BuildCommand cmd;a;prefix;ns;c;eh;name2;name1;ceh;cb;pr
      :If 1≥≡cmd ⋄ cmd←cmd cmd
      :EndIf
      cmd←cmd~¨' '
      :If 0≠⍴1⊃cmd
          :If 0=⎕NC'_Commands' ⋄ _Commands←FullCommand'Cut' 'Copy' 'Paste'
          :EndIf
          (1⊃cmd)←FullCommand 1⊃cmd
          (2⊃cmd)←((1+⍴2⊃cmd)|(2⊃cmd)⍳'.')↓2⊃cmd
          :If ~cmd[1]∊_Commands
              :If '#'≠1↑2⊃cmd ⋄ prefix←((prefix∊1↑prefix)⍳0)⊃prefix←⎕NSI,⊂''
              :Else ⋄ prefix←''
              :EndIf
              :If 0≠⍴ns←(+/∨\⌽(2⊃cmd)∊'.')↑2⊃cmd ⋄ prefix←prefix,'.',ns
              :EndIf
              :If 0=×/⍴prefix ⋄ pr←''⎕NS''
              :Else ⋄ pr←⍎prefix
              :EndIf
              c←(⍴ns)↓2⊃cmd
              :If 0=pr.⎕NC name1←'CanExecute',c
                  a←⊂name1,'(sender args)'
                  a,←⊂':Signature ',name1,' Object sender,CanExecuteRoutedEventArgs args'
                  pr.⎕FX a
              :EndIf
              :If 0=pr.⎕NC name2←'Executed',c
                  a←⊂name2,'(sender args)'
                  a,←⊂':Signature ',name2,' Object sender,ExecutedRoutedEventArgs args'
                  pr.⎕FX a
              :EndIf
              prefix←prefix,(0≠⍴prefix)/'.'
              ⎕USING←'System.Windows.Input,WPF/PresentationCore.dll' 'System.Windows,WPF/PresentationFramework.dll'
              eh←⎕NEW ExecutedRoutedEventHandler(⎕OR prefix,name2)
              ceh←⎕NEW CanExecuteRoutedEventHandler(⎕OR prefix,name1)
              cb←⎕NEW CommandBinding((⍎1⊃cmd)eh ceh)
              CommandManager.RegisterClassCommandBinding FrameworkElement cb
              _Commands,←cmd[1]
          :EndIf
      :EndIf
    ∇

    ∇ r←FullCommand cmd;a;⎕USING;b;i;prefix;flg
      :If flg←1≥≡cmd ⋄ cmd←,⊂cmd
      :EndIf
      ⎕USING←'System.Windows.Input,WPF/PresentationCore.dll' 'System.Windows.Documents,WPF/PresentationFramework.dll'
      a←(ApplicationCommands ComponentCommands MediaCommands NavigationCommands EditingCommands).⎕NL-2
      b←⊃,/⍴¨a ⋄ b,←1
      a←⊃,/a,⊂''
      i←a⍳cmd
      prefix←('ApplicationCommands' 'ComponentCommands' 'MediaCommands' 'NavigationCommands' 'EditingCommands' '')[(+\⊃,/b↑¨1)[i]]
      prefix←{⍵,(0≠×/⍴⍵)/'.'}¨prefix
      r←prefix,¨cmd
      :If flg ⋄ r←1⊃r
      :EndIf
    ∇

    ∇ (data events names)←HandleTags(data events);m;i;start;header;name;d;n;in
      m←(1↑⍴data)⍴1
      names←0⍴⊂''
      :For i :In Where(¯4↑¨data[;2])∊⊂'.Tag'
          name←¯4↓⊃data[i;2]
          header←⊖i↑[1]data
          in←header[;2]⍳⊂name
          name←⊃header[in;4]
          :If 0=⍴n←Where name[;1]∊'x:Name' 'Name'
              name←'Tag',⍕i
              names,←⊂name
              d←⊃data[i-in-1;4]
              d,[1]←'Name'name
              data[i-in-1;4]←⊂d
          :Else ⋄ name←name[n;2]
          :EndIf
          start←i+0,Where∧\i↓data[i;1]<data[;1]
          m[start]←0
          header←data[start;]
          :If 0≠⍴n←Where header[;1]=data[i;1]+3
              header←(2,⌊0.5×⍴n)⍴header[n;3]
              :If 0≠⍴n←Where((2↑¨header[1;])∊⊂'on')∧~(8↑¨header[1;])∊⊂'onhandle'
                  events,[1]←(⊂name),(2↓¨header[1;n]),[1.5]header[2;n]
              :EndIf
          :EndIf
      :EndFor
      data←m⌿data
    ∇

    ∇ r←GetSettings;filename;names;values;XML;⎕USING
     ⍝ Get the Settings from file XED.Settings.xml
     
      r←0  ⍝ Default Value in Case of Error
     
      :Trap DEBUG↓0
         ⍝ Add the Name of the File
          filename←sourcePath'XED.Settings.xml'
     
         ⍝ Read the File
          XML←⎕XML ⎕SE.Dyalog.Callbacks.GetUTF8File filename
          →0 If 0 'Settings'≢2⍴XML ⍝ sanity check
     
          names←'EditorFont' 'EditorFontSize' 'ReducePrecision' 'Precision' 'AttributeCounteTolerance'
          names,←'ReorderAttributes' 'RemoveCommonDefaultValues' 'ConvertTabsToSpaces' 'SpaceCount'
          values←{(b v)←⎕VFI ⍵ ⋄ 0∊b:⍵ ⋄ 1∊⍴b:⍬⍴v ⋄ v}¨(XML[;3],⊂'')[XML[;2]⍳names]
          r←1(names,[1.5]values)
      :EndTrap
    ∇

    ∇ MainButtonClickEvent(sender event);⎕USING;OFD;SFD
     ⍝ Main Handler for the Click Events of the Buttons
     ⍝ This handler was setup by a Bubble Event to Catch all the 'Click' event of any buttons on the form.
      :If DEBUG ⋄ ⎕←'In click callback'sender event ⋄ :EndIf
      :Select event.OriginalSource.Name
     
      :Case 'bnClearClipboard'
          ⎕USING←'System.Windows,',_NetPath,'PresentationCore.dll'
     
         ⍝ To Clear the Content of the Clipboard
          Clipboard.Clear
     
      :Case 'bnExec'
          fmXED_obj ExecXaml(textEditor_obj.Text)
     
      :Case 'bnOpenFile'
          ⍝ Show the Open File Dialog
     
          ⎕USING←'System.Windows.Forms,System.Windows.Forms.dll'
     
          OFD←⎕NEW OpenFileDialog
          OFD.Title←'Select a File'
          OFD.AddExtension←1
          OFD.DefaultExt←'.xaml'
          OFD.Filter←'XAML files|*.xaml|Backup files|*.backup|All files|*.*'
          OFD.Multiselect←0
          OFD.CheckFileExists←1
          OFD.CheckPathExists←1
          OFD.RestoreDirectory←1
     
         ⍝ Set the opacity of the owner form
          fmXED_obj.Opacity←0.6
     
          :If (DialogResult.OK)=OFD.ShowDialog ⍬
             ⍝ Open Button Was Clicked
     
             ⍝ Store the File Name in a global variable
              _currentFileName←OFD.FileName
     
             ⍝ Have the Text Editor Load the File
              textEditor_obj.Load(⊂_currentFileName)
     
             ⍝ Show the File Name in the Title Bar of the Window
              fmXED_obj.Title←'XAML Editor - [ ',(∆DBR _currentFileName),' ]'
     
          :Else
             ⍝ Cancel Button Was Clicked. Do Nothing.
          :End
     
         ⍝ Set back the opacity
          fmXED_obj.Opacity←1
     
      :Case 'bnOpenObject'
          bnOpenObject_Click
     
      :Case 'bnPrint'
          bnPrint_Click
     
      :Case 'bnPrintPreview'
          bnPrintPreview_Click
     
      :Case 'bnSaveFile'
         ⍝ Show the Save File Dialog
     
          ⎕USING←'System.Windows.Forms,System.Windows.Forms.dll'
     
          SFD←⎕NEW SaveFileDialog
          SFD.AddExtension←1
          SFD.DefaultExt←'.xaml'
          SFD.Filter←'XAML file|*.xaml|All files|*.*'
     
         ⍝ Set the opacity of the owner form
          fmXED_obj.Opacity←0.6
     
          :If DialogResult.OK=SFD.ShowDialog ⍬
             ⍝ Save Button Was Clicked.
     
             ⍝ Store the File Name
              _currentFileName←SFD.FileName
     
             ⍝ Have the Text Editor Save the File
              textEditor_obj.Save(⊂_currentFileName)
     
             ⍝ Show the File Name in the Title Bar of the Window
              fmXED_obj.Title←'XAML Editor - [ ',(∆DBR _currentFileName),' ]'
     
          :Else
             ⍝ Cancel Button Was Clicked. Do Nothing.
          :End
     
         ⍝ Set back the opacity
          fmXED_obj.Opacity←1
     
      :Case 'bnSaveObject'
          bnSaveObject_Click
     
      :Case 'bnScrubber'
          ⍝ bnScrubber_Click
          textEditor_obj.Text←fmXED_obj SCRUB.Run(textEditor_obj.Text _XED_settings)
     
      :Case 'bnSettings'
          bnSettings_Click
     
      :EndSelect
    ∇

    ∇ PrintPreview flowDoc;docViewer;flowDocPaginator;package;packageUri;packageUriString;printDLG;sink;wDV;xpsDoc;xpsSm;xpsStream;⎕USING
     ⍝ Program to Preview a FlowDocument in a DocumentViewer Object before Printing.
     ⍝
     ⍝ Based On: http://khason.net/blog/printing-more-then-one-page-creation-in-memory-xps-document-and-documentviewer-customization/
     
     ⍝ OBTAIN THE PROPERTIES OF THE DEFAULT PRINTER
      ⎕USING←'System.Windows.Controls,',_NetPath,'PresentationFramework.dll'
      printDLG←⎕NEW PrintDialog
     
     ⍝ SET THE FlowDocument ACCORDING TO THE PROPERTIES OF THE DEFAULT PRINTER
      ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
      flowDoc.PageHeight←printDLG.PrintableAreaHeight
      flowDoc.PageWidth←printDLG.PrintableAreaWidth
      flowDoc.PagePadding←(⎕NEW Thickness 50)
      flowDoc.ColumnWidth←printDLG.PrintableAreaWidth
      flowDoc.ColumnGap←0
     
     ⍝ GET A DocumentPaginator FROM THE FlowDocument
      ⎕USING←'System.Windows.Documents,',_NetPath,'PresentationCore.dll'
      flowDocPaginator←(IDocumentPaginatorSource ⎕CLASS flowDoc).DocumentPaginator
     
     ⍝ PREPARE A package FROM A MemoryStream
      ⎕USING←'System.IO,mscorlib.dll' 'System,System.dll'('System.IO.Packaging,',_NetPath,'WindowsBase.dll')
      ⎕USING,←⊂'System.IO.Packaging,',_NetPath,'PresentationCore.dll'
      xpsStream←⎕NEW MemoryStream
      package←Package.Open(xpsStream,FileMode.Create,FileAccess.ReadWrite)
      packageUriString←'memorystream://printstream'
      packageUri←⎕NEW Uri(⊂packageUriString)
      PackageStore.AddPackage(packageUri,package)
     
     ⍝ CREATE A XpsDocument WITH THE Package
      ⎕USING←'System.Windows.Xps.Packaging,',_NetPath,'ReachFramework.dll'
      ⎕USING,←⊂'System.Windows.Xps.Serialization,',_NetPath,'ReachFramework.dll'
      ⎕USING,←⊂'System.IO.Packaging,',_NetPath,'WindowsBase.dll'
      xpsDoc←⎕NEW XpsDocument(package,(CompressionOption.SuperFast),⊂packageUriString)
      xpsSm←⎕NEW XpsSerializationManager((⎕NEW XpsPackagingPolicy(xpsDoc)),0)
      xpsSm.SaveAsXaml flowDocPaginator
     
     ⍝ INITIALIZE A DocumentViewer FROM THE XpsDocument
      ⎕USING←'System.Windows.Controls,',_NetPath,'PresentationFramework.dll'
      docViewer←⎕NEW DocumentViewer
      docViewer.Document←xpsDoc.GetFixedDocumentSequence
     
     ⍝ ADD THE DocumentViewer TO A Window
      ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
      wDV←⎕NEW Window
      wDV.Title←'Print Preview'
      wDV.Content←docViewer
      wDV.WindowStartupLocation←wDV.WindowStartupLocation.CenterScreen
      wDV SetIcon(⊃fmXED_toi[fmXED_toi[;1]⍳⊂'PrintPreview24.png';2])
     
     ⍝ SHOW THE Window AS MODAL
      sink←wDV.ShowDialog
     
     ⍝ CLEAN-UP AFTER THE WINDOW IS CLOSED
      ⎕USING←('System.IO.Packaging,',_NetPath,'PresentationCore.dll')
      PackageStore.RemovePackage(packageUri)
      xpsDoc.Close
      package.Close
      xpsStream.Close
      xpsStream.Dispose
    ∇

    ∇ RunEd arg;HM;inter;progFiles;⎕USING;⎕ML;⎕IO;app;isVar;tmp;Text
     ⍝ Main Program to Start XED
     
      ⎕ML←3 ⋄ ⎕IO←1 ⋄ _currentObject←_currentFileName←''
     
      ⍎'CloseAppDomain'If APPmode
      DEBUG←{×⎕NC ⍵:⍎⍵ ⋄ s←⎕SE.SALTUtils ⋄ ×s.⎕NC ⍵:×s⍎⍵ ⋄ 0}'DEBUG'
     
     ⍝ *** SECTION TO SETUP THE INITIAL SETTINGS ***
     ⍝ Get the Settings from the XML file.
      :If 1=↑inter←GetSettings
          _XED_settings←2⊃inter  ⍝ success
      :Else
         ⍝ Failure. The XML file is Corrupted or Does Not Exists.
         ⍝ Take the Following Default Value.
          _XED_settings←1 2⍴'EditorFont' 'APL385 Unicode'
          _XED_settings⍪←'EditorFontSize' 14
          _XED_settings⍪←'ReducePrecision' 0
          _XED_settings⍪←'Precision' 10
          _XED_settings⍪←'AttributeCounteTolerance' 2
          _XED_settings⍪←'ReorderAttributes' 1
          _XED_settings⍪←'RemoveCommonDefaultValues' 1
          _XED_settings⍪←'ConvertTabsToSpaces' 1
          _XED_settings⍪←'SpaceCount' 2
     
          SaveSettings
      :End
     
     ⍝ *** SECTION TO SETUP ⎕USING ***
     
     ⍝ Section to Add the AvalonEdit Namespace
      _TextEditorPath←sourcePath'ICSharpCode.AvalonEdit.dll'
     
     ⍝ Verify the version of the DLL
      :If 0<⍴Text←'4.0.0.5471'IO.ChkFileVersion _TextEditorPath
          ⎕←'XED Error: ',Text
          →0
      :End
     
     ⍝ The DLL is Present and with the right version
     ⍝ You need to call (load) the DLL once so it will be found by the XAML parser
      ⎕USING←'ICSharpCode.AvalonEdit,',_TextEditorPath
      {}TextEditor  ⍝ prime the ICSharpCode.AvalonEdit DLL (otherwise it won't be seen)
     
     ⍝           *** SECTION TO PREPARE THE MAIN FORM USING XAML ***
     
      :If isVar←arg∧.=' ' ⋄ Text←_Default_xaml
     ⍝ Could  be a var
      :ElseIf 2⊃tmp isVar←{2=##.⎕NC'THIS':t,2∊(t←##.THIS).⎕NC ⍵ ⋄ #,2∊#.⎕NC ⍵}arg
          'Variable not character'⎕SIGNAL 11 If~isChar Text←tmp⍎arg
      :Else ⍝ must be a file
          tmp←⎕SE.Dyalog.Callbacks.GetUTF8File arg
          Text←tmp
      :EndIf
     
      :If 9.2≠⎕NC⊂'fmXED_obj' ⍝ is it already setup?
      :OrIf ~fmXED_obj{6::0 ⋄ ⍺⍎⍵}'IsActive'
        ⍝ Get the Root Object
          →0 If ⎕NULL≡fmXED_obj←GetObjectFromString fmXED_xaml  ⍝ any good?
      :Else ⍝ must already be up
        ⍝ The Window exists already. Reset the Content and show it.
          Text←⍴textEditor_obj.Text←Text
          fmXED_obj.Title←'Xaml Editor'
          fmXED_obj.WindowState←fmXED_obj.WindowState.Normal
          tmfoldingUpdate_obj.Start
          fmXED_obj.ShowDialog
          tmfoldingUpdate_obj.Stop
          →0
      :EndIf
     
     ⍝ Set The Bubble Event for the Buttons
      ⎕USING←'System.Windows,',_NetPath,'PresentationCore.dll'
      ⎕USING,←⊂'System.Windows.Controls.Primitives,',_NetPath,'PresentationFramework.dll'
      fmXED_obj.AddHandler(ButtonBase.ClickEvent)(⎕NEW RoutedEventHandler(⎕OR'MainButtonClickEvent'))
     
     ⍝ Automatically Fix The Bitmaps, Icons and Cursors.
      fmXED_toi←fmXED_obj FixImages fmXED_toi
     
     ⍝ *** SECTION TO SETUP THE AvalonEdit TEXT EDITOR WITH THE DEFAULT CONTENT ***
      textEditor_obj←fmXED_obj.textEditor
      textEditor_obj.Text←Text
      textEditor_obj.FontSize←_XED_settings[2;2]
      textEditor_obj.FontFamilyProperty←⎕NEW Media.FontFamily(_XED_settings[1;2])
     
     ⍝ Setup the Folding Manager.
      ⎕USING←('ICSharpCode.AvalonEdit,',_TextEditorPath)
      FoldingManager_obj←Folding.FoldingManager.Install textEditor_obj.TextArea
      FoldingStrategy_obj←⎕NEW Folding.XmlFoldingStrategy
      FoldingStrategy_obj.UpdateFoldings FoldingManager_obj textEditor_obj.Document
     
     ⍝ Setup the Syntax Highlighting for a XML Document.
      HM←⎕NEW Highlighting.HighlightingManager
      textEditor_obj.SyntaxHighlighting←HM.Instance.GetDefinition(⊂'XML')
     
     ⍝ Show the Line Numbers.
      textEditor_obj.ShowLineNumbers←1
     
     ⍝ Set-Up a Timer to Refresh the Folding Every 2 Seconds.
      ⎕USING←('System.Windows,',_NetPath,'WindowsBase.dll')'System,mscorlib.dll'
      tmfoldingUpdate_obj←⎕NEW Threading.DispatcherTimer
      tmfoldingUpdate_obj.Interval←TimeSpan.FromSeconds 2
      tmfoldingUpdate_obj.onTick←'FoldingUpdate_Tick' ⍝ will Stop the timer when the Window is Closed
      tmfoldingUpdate_obj.Start
     
     ⍝ Show The Root Object
      :If APPmode
          ⎕USING,←'System.Windows,WPF/PresentationFramework.dll' 'System.Windows.Controls'
          app←⎕NEW Application
          {}app.Run fmXED_obj
          CloseAppDomain
      :Else
          ⍝ Add the Closing Event Handler
          fmXED_obj.onClosing←'fmXED_Closing'
          fmXED_obj.Hide
          fmXED_obj.ShowDialog
          tmfoldingUpdate_obj.Stop
      :EndIf
    ∇

    ∇ {r}←SaveSettings;filename;XML
     ⍝ Save the XED Settings as a XML file
     
      :Trap r←0
         ⍝ Prepare the XML content
          XML←⎕XML(0 'Settings' '')⍪1,_XED_settings
          XML←'<?xml version="1.0" encoding="utf-8"?>',CR,XML
     
         ⍝ Add the Name on the File
          filename←sourcePath'XED.Settings.xml'
     
         ⍝ Save the File
          r←¯1≠XML IO.UFileWrite filename
      :Else
          ShowLastError
      :End
    ∇

    ∇ object SetIcon FileStream;BF;MS;obj;⎕USING
     ⍝ Sets the Icon Property of an Object from a FileStream
     ⍝
     ⍝ FileStream = File Stream Saved in the WS from function: FileStream
     ⍝ object     = Image's Object Name Present in the RootObject
     
      ⎕USING←('System.Windows.Media.Imaging,',_NetPath,'PresentationCore.dll')'System.IO,mscorlib.dll'
     
     ⍝ Create a BitmapFrame Object with the File Stream
      MS←⎕NEW MemoryStream(⊂⎕UCS 3↓FileStream)
      BF←BitmapFrame.Create MS
     
     ⍝ Set the Icon Property with the BitmapFrame Object
      object.Icon←BF
    ∇

    ∇ ShowLastError;BN;err;IMG;inter;MB;sink;title;⎕USING
     ⍝ Show the Last Error using a standard MessageBox
     
      ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
     
      title←'Error Executing XED'
      err←''
     
     ⍝ Build the Error Message
      :If 90≠⎕EN
         ⍝ APL Error
          err←(1⊃⎕DM),': ',{(' '=1↑⍵)↓((1↓a,0)∨a←' '≠⍵)/⍵}(2⊃⎕DM),CR,err
     
      :Else
         ⍝ .Net Error
          err←('EXCEPTION: ',⎕EXCEPTION.Message),CR,err
     
      :End
     
     
     ⍝ Show the Error Message
      :If 'System.Windows.Window'≡⍕fmXED_obj
          MB←MessageBox
          BN←MessageBoxButton
          IMG←MessageBoxImage
     
          sink←MB.Show fmXED_obj err title(BN.OK)(IMG.Error)
     
      :Else
          ⎕←'XED ERRROR: ',err
     
      :End
    ∇

    ∇ fs←{color_obj}ToFileStream file_name;BA;ext;file_name;handle;inter;MS;object;sink;size;⎕USING
     ⍝ Creates a FileStream Variable From an Image File on Disk.
     ⍝ Optionnaly a Color Object (System.Drawing.Color) can be Provided to Specify the Transparent Color on Bitmaps.
     
      ⎕USING←'System.Drawing,System.Drawing.dll' 'System.IO,mscorlib.dll'
     
      fs←'' ⍝ Default Value in Case of Error
     
     ⍝ VERIFY IF file_name IS VALID
      :If 0=↑⍴file_name  ⍝ Empty ?
      :AndIf 0=↑⍴file_name←IO.DialogOpenFile
          ⍝ Cancel Button Was Pressed
          ⎕←'ToFileStream ERROR: No File Name !' ⋄ →0
     
      :ElseIf 0=↑inter←IO.FileExists file_name
          ⍝ Abort. Invalid File Name.
          ⎕←'ToFileStream ERROR: invalid file' ⋄ →0
     
      :End
     
     ⍝ CREATE A FILE STREAM FROM A VALID FILE NAME
     ⍝ TESTS ON FILE NAME EXTENSION
      file_name←∆CharUpper ∆DBR file_name   ⍝ Remove Trealing and Leading Blanks
      :If 0=↑⍴ext←(file_name⍳'.')↓file_name
          ⎕←'ToFileStream ERROR: Use an Image File Name with an Extension, ie. C:\MyBitmap.bmp' ⋄ →0
     
      :ElseIf (⊂ext)∊'BMP' 'GIF' 'EXIG' 'JPG' 'PNG' 'TIFF'
          ext←'BMP'
     
      :ElseIf (⊂ext)∊'ICO' 'CUR' 'ANI'
         ⍝ Do Nothing, ext has the proper value
     
      :Else
          ⎕←'ToFileStream ERROR: Extension of File Not Supported !'
          ⎕←'Supported Extensions Are: .BMP .ICO .CUR .ANI .GIF .JPG .PNG .TIFF .EXIG'
          →0
     
      :End
     
       ⍝ CREATE THE FILE STREAM FROM FILE NAME
      :If 0=↑inter←IO.BFileRead file_name
         ⍝ Failure. Show Error.
          ⎕←'ToFileStream ERROR: ',2⊃inter ⋄ →0
     
      :Else
         ⍝ Success. Get the Byte Array
          BA←2⊃inter  ⍝ Byte Array in Characters
     
      :End
     
      fs←ext,BA  ⍝ Create the File Stream from the Byte Array
     
      :Trap DEBUG↓0
         ⍝ IF AN ORIGINAL .BMP FILE, CONVERT TO PNG FORMAT THAT IS BETTER AND
         ⍝ USES A LOSSLESS COMPRESSION ALGORITHM THAT SUPPORTS TRANSPARENCY
          :If 'BMP'≡∆CharUpper ¯3↑file_name
             ⍝ For a BMP file, reformat to PNG
              MS←⎕NEW MemoryStream(⊂⎕UCS 3↓fs)
              object←⎕NEW Bitmap MS                     ⍝ ReCreate the Object from the File Stream
              MS←⎕NEW MemoryStream                      ⍝ Get an Empty Memory Stream
              object.Save MS object.RawFormat.Png       ⍝ Save the Object into the Memory Stream with a PNG Format
              fs←'BMP',⎕UCS MS.ToArray                  ⍝ Create the BMP File Stream of Object in PNG Format
              MS.Close ⋄ MS.Dispose ⋄ object.Dispose    ⍝ Clean-up
          :End
     
         ⍝ IF BMP FILE STREAM WITH TRANSPARENT COLOR
          :If 'BMP'≡ext
          :AndIf 0≠⎕NC'color_obj' ⍝ color exist ?
             ⍝ Bitmap with an Optional Color Object
              MS←⎕NEW MemoryStream(⊂⎕UCS 3↓fs)
              object←⎕NEW Bitmap MS                     ⍝ ReCreate the Object from the FileStream
              object.MakeTransparent color_obj          ⍝ Set the Transparent Color of the Bitmap
              MS←⎕NEW MemoryStream                      ⍝ Create an Empty Memory Stream
              object.Save MS object.RawFormat.Png       ⍝ Save the Object into a Memory Stream with a PNG format
              fs←'BMP',⎕UCS MS.ToArray                  ⍝ Create the BMP File Stream
              MS.Close ⋄ MS.Dispose ⋄ object.Dispose    ⍝ Clean-up
     
          :ElseIf 0≠⎕NC'color_obj'
              ⎕←'ToFileStream WARNING: Transparent Color Used Only On Bitmap !'
     
         ⍝ Else do Nothing, Optional Color Does Not Exists.
          :End
      :Else
          ⎕←'ToFileStream ERROR: ',GetLastError ⋄ →0
     
      :EndTrap
    ∇

    ∇ obj←fillTree obj;ns_names;ns_var;test;ndNamespace_obj;i;ii;inter;NS;vars;node;add
    ⍝ Fill Tree view object with data
      obj.ndRoot.Tag←''                  ⍝ Needs to Be Initialized Empty
      ns_names←⎕THIS∪#⍎¨'#',#.⎕NL-9.1    ⍝ List the namespaces
      ns_var←ns_names.⎕NL 2              ⍝ Variables in namespaces
      test←(⊂0 0)≢¨⍴¨ns_var              ⍝ namespace empty ?
      ns_names←test/ns_names             ⍝ keep the non-empty namespace
      ns_var←test/ns_var                 ⍝ keep the var of non-empty namespace
     
     ⍝ ITERATE TO FILL THE Namespace NODES
      :For i :In ⍳⍴ns_names
     
          ndNamespace_obj←⎕NEW TreeViewItem
          ndNamespace_obj.Header←⍕NS←i⊃ns_names
          ndNamespace_obj.IsExpanded←1
          ndNamespace_obj.Tag←''
          add←0
          :For ii :In ⍳1⍴⍴vars←i⊃ns_var
              :If 1=≡inter←NS⍎vars[ii;]      ⍝ not nested?
              :AndIf 1=⍴⍴inter               ⍝ Vector ?
              :AndIf 0≠⍴inter                ⍝ Not empty ?
              :AndIf (10|⎕DR inter)∊2 0      ⍝ Characters ?
              :AndIf '<'=1↑∆DBR inter        ⍝ XML ?
                  node←⎕NEW TreeViewItem
                  node.Header←(⍕NS),'.',vars[ii;]~' '
                  node.Tag←node.Header
                  {}ndNamespace_obj.Items.Add node
                  add←1
              :End
          :EndFor
          :If add ⋄ {}obj.ndRoot.Items.Add ndNamespace_obj ⋄ :EndIf
      :EndFor
    ∇

    ∇ bnOpenObject_Click;bnOK_Click;fmSelectObject;i;ii;inter;ndNamespace_obj;node;ns_names;ns_var;sink;test;tvObjectView_SelectedItemChanged;⎕USING;tmp;NS;vars
     ⍝ Form to Select a Variable in this WS via a Tree View Object
     
      ⎕USING←'System.Windows.Controls,',_NetPath,'PresentationFramework.dll'
     
     ⍝ OBTAIN THE ROOT OBJECT OF THE OpenObject DIALOG FROM XAML STRING
      fmSelectObject←GetObjectFromString fmSelectObject_xaml
     
      →(⎕NULL≡fmSelectObject)/0          ⍝ Exit if not valid
      fmSelectObject←fillTree fmSelectObject
     
     ⍝ SET THE EVENTS
      fmSelectObject.bnOK.onClick←⎕OR ⎕FX'bnOK_Click' 'fmSelectObject.DialogResult←1 ⋄ fmSelectObject.Close'
      fmSelectObject.tvObjectView.onSelectedItemChanged←⎕OR ⎕FX'tvObjectView_SelectedItemChanged' 'fmSelectObject.tbObjectName.Text←fmSelectObject.tvObjectView.SelectedItem.Tag'
     
     ⍝ Set the Caption of the OK button
      fmSelectObject.bnOK.Content←'Open'
     
     ⍝ SET THE ICON OF THE FORM
      fmSelectObject SetIcon(⊃fmXED_toi[fmXED_toi[;1]⍳⊂'OpenObject24.png';2])
     
     ⍝ Set the Owner of the Dialog form so it will be modal.
      fmSelectObject.Owner←fmXED_obj
     
     ⍝ Set the Opacity of the Owner form
      fmXED_obj.Opacity←0.6
     
     ⍝ Show and Wait while the Form is Visible
      inter←fmSelectObject.ShowDialog
     
     ⍝ A button was clicked. Set the Opacity back of the main form.
      fmXED_obj.Opacity←1
     
     ⍝ Is Cancel was Clicked ?
      :If 0=inter
          ⍝ Cancel Button was Clicked. Do Nothing.
      :Else
          ⍝ OK Button was Clicked
          :If 0=↑⍴fmSelectObject.tbObjectName.Text
              ⍝ Nothing was Selected with OK Button Clicked. Do Nothing.
          :Else
              ⍝ A Variable Name was Selected with OK Button Clicked
              inter←fmSelectObject.tbObjectName.Text
              _currentObject←inter
              textEditor_obj.Text←⍎inter
              fmXED_obj.Title←'XAML Editor - [ ',(∆DBR inter),' ]'
          :End
      :End
    ∇

    ∇ bnPrintPreview_Click;documentLine;flowDoc;highlightedLine;highlighter;inlineBuilder;lineNumber;lineStartOffset;par;section;⎕USING
     ⍝ PrintPreview the Document
     
     ⍝ Exit if Document is Empty
      →(0=textEditor_obj.Document.TextLength)/0
     
     ⍝ Set the Wait Cursor
      ⎕USING←'System.Windows.Input,',_NetPath,'PresentationCore.dll'
      fmXED_obj.Cursor←Cursors.Wait
     
     ⍝ Section to Create the Paragraphs
      ⎕USING←'ICSharpCode.AvalonEdit,',_TextEditorPath
      ⎕USING,←⊂'System.Windows.Documents,',_NetPath,'PresentationFramework.dll'
      highlighter←textEditor_obj.TextArea.GetService Highlighting.IHighlighter
      par←⎕NEW Paragraph
     
      :For lineNumber :In ⍳textEditor_obj.Document.LineCount
     
          documentLine←textEditor_obj.Document.GetLineByNumber lineNumber
          inlineBuilder←⎕NEW Highlighting.HighlightedInlineBuilder(⊂documentLine.Text)
          highlightedLine←highlighter.HighlightLine documentLine
          lineStartOffset←documentLine.Offset
     
          :For section :In highlightedLine.Sections
              inlineBuilder.SetHighlighting(section.Offset-lineStartOffset)(section.Length)(section.Color)
          :EndFor
     
          par.Inlines.Add¨(inlineBuilder.CreateRuns)
          par.Inlines.Add(⎕NEW LineBreak)
     
      :EndFor
     
     ⍝ Initialize a New Flow Document
      flowDoc←⎕NEW FlowDocument par
      flowDoc.FontFamily←textEditor_obj.FontFamily
      flowDoc.FontSize←textEditor_obj.FontSize
     
     ⍝ Set the Opacity of the Owner form
      fmXED_obj.Opacity←0.6
     
     ⍝ Call the PrintPreview subroutine with the Flow Document
      PrintPreview flowDoc
     
     ⍝ Finished waiting. Set the normal opacity
      fmXED_obj.Opacity←1
     
     ⍝ Set Normal Cursor
      ⎕USING←'System.Windows.Input,',_NetPath,'PresentationCore.dll'
      fmXED_obj.Cursor←Cursors.Arrow
    ∇

    ∇ bnPrint_Click;documentLine;flowDoc;highlightedLine;highlighter;idpSource;inlineBuilder;lineNumber;lineStartOffset;par;printDLG;scrollViewer;section;sink;⎕USING
     ⍝ Print the Document using the Selected Printer
     ⍝ Based on: http://www.codeproject.com/KB/edit/AvalonEdit.aspx?msg=3337439#xx3337439xx
     ⍝ and:      http://blogs.vbcity.com/canoz/archive/2008/10/18/9184.aspx
     
     ⍝ Exit if Document is Empty
      →(0=textEditor_obj.Document.TextLength)/0
     
     ⍝ Obtain the Properties of the Selected Printer
      ⎕USING←'System.Windows.Controls,',_NetPath,'PresentationFramework.dll'
      printDLG←⎕NEW PrintDialog
     
     ⍝ Change the opacity of the main window
      fmXED_obj.Opacity←0.6
     
      :If 0=printDLG.ShowDialog
         ⍝ Cancel Button was Pressed. Show an Information Message Box
          ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
          sink←MessageBox.Show fmXED_obj'Printing Cancelled' 'XED'(MessageBoxButton.OK)(MessageBoxImage.Information)
     
         ⍝ Set normal opacity and Exit
          fmXED_obj.Opacity←1
          →0
     
      :Else
         ⍝ A printer was Selected. Set normal opacity and Continue
          fmXED_obj.Opacity←1
     
      :End
     
     ⍝ Section to Create the Paragraphs
      ⎕USING←'ICSharpCode.AvalonEdit,',_TextEditorPath
      ⎕USING,←⊂'System.Windows.Documents,',_NetPath,'PresentationFramework.dll'
      highlighter←textEditor_obj.TextArea.GetService(Highlighting.IHighlighter)
      par←⎕NEW Paragraph
     
      :For lineNumber :In ⍳textEditor_obj.Document.LineCount
     
          documentLine←textEditor_obj.Document.GetLineByNumber lineNumber
          inlineBuilder←⎕NEW Highlighting.HighlightedInlineBuilder(⊂documentLine.Text)
          highlightedLine←highlighter.HighlightLine documentLine
          lineStartOffset←documentLine.Offset
     
          :For section :In highlightedLine.Sections
              inlineBuilder.SetHighlighting(section.Offset-lineStartOffset)(section.Length)(section.Color)
          :EndFor
     
          par.Inlines.Add¨(inlineBuilder.CreateRuns)
          par.Inlines.Add(⎕NEW LineBreak)
     
      :EndFor
     
     ⍝ Initialize a New Flow Document
      flowDoc←⎕NEW FlowDocument par
      flowDoc.FontFamily←textEditor_obj.FontFamily
      flowDoc.FontSize←textEditor_obj.FontSize
     
     ⍝ Initialize a FlowDocumentScrollViewer
      ⎕USING←'System.Windows.Controls,',_NetPath,'PresentationFramework.dll'
      scrollViewer←⎕NEW FlowDocumentScrollViewer
      scrollViewer.Document←flowDoc
     
     ⍝ Set the Document According to the Properties of the Selected Printer
      ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
      scrollViewer.Document.PageHeight←printDLG.PrintableAreaHeight
      scrollViewer.Document.PageWidth←printDLG.PrintableAreaWidth
      scrollViewer.Document.PagePadding←(⎕NEW Thickness 25)
      scrollViewer.Document.ColumnWidth←printDLG.PrintableAreaWidth
      scrollViewer.Document.ColumnGap←0
     
      ⎕USING←'System.Windows.Documents,',_NetPath,'PresentationCore.dll'
      idpSource←IDocumentPaginatorSource ⎕CLASS(scrollViewer.Document)
      printDLG.PrintDocument(idpSource.DocumentPaginator(fmXED_obj.Title))
    ∇

    ∇ bnSaveObject_Click;bnOK_Click;fmSelectObject;i;ii;inter;ndNamespace_obj;node;ns_names;ns_var;sink;test;tvObjectView_SelectedItemChanged;⎕USING
     ⍝ Form to Save a Variable in this WS via a Tree View Object
     
      ⎕USING←'System.Windows.Controls,',_NetPath,'PresentationFramework.dll'
     
     ⍝ OBTAIN THE ROOT OBJECT OF THE DIALOG FROM XAML STRING
      fmSelectObject←fillTree GetObjectFromString fmSelectObject_xaml
     
     ⍝ SET THE EVENTS
      fmSelectObject.bnOK.onClick←⎕OR ⎕FX'bnOK_Click' 'fmSelectObject.DialogResult←1 ⋄ fmSelectObject.Close'
      fmSelectObject.tvObjectView.onSelectedItemChanged←⎕OR ⎕FX'tvObjectView_SelectedItemChanged' 'fmSelectObject.tbObjectName.Text←fmSelectObject.tvObjectView.SelectedItem.Tag'
     
     ⍝ Set the Caption of the OK button
      fmSelectObject.bnOK.Content←'Save'
     
     ⍝ Set the Content of the text block
      fmSelectObject.tbObjectName.IsEnabled←1
      fmSelectObject.tbObjectName.Text←_currentObject
     
     ⍝ SET THE ICON OF THE FORM
      fmSelectObject SetIcon(⊃fmXED_toi[fmXED_toi[;1]⍳⊂'SaveObject24.png';2])
     
     ⍝ Set the Owner of the Dialog form
      fmSelectObject.Owner←fmXED_obj
     
     ⍝ Set the Opacity of the Owner form
      fmXED_obj.Opacity←0.6
     
     ⍝ Show and Wait while the Form is Visible
      inter←fmSelectObject.ShowDialog
     
     ⍝ A button was clicked. Set the Opacity back of the main form.
      fmXED_obj.Opacity←1
     
     ⍝ Is Cancel was Clicked ?
      :If 0=inter
          ⍝ Cancel Button was Clicked. Do Nothing.
      :Else
          ⍝ OK Button was Clicked
          :If 0=↑⍴fmSelectObject.tbObjectName.Text
              ⍝ Nothing was Selected with OK Button Clicked. Do Nothing.
          :Else
              :Trap DEBUG↓0
                 ⍝ A Variable Name was Selected with OK Button Clicked
                  _currentObject←fmSelectObject.tbObjectName.Text
                  ⍎_currentObject,'←',(⍕⎕THIS),'.textEditor_obj.Text'
                  fmXED_obj.Title←'XAML Editor - [ ',(∆DBR _currentObject),' ]'
              :Else
                  ShowLastError
              :End
          :End
      :End
    ∇

    ∇ bnSettings_Click;bnSaveClick;CVT;fmSettings;fontsize;inter;settings;⎕USING
     ⍝ Function to Change the Settings of XED
     
      ⎕USING←'System.Windows,',_NetPath,'PresentationCore.dll'
     
     ⍝ Get the Root Object of the Settings Form
      fmSettings←GetObjectFromString fmSettings_xaml
     
     ⍝ Set FontName
      fmSettings.cbEditorFont.SelectedValue←⎕NEW Media.FontFamily(_XED_settings[1;2])
     
     ⍝ Set FontSize
      fontsize←8 9 10 11 12 13 14 15 16 17 18 20 22 24 28 32 36
      fmSettings.cbEditorFontSize.SelectedIndex←¯1+fontsize⍳_XED_settings[2;2]
     
     ⍝ Set Numeric Precision
      :If 1=_XED_settings[3;2]
       ⍝ Reduce Precision
          fmSettings.cbPrecision.SelectedIndex←¯1+(1 2 3 4 5 10)⍳_XED_settings[4;2]
      :Else
       ⍝ Do Not Reduce Precision
          fmSettings.cbPrecision.SelectedIndex←6 ⍝ Disabled
      :End
     
     ⍝ Set Attribute Tolerance
      fmSettings.cbAttributeTolerance.SelectedIndex←¯1+(0 1 2 3 4 5 99)⍳_XED_settings[5;2]
     
     ⍝ Set Reorder Attributes
      fmSettings.ckReorderAttributes.IsChecked←_XED_settings[6;2]
     
     ⍝ Set Remove Common Default
      fmSettings.ckRemoveCommonDefaultValues.IsChecked←_XED_settings[7;2]
     
     ⍝ Set Convert Tab to Space
      fmSettings.ckConvertTabsToSpaces.IsChecked←_XED_settings[8;2]
     
     ⍝ Set Space Count
      fmSettings.cbSpaceCount.SelectedIndex←¯1+(1 2 3 4 5)⍳_XED_settings[9;2]
     
     ⍝ Set the Buttons Events
      fmSettings.bnSave.onClick←⎕OR ⎕FX'bnSaveClick' 'fmSettings.DialogResult←1 ⋄ fmSettings.Close'
     
      ⍝ Set the Owner of the Dialog form
      fmSettings.Owner←fmXED_obj
     
      ⍝ Set the Opacity of the Owner form
      fmXED_obj.Opacity←0.6
     
      ⍝ Show and Wait while the Form is Visible
      inter←fmSettings.ShowDialog
     
      ⍝ A button was clicked. Set the Opacity back of the main form.
      fmXED_obj.Opacity←1
     
     ⍝ Is Cancel was Clicked ?
      :If 0=inter
          ⍝ Cancel Button was Clicked. Exit.
          →0
      :End
     
     ⍝ Save the Settings
      settings←_XED_settings
     
     ⍝ Get FontName
      settings[1;2]←⊂fmSettings.cbEditorFont.SelectedValue.Source
     
     ⍝ Get FontSize
      settings[2;2]←fontsize[1+fmSettings.cbEditorFontSize.SelectedIndex]
     
     ⍝ Get Numeric Precision
      :If 6=inter←fmSettings.cbPrecision.SelectedIndex
          ⍝ Do Not Reduce Precision
          settings[3;2]←0
      :Else
          ⍝ Reduce Precision
          settings[3;2]←1
          settings[4;2]←(1 2 3 4 5 10)[1+inter]
      :End
     
     ⍝ Get Attribute Tolerance
      settings[5;2]←0 1 2 3 4 5 99[1+fmSettings.cbAttributeTolerance.SelectedIndex]
     
     ⍝ Set Reorder Attributes
      settings[6;2]←fmSettings.ckReorderAttributes.IsChecked
     
     ⍝ Set Remove Common Default
      settings[7;2]←fmSettings.ckRemoveCommonDefaultValues.IsChecked
     
     ⍝ Set Convert Tab to Space
      settings[8;2]←fmSettings.ckConvertTabsToSpaces.IsChecked
     
     ⍝ Set Space Count
      settings[9;2]←(1 2 3 4 5)[1+fmSettings.cbSpaceCount.SelectedIndex]
     
     ⍝ Save the Settings into the Class
      _XED_settings←settings
     
     ⍝ Set Font Name and Size to the Editor
      textEditor_obj.FontFamily←⎕NEW Media.FontFamily(_XED_settings[1;2])
      textEditor_obj.FontSize←_XED_settings[2;2]
     
     ⍝ Save the Settings to a XML File
      :If 0=↑inter←SaveSettings
          ⎕←'bnSettings_Click Error: ',⍕2↑⎕DM
      :Else
          ⍝ Success Writing XML File. Do Nothing.
      :End
    ∇

    ∆CharUpper←{⎕USING←'' ⋄ (⎕NEW System.String(⊂,⍵)).ToUpperInvariant}

    ∆DBR←{(' '=1↑⍵)↓((1↓a,0)∨a←' '≠⍵)/⍵}

      ∆SS←{⎕ML←3                          ⍝ Approx alternative to xutils' ss.
          srce find repl←,¨⍵              ⍝ Source, find and replace vectors.
          cvex←{(+\find⍷⍵)⊂⍵}find,srce    ⍝ Partitioned at find points.
          (⍴repl)↓∊{repl,(⍴find)↓⍵}¨cvex}


    :Namespace IO
        ⎕USING←,⊂''
        ⎕IO ⎕ML ⎕WX←1 3 3

        ∇ r←BFileRead file_name;BA;FS;sink;size;⎕USING
          :Access Public Shared
           ⍝ Reads The Contents of a Binary File Into a Character Vector Without Encoding.
           ⍝ If file_name is empty (''), DialogOpenFile is called to retrieve the file_name.
           ⍝
           ⍝ file_name = Fully Qualified File Name, if Empty Uses DialogOpenFile
           ⍝         r = 1 (Bytes as 8 bits characters) (File Name) if Successfull
           ⍝         r = 0 (Error Description) (File Name) if Failure
           ⍝
           ⍝ Example: R ← DIO.BFileRead 'C:\MyBinaryFile'
           ⍝  Success: R ← 1 (Bytes as 8 bits characters) 'C:\MyBinaryFile'
           ⍝  Failure: R ← 0 'File Name Does Not Exists' 'C:\MyBinaryFile'
           ⍝
           ⍝ Example: R ← DIO.BFileRead '' ⍝ Select the File to Read Using the Open File Dialog
         
          ⎕USING←'System.IO,mscorlib.dll'
         
           ⍝ VERIFY IF 'file_name' IS EMPTY OR HAS CHARACTERS
          :If 0=↑⍴file_name←∊file_name  ⍝ Empty ?
               ⍝ Call the Open File Dialog
              :If 0=↑⍴file_name←DialogOpenFile
                   ⍝ Failure: No File Name. Exit the program.
                  r←0 'No File Name' '' ⋄ →0
              :Else
                   ⍝ Do Nothing, 'file_name' is not Empty.
              :End
          :ElseIf 80≠⎕DR file_name
               ⍝ Failure: 'file_name' Not Characters.
              r←0 'File Name Must Be Characters'file_name ⋄ →0
          :Else
               ⍝ Do Nothing 'file_name' is a Character Vector and Not Empty.
          :End
         
           ⍝ MAKE THE CALL
          :Trap 0
               ⍝ No Conversion is made with this Call.
              r←1(⎕UCS File.ReadAllBytes(⊂,file_name))file_name
          :Else
               ⍝ Failure: Unexpected Error While Reading the File
              r←0 GetLastError file_name
          :EndTrap
        ∇

        ∇ yes←FileExists file
          :Trap 22
              ⎕NUNTIE file ⎕NTIE yes←0
              yes←1
          :EndTrap
        ∇

        ∇ r←ver ChkFileVersion file_name;FVI;v;⎕USING
         ⍝ Check and return a message indicating the status of a File in Characters.
         
          ⎕USING←'System.Diagnostics,System.dll'
         
           ⍝ CHECK IF FILE EXISTS
          :If 0=↑FileExists file_name
              r←'File Does Not Exist: ',file_name ⋄ →0
          :End
         
           ⍝ MAKE THE CALL
          :Trap 0
               ⍝ FileVersionInfo Object is Returned
              FVI←FileVersionInfo.GetVersionInfo(⊂,file_name)
         
               ⍝ The File Version is a Vector of Characters
              r←(ver≢FVI.FileVersion)/'Invalid version: ',file_name
          :Case 6
               ⍝ VALUE ERROR. There is No File Version Available
              r←'No version available in :',file_name
          :Else
               ⍝ Unexpected Error While Obtaining The File Version
              r←0 GetLastError file_name
          :EndTrap
        ∇

        ∇ r←GetFiles arg;arg_save;path;search_option;search_pattern;⎕USING
          :Access Public Shared
           ⍝ Returns The Names of Files in a Directory (without their paths).
           ⍝
           ⍝ USAGE: r ← DIO.GetFiles Path [SearchPattern] [SearchOption]
           ⍝
           ⍝          Path = Directory Path to search
           ⍝ SearchPattern = The search string to match against the names of files found
           ⍝  SearchOption = 'TopDirectoryOnly' to search only the specified directory (default)
           ⍝                 'AllDirectories' to search all subdirectories
           ⍝
           ⍝             r = 1 (Nested vector with the fully qualified files names) arg, if Successfull
           ⍝             r = 0 (ERROR) arg, if Failure
           ⍝
           ⍝ EXAMPLES:
           ⍝ r←DIO.GetFiles 'C:\MyDir' '*' 'AllDirectories' ⍝ All Files in All Subdirectories of 'C:\MyDir'
           ⍝ r←DIO.GetFiles 'C:\MyDir' 'p*'                 ⍝ Files of 'C:\MyDir' beginning with 'p'
           ⍝ r←DIO.GetFiles 'C:\MyDir' '*.exe'              ⍝ Files of 'C:\MyDir' ending with '.exe'
         
          ⎕USING←'System.IO,mscorlib.dll'
         
           ⍝ PARSE THE ARGUMENT
          search_pattern←'*.*'
          :If 2=≡arg_save←arg
              :Select ↑⍴arg
              :CaseList 0 1
                  path←∊arg
              :Case 2
                  (path search_pattern)←arg
              :Case 3
                  (path search_pattern search_option)←arg
              :Else
                  r←0 'Argument Ill Formed'arg_save ⋄ →0
              :EndSelect
          :ElseIf 80=⎕DR arg
              path←arg
          :Else
              r←0 'Argument Ill Formed'arg_save ⋄ →0
          :End
         
           ⍝ TEST IF DIRECTORY EXISTS
          :If 0=↑DirExists path
              r←0 'Directory does not Exists'arg_save ⋄ →0
          :End
         
           ⍝ TEST IF SEARCH OPTION IS VALID
          :If 0≠⎕NC'search_option'  ⍝ Exists ?
               ⍝ search_option is defined
              :If 80=⎕DR search_option  ⍝ Characters ?
                  :If 1=(⊂search_option)∊(SearchOption.⎕NL ¯2)
                       ⍝ Valid Search Option
                      search_option←⍎'SearchOption.',search_option
                      (3⊃arg)←search_option
                  :Else
                       ⍝ Search Option is not Valid
                       ⍝ 'AllDirectories' or 'TopDirectoryOnly' are Valid
                      r←0 'Search Option is Not Valid'arg ⋄ →0
                  :End
              :Else
                   ⍝ Search Option is not Valid
                   ⍝ 'AllDirectories' or 'TopDirectoryOnly' are Valid
                  r←0 'Search Option is Not Valid'arg ⋄ →0
              :End
          :Else
               ⍝ Default Value
              search_option←SearchOption.TopDirectoryOnly
          :End
         
           ⍝ MAKE THE CALL
          :Trap 0
              r←Directory.GetFiles((⊂,path),(⊂,search_pattern),(search_option))
              :If 0≠↑⍴r
                  r←Path.GetFileName¨⊂¨,¨r  ⍝ Removes the Directory Path
                  r←∆DBR¨⊂[2]r[⍋r←⊃[2]r;]   ⍝ Sorts the Names
              :End
              r←1 r arg_save
          :Else
               ⍝ Failure: Unexpected Error
              r←0 GetLastError arg_save
          :End
        ∇

        ∇ r←GetLastError
           ⍝ Returns the Last Error
          :If 90=⎕EN
              r←'EXCEPTION: ',⎕EXCEPTION.Message
          :Else
              r←(1⊃⎕DM),': ',(2⊃⎕DM)
          :EndIf
        ∇

        ∇ {r}←text UFileWrite file;header;tie
     ⍝ Write UTF-8 text to  native file
          r←¯1
          header←¯17 ¯69 ¯65 ⍝ FEFF header in UTF-8 (unsigned)
          :Trap 22
              tie←file ⎕NCREATE 0
          :Else
              0 ⎕NRESIZE tie←file ⎕NTIE 0
          :EndTrap
          header ⎕NAPPEND tie
          r←text ⎕NAPPEND tie
          ⎕NUNTIE tie
        ∇


    :EndNamespace

    :Class SCRUB

        ⎕ML←3 ⋄ ⎕IO←1 ⋄ CR←⎕ucs 13

        ∆DBR←{(' '=1↑⍵)↓((1↓a,0)∨a←' '≠⍵)/⍵}

          ∆SS←{⎕ML←3                          ⍝ Approx alternative to xutils' ss.
              srce find repl←,¨⍵              ⍝ Source, find and replace vectors.
              cvex←{(+\find⍷⍵)⊂⍵}find,srce    ⍝ Partitioned at find points.
              (⍴repl)↓∊{repl,(⍴find)↓⍵}¨cvex}


        ∇ str←{owner}Run(XamlString _XED_settings);attribute;AttributeCounteTolerance;attributes;ConvertTabsToSpaces;elementName;ForceLineMin;i;isEmptyElement;k;ms;Precision;progFiles;ReducePrecision;RemoveCommonDefaultValues;ReorderAttributes;sink;SpaceCount;str4;sw;xml;xmlReader;_NetPath;⎕USING
          :Access Public Shared
    ⍝ Scrubs a XAML file
    ⍝ Based on the Work of Robby Ingebretsen (www.Kaxaml.com) and Pete OHanlon (http://peteohanlon.wordpress.com/moxaml-power-toys/)
         
          ⎕USING←'System,mscorlib.dll'
          progFiles←Environment.GetFolderPath Environment.SpecialFolder.ProgramFiles
          _NetPath←progFiles,'\Reference Assemblies\Microsoft\Framework\v3.0\'
         
          ⎕USING←'System,mscorlib.dll' 'System.IO,mscorlib.dll'
          ⎕USING,←⊂'System.Xml,System.Xml.dll'
          ⎕USING,←⊂('System.Windows,',_NetPath,'PresentationCore.dll')
         
          :Trap 0
         
         ⍝ Exit if Document is Empty
              :If 0=↑⍴XamlString
                  str←XamlString
              :End
         
              :If 0≠⎕NC'owner'
            ⍝ Set the Wait Cursor
                  owner.Cursor←Input.Cursors.Wait
              :End
         
         ⍝ Initialise the Settings to be used
              ReducePrecision←_XED_settings[3;2]
              Precision←_XED_settings[4;2]
              AttributeCounteTolerance←_XED_settings[5;2]
              ReorderAttributes←_XED_settings[6;2]
              RemoveCommonDefaultValues←_XED_settings[7;2]
              ConvertTabsToSpaces←_XED_settings[8;2]
              SpaceCount←_XED_settings[9;2]
         
              XamlString←∆SS XamlString'&'(⎕UCS 172 172)
              str←''
         
              :If 1=ReducePrecision
                  XamlString←Precision Trunk XamlString
              :EndIf
         
              xmlReader←⎕NEW XmlTextReader(⎕NEW StringReader(⊂XamlString))
              xmlReader.Normalization←0
              sink←xmlReader.Read
              xmlReader.Normalization←0
         
              :While 0=xmlReader.EOF
         
                  :Select xmlReader.NodeType
         
                  :Case XmlNodeType.Element
                ⍝ Element Type of Node (1)
         
                      :If ConvertTabsToSpaces
                          xml←(SpaceCount×xmlReader.Depth)⍴' '
                      :Else
                          xml←(xmlReader.Depth)⍴⎕UCS 9 ⍝ TAB
                      :End
         
                      elementName←xmlReader.Name
                      str,←CR,xml,'<',elementName
                      isEmptyElement←xmlReader.IsEmptyElement
         
                      :If 1=xmlReader.HasAttributes
                          attributes←''
                          :For k :In ¯1+⍳(xmlReader.AttributeCount)
                              xmlReader.MoveToAttribute k
                              attributes←attributes,⊂elementName(xmlReader.Name)(xmlReader.Value)
                          :EndFor
         
                          :If 1=RemoveCommonDefaultValues
                              attributes←(~IsCommonDefault¨attributes)/attributes
                          :EndIf
         
                          :If 1=ReorderAttributes
                              attributes←AttributeSort attributes
                          :EndIf
         
                          :If ConvertTabsToSpaces
                              xml←(SpaceCount×xmlReader.Depth)⍴' '
                          :Else
                              xml←(xmlReader.Depth)⍴⎕UCS 9 ⍝ TAB
                          :End
         
                          :For i :In ⍳⍴attributes
                              attribute←i⊃attributes
         
                              :If ((xmlReader.AttributeCount)>AttributeCounteTolerance)∧(~ForceNoLineBreaks 1⊃attribute)
                            ⍝ Break up attributes into different lines
                                  str,←CR,xml,(2⊃attribute),'="',(3⊃attribute),'"'
                              :Else
                            ⍝ attributes on one line
                                  str,←' ',(2⊃attribute),'="',(3⊃attribute),'"'
                              :EndIf
                          :EndFor
                      :EndIf
         
                      :If isEmptyElement
                          str,←'/'
                      :EndIf
         
                      str,←'>'
                      sink←xmlReader.Read
         
                  :Case XmlNodeType.Text
                ⍝ Text Type of Node (3)
         
                      str4←xmlReader.Value
                      str4←∆SS str4'&' '&amp;'
                      str4←∆SS str4'<' '&lt;'
                      str4←∆SS str4'>' '&gt;'
                      str4←∆SS str4'\''' '&quot;'
                      str4←(∆DBR str4)~⎕UCS 13 10
                      str,←str4
         
                      sink←xmlReader.Read
         
                  :Case XmlNodeType.ProcessingInstruction
                ⍝ ProcessingInstruction Type of Node (7)
         
                      xml←(4×xmlReader.Depth)⍴' '
                      str,←CR,xml,'<?Mapping',(xmlReader.Value),'?>'
                      sink←xmlReader.Read
         
                  :Case XmlNodeType.Comment
                ⍝ Comment Type of Node (8)
         
                      :If 1=ConvertTabsToSpaces
                          xml←(SpaceCount×xmlReader.Depth)⍴' '
                      :Else
                          xml←(xmlReader.Depth)⍴⎕UCS 9
                      :End
         
                      str,←CR,xml,'<!--',(xmlReader.Value),'-->'
                      sink←xmlReader.Read
         
                  :Case XmlNodeType.Whitespace
                ⍝ WhiteSpace Type of Node (13)
         
                      sink←xmlReader.Read
         
                  :Case XmlNodeType.EndElement
                ⍝ EndElement Type of Node (15)
         
                      :If ConvertTabsToSpaces
                          xml←(SpaceCount×xmlReader.Depth)⍴' '
                      :Else
                          xml←(xmlReader.Depth)⍴⎕UCS 9
                      :End
         
                      str,←CR,xml,'</',(xmlReader.Name),'>'
                      sink←xmlReader.Read
         
                  :Else
                ⍝ All the Other Type of Node
                      sink←xmlReader.Read
         
                  :EndSelect
              :EndWhile
         
              xmlReader.Close
         
              str←∆SS str(⎕UCS 172 172)'&'
              str←1↓str  ⍝ To Remove the First Character
         
          :Else
              str←XamlString ⋄ ShowLastError      ⍝ Failure. Show Last Error
          :EndTrap
         
          :If 0≠⎕NC'owner'
              owner.Cursor←Input.Cursors.Arrow    ⍝ Set Normal Cursor
          :End
        ∇

        ∇ r←AttributeSort v;attributename;i;index
    ⍝ Sorts the Attributes
    ⍝ Subroutine of bnScrubber_Click
         
          :Trap 0
              :If 0=⍴v
             ⍝ If Empty, Exit.
                  r←v ⋄ →0
              :End
         
         ⍝ Sort Alphabetically First
              v←v[⍋⊃[2](2⊃¨v)]
         
         ⍝ Initialize the Sort Index
              index←(⍴v)⍴0
         
         ⍝ Find the Weight of Each Attributes
              :For i :In ⍳⍴v
                  attributename←(2⊃i⊃v)
         
                  :If (5↑attributename)≡'xmlns'
                      index[i]←40 ⍝ Namespace
                  :Else
                      :Select attributename
                      :CaseList 'Key' 'x:Key'
                          index[i]←10 ⍝ Key
                      :CaseList 'Name' 'x:Name'
                          index[i]←20 ⍝ Name
                      :Case 'x:Class'
                          index[i]←30 ⍝ Class
                      :CaseList 'Canvas.Top' 'Canvas.Left' 'Canvas.Bottom' 'Canvas.Right' 'Grid.Row' 'Grid.RowSpan' 'Grid.Column' 'Grid.ColumnSpan'
                          index[i]←60 ⍝ Attached Layout
                      :CaseList 'Width' 'Height' 'MaxWidth' 'MinWidth' 'MinHeight' 'MaxHeight'
                          index[i]←50 ⍝ Core Layout
                      :CaseList 'Margin' 'VerticalAlignment' 'HorizontalAlignment' 'Panel.ZIndex'
                          index[i]←70 ⍝ Standard Layout
                      :CaseList 'mc:Ignorable' 'd:IsDataSource' 'd:LayoutOverrides' 'd:IsStaticText'
                          index[i]←2000 ⍝ BlendGoo
                      :Else
                          index[i]←1000 ⍝ Other
                      :EndSelect
                  :EndIf
              :EndFor
         
              r←v[⍋index]  ⍝ Sorts the Attributes List According to Their Rexpective Weight
         
          :Else
              r←v ⋄ ShowLastError
          :EndTrap
        ∇

        ∇ r←ForceNoLineBreaks elementname     ⍝ Subroutine of bnScrubber_Click
          r←(⊂elementname)∊'RadialGradientBrush' 'GradientStop' 'LinearGradientBrush' 'ScaleTransform' 'SkewTransform' 'RotateTransform' 'TranslateTransform' 'Trigger' 'Setter'
        ∇

        ∇ r←IsCommonDefault(elementname attributename attributevalue)
    ⍝ Determine if the Attribute is Common Default
    ⍝ Subroutine of bnScrubber_Click
    ⍝
    ⍝ r = 1 if Commnon Default, 0 Otherwise
         
          :Trap r←0
         
         ⍝ Do the tests
              :If (attributename≡'HorizontalAlignment')∧(attributevalue≡'Stretch')
              :OrIf (attributename≡'VerticalAlignment')∧(attributevalue≡'Stretch')
              :OrIf (attributename≡'Margin')∧(attributevalue≡'0')
              :OrIf (attributename≡'Margin')∧(attributevalue≡'0,0,0,0')
              :OrIf (attributename≡'Opacity')∧(attributevalue≡'1')
              :OrIf (attributename≡'FontWeight')∧(attributevalue≡'{x:Null}')
              :OrIf (attributename≡'Background')∧(attributevalue≡'{x:Null}')
              :OrIf (attributename≡'Stroke')∧(attributevalue≡'{x:Null}')
              :OrIf (attributename≡'Fill')∧(attributevalue≡'{x:Null}')
              :OrIf (attributename≡'Visibility')∧(attributevalue≡'Visible')
              :OrIf (attributename≡'Grid.RowSpan')∧(attributevalue≡'1')
              :OrIf (attributename≡'Grid.ColumnSpan')∧(attributevalue≡'1')
              :OrIf (attributename≡'BasedOn')∧(attributevalue≡'{x:Null}')
              :OrIf (elementname≢'ColumnDefinition')∧(elementname≢'RowDefinition')∧(attributename≡'Width')∧(attributevalue≡'Auto')
              :OrIf (elementname≢'ColumnDefinition')∧(elementname≢'RowDefinition')∧(attributename≡'Height')∧(attributevalue≡'Auto')
                  r←1
              :EndIf
          :Else
              ShowLastError
          :EndTrap
        ∇

        ∇ v←precision Trunk v;digits;i;segments
    ⍝ Trunks the digits past the precision required in a XAML file.
    ⍝ Subroutine of bnScrubber_Click
         
          →(~'.'∊v)/0  ⍝ Nothing to Trunk. Exit.
         
          segments←(v≠'.')⊂v
          digits←¯1+(segments∊¨⊂⎕D)⍳¨0
         
          :For i :In 1↓⍳⍴segments
              :If (i⊃digits)≤precision
            ⍝ Do Nothing. Nothing to Trunk
              :Else
                  (i⊃segments)←(precision↑i⊃segments),(i⊃digits)↓(i⊃segments)
              :End
          :EndFor
         
          v←¯1↓∊segments,¨⊂'.'
        ∇

        ∇ ShowLastError;BN;err;IMG;MB;progFiles;sink;title;_NetPath;⎕USING
    ⍝ Show the Last Error using a standard MessageBox
         
          ⎕USING←'System,mscorlib.dll'
          progFiles←Environment.GetFolderPath Environment.SpecialFolder.ProgramFiles
          _NetPath←progFiles,'\Reference Assemblies\Microsoft\Framework\v3.0\'
         
          ⎕USING←'System.Windows,',_NetPath,'PresentationFramework.dll'
         
          title←'Error Executing XED'
          err←''
         
     ⍝ Build the Error Message
          :If 90≠⎕EN
             ⍝ APL Error
              err←(1⊃⎕DM),': ',{(' '=1↑⍵)↓((1↓a,0)∨a←' '≠⍵)/⍵}(2⊃⎕DM),CR,err
          :Else
             ⍝ .Net Error
              err←('EXCEPTION: ',⎕EXCEPTION.Message),CR,err
          :End
         
     ⍝ Show the Error Message
          :If 0≠⎕NC'owner'
          :AndIf 'System.Windows.Window'≡⍕owner
              MB←MessageBox
              BN←MessageBoxButton
              IMG←MessageBoxImage
         
              {}MB.Show owner err title(BN.OK)(IMG.Error)
          :Else
              ⎕←'XED ERRROR: ',err
          :End
        ∇

    :EndClass

    L←⎕av[3+⎕io]              ⍝ These are XAML examples to play with
    _←⍬
    _,←'<Window xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"'
    _,←L,'xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml">',L,'</Window>'
    _Default_xaml←_

    _←⍬
    _,←'<Window xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"'
    _,←L,'  xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"'
    _,←L,'  MaxWidth="800"'
    _,←L,'  ResizeMode="NoResize"'
    _,←L,'  SizeToContent="WidthAndHeight"'
    _,←L,'  SnapsToDevicePixels="True"'
    _,←L,'  Title="Execution Error"'
    _,←L,'  WindowStartupLocation="CenterOwner">'
    _,←L,'  <StackPanel>'
    _,←L,'    <Grid'
    _,←L,'      Margin="0,-1,0,0"'
    _,←L,'      Background="White">'
    _,←L,'      <Grid.ColumnDefinitions>'
    _,←L,'        <ColumnDefinition Width="Auto"/>'
    _,←L,'        <ColumnDefinition/>'
    _,←L,'      </Grid.ColumnDefinitions>'
    _,←L,'      <Grid.RowDefinitions>'
    _,←L,'        <RowDefinition Height="Auto"/>'
    _,←L,'        <RowDefinition Height="Auto"/>'
    _,←L,'      </Grid.RowDefinitions>'
    _,←L,'      <Image'
    _,←L,'        x:Name="imgIcon"'
    _,←L,'        Grid.ColumnSpan="1"'
    _,←L,'        Grid.Row="0"'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="10,10,16,10"'
    _,←L,'        VerticalAlignment="Top"/>'
    _,←L,'      <ScrollViewer'
    _,←L,'        Name="Scroller"'
    _,←L,'        MaxHeight="600"'
    _,←L,'        MinWidth="200"'
    _,←L,'        Grid.Column="1"'
    _,←L,'        Grid.ColumnSpan="1"'
    _,←L,'        Grid.RowSpan="1"'
    _,←L,'        Margin="8,18,12,8"'
    _,←L,'        HorizontalScrollBarVisibility="Auto"'
    _,←L,'        VerticalScrollBarVisibility="Auto">'
    _,←L,'        <TextBlock'
    _,←L,'          x:Name="tbError"'
    _,←L,'          Width="{Binding ElementName=Scroller}"'
    _,←L,'          FontFamily="APL385 Unicode"'
    _,←L,'          FontSize="14"'
    _,←L,'          Text="TextBlock"/>'
    _,←L,'      </ScrollViewer>'
    _,←L,'      <Border'
    _,←L,'        Grid.ColumnSpan="2"'
    _,←L,'        Grid.Row="1"'
    _,←L,'        Margin="0,10,0,0"'
    _,←L,'        VerticalAlignment="Top"'
    _,←L,'        Background="#FFECECEC"'
    _,←L,'        BorderBrush="#FFC9C9C9"'
    _,←L,'        BorderThickness="0,1,0,0"'
    _,←L,'        Padding="8,0,0,0">'
    _,←L,'        <StackPanel>'
    _,←L,'          <Button'
    _,←L,'            x:Name="bnOK"'
    _,←L,'            Height="24"'
    _,←L,'            Width="86"'
    _,←L,'            HorizontalAlignment="Right"'
    _,←L,'            Margin="0,8,12,8"'
    _,←L,'            VerticalAlignment="Center"'
    _,←L,'            Content="OK"'
    _,←L,'            IsCancel="True"'
    _,←L,'            IsDefault="True"/>'
    _,←L,'        </StackPanel>'
    _,←L,'      </Border>'
    _,←L,'    </Grid>'
    _,←L,'  </StackPanel>'
    _,←L,'</Window>'
    fmMsgBox_xaml←_

    _←⍬
    _,←'<Window'
    _,←L,'  xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"'
    _,←L,'  xmlns:d="http://schemas.microsoft.com/expression/blend/2008"'
    _,←L,'  xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"'
    _,←L,'  xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"'
    _,←L,'  Height="518"'
    _,←L,'  Width="400"'
    _,←L,'  ResizeMode="NoResize"'
    _,←L,'  Title="Select a Variable"'
    _,←L,'  WindowStartupLocation="CenterOwner"'
    _,←L,'  mc:Ignorable="d">'
    _,←L,'  <Grid>'
    _,←L,'    <Grid.RowDefinitions>'
    _,←L,'      <RowDefinition Height="360"/>'
    _,←L,'      <RowDefinition Height="71"/>'
    _,←L,'      <RowDefinition Height="21"/>'
    _,←L,'    </Grid.RowDefinitions>'
    _,←L,'    <TreeView'
    _,←L,'      x:Name="tvObjectView"'
    _,←L,'      MaxHeight="356"'
    _,←L,'      Margin="20,20,20,8">'
    _,←L,'      <TreeViewItem'
    _,←L,'        x:Name="ndRoot"'
    _,←L,'        Header="Namespace"'
    _,←L,'        IsExpanded="True"/>'
    _,←L,'    </TreeView>'
    _,←L,'    <Button'
    _,←L,'      x:Name="bnCancel"'
    _,←L,'      Height="30"'
    _,←L,'      Width="135"'
    _,←L,'      Grid.Row="2"'
    _,←L,'      HorizontalAlignment="Right"'
    _,←L,'      Margin="0,8,20,-17"'
    _,←L,'      Content="Cancel"'
    _,←L,'      IsCancel="True"/>'
    _,←L,'    <Button'
    _,←L,'      x:Name="bnOK"'
    _,←L,'      Height="30"'
    _,←L,'      Width="135"'
    _,←L,'      Grid.Row="2"'
    _,←L,'      Margin="20,8,239,-17"'
    _,←L,'      Content="OK"'
    _,←L,'      IsDefault="True"/>'
    _,←L,'    <Label'
    _,←L,'      Grid.Row="1"'
    _,←L,'      Margin="20,5,290,40"'
    _,←L,'      Content="Object Name:"'
    _,←L,'      d:LayoutOverrides="Width, Height"/>'
    _,←L,'    <TextBox'
    _,←L,'      x:Name="tbObjectName"'
    _,←L,'      Height="24"'
    _,←L,'      Grid.Row="1"'
    _,←L,'      Margin="20,35,20,8"'
    _,←L,'      IsEnabled="False"/>'
    _,←L,'  </Grid>'
    _,←L,'</Window>'
    fmSelectObject_xaml←_

    _←⍬
    _,←'<Window'
    _,←L,'  xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"'
    _,←L,'  xmlns:d="http://schemas.microsoft.com/expression/blend/2008"'
    _,←L,'  xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"'
    _,←L,'  xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"'
    _,←L,'  Background="#FFEBECCA"'
    _,←L,'  ResizeMode="NoResize"'
    _,←L,'  SizeToContent="WidthAndHeight"'
    _,←L,'  Title="XED Settings"'
    _,←L,'  Topmost="True"'
    _,←L,'  WindowStartupLocation="CenterOwner"'
    _,←L,'  WindowStyle="ToolWindow"'
    _,←L,'  d:DesignHeight="322"'
    _,←L,'  d:DesignWidth="250"'
    _,←L,'  mc:Ignorable="d">'
    _,←L,'  <StackPanel'
    _,←L,'    Height="278"'
    _,←L,'    Width="227"'
    _,←L,'    HorizontalAlignment="Left"'
    _,←L,'    VerticalAlignment="Top">'
    _,←L,'    <!-- Font -->'
    _,←L,'    <StackPanel'
    _,←L,'      VerticalAlignment="Top"'
    _,←L,'      Orientation="Horizontal">'
    _,←L,'      <StackPanel.ToolTip>'
    _,←L,'        <TextBlock'
    _,←L,'          Text="Font Used by XED"'
    _,←L,'          TextWrapping="Wrap"/>'
    _,←L,'      </StackPanel.ToolTip>'
    _,←L,'      <Label'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="5,5,0,5"'
    _,←L,'        Content="Font:"/>'
    _,←L,'      <ComboBox'
    _,←L,'        Name="cbEditorFont"'
    _,←L,'        Height="20"'
    _,←L,'        Width="163"'
    _,←L,'        Margin="10,5,0,5"'
    _,←L,'        DataContext="{x:Static Fonts.SystemFontFamilies}"'
    _,←L,'        HorizontalContentAlignment="Stretch"'
    _,←L,'        ItemsSource="{Binding}"'
    _,←L,'        Padding="4,1">'
    _,←L,'      </ComboBox>'
    _,←L,'    </StackPanel>'
    _,←L,'    <!-- Font Size -->'
    _,←L,'    <StackPanel'
    _,←L,'      VerticalAlignment="Top"'
    _,←L,'      Orientation="Horizontal">'
    _,←L,'      <StackPanel.ToolTip>'
    _,←L,'        <TextBlock TextWrapping="Wrap">Size of the Font Used by XED'
    _,←L,'        </TextBlock>'
    _,←L,'      </StackPanel.ToolTip>'
    _,←L,'      <Label'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="5,5,0,5"'
    _,←L,'        Content="Font Size:"/>'
    _,←L,'      <ComboBox'
    _,←L,'        Name="cbEditorFontSize"'
    _,←L,'        Height="20"'
    _,←L,'        Width="60"'
    _,←L,'        Margin="20,5,5,5"'
    _,←L,'        Padding="4,1">'
    _,←L,'        <ComboBoxItem Content="8 pt"/>'
    _,←L,'        <ComboBoxItem Content="9 pt"/>'
    _,←L,'        <ComboBoxItem Content="10 pt"/>'
    _,←L,'        <ComboBoxItem Content="11 pt"/>'
    _,←L,'        <ComboBoxItem Content="12 pt"/>'
    _,←L,'        <ComboBoxItem Content="13 pt"/>'
    _,←L,'        <ComboBoxItem Content="14 pt"/>'
    _,←L,'        <ComboBoxItem Content="15 pt"/>'
    _,←L,'        <ComboBoxItem Content="16 pt"/>'
    _,←L,'        <ComboBoxItem Content="17 pt"/>'
    _,←L,'        <ComboBoxItem Content="18 pt"/>'
    _,←L,'        <ComboBoxItem Content="20 pt"/>'
    _,←L,'        <ComboBoxItem Content="22 pt"/>'
    _,←L,'        <ComboBoxItem Content="24 pt"/>'
    _,←L,'        <ComboBoxItem Content="28 pt"/>'
    _,←L,'        <ComboBoxItem Content="32 pt"/>'
    _,←L,'        <ComboBoxItem Content="36 pt"/>'
    _,←L,'        <ComboBoxItem/>'
    _,←L,'      </ComboBox>'
    _,←L,'    </StackPanel>'
    _,←L,'    <!-- Numeric Precision -->'
    _,←L,'    <StackPanel'
    _,←L,'      VerticalAlignment="Top"'
    _,←L,'      Orientation="Horizontal">'
    _,←L,'      <StackPanel.ToolTip>'
    _,←L,'        <TextBlock'
    _,←L,'          Width="240"'
    _,←L,'          TextWrapping="Wrap">Reduce the Precision of Large Numbers to a Maximum of this Number of Digits Beyond the Decimal Point.'
    _,←L,'        </TextBlock>'
    _,←L,'      </StackPanel.ToolTip>'
    _,←L,'      <Label'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="5,5,16,5"'
    _,←L,'        VerticalAlignment="Top"'
    _,←L,'        Content="Numeric  Precision:"/>'
    _,←L,'      <ComboBox'
    _,←L,'        Name="cbPrecision"'
    _,←L,'        Height="20"'
    _,←L,'        Width="76"'
    _,←L,'        Margin="5,10,5,5"'
    _,←L,'        VerticalAlignment="Top"'
    _,←L,'        Padding="4,1">'
    _,←L,'        <ComboBoxItem Content="1"/>'
    _,←L,'        <ComboBoxItem Content="2"/>'
    _,←L,'        <ComboBoxItem Content="3"/>'
    _,←L,'        <ComboBoxItem Content="4"/>'
    _,←L,'        <ComboBoxItem Content="5"/>'
    _,←L,'        <ComboBoxItem Content="10"/>'
    _,←L,'        <ComboBoxItem Content="Disabled"/>'
    _,←L,'      </ComboBox>'
    _,←L,'    </StackPanel>'
    _,←L,'    <!-- Attribute Tolerance -->'
    _,←L,'    <StackPanel'
    _,←L,'      VerticalAlignment="Top"'
    _,←L,'      Orientation="Horizontal">'
    _,←L,'      <StackPanel.ToolTip>'
    _,←L,'        <TextBlock'
    _,←L,'          Width="240"'
    _,←L,'          TextWrapping="Wrap">If More Than This Number of Attributes are Defined on an Element, Scrubber Will Move Each Attribute onto its Own Line.'
    _,←L,'        </TextBlock>'
    _,←L,'      </StackPanel.ToolTip>'
    _,←L,'      <Label'
    _,←L,'        Height="28"'
    _,←L,'        Width="120"'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="5,5,8,5"'
    _,←L,'        VerticalAlignment="Top"'
    _,←L,'        Content="Attribute  Tolerance:"/>'
    _,←L,'      <ComboBox'
    _,←L,'        Name="cbAttributeTolerance"'
    _,←L,'        Height="20"'
    _,←L,'        Width="76"'
    _,←L,'        Margin="5,10,5,5"'
    _,←L,'        VerticalAlignment="Top"'
    _,←L,'        Padding="4,1">'
    _,←L,'        <ComboBoxItem Content="0"/>'
    _,←L,'        <ComboBoxItem Content="1"/>'
    _,←L,'        <ComboBoxItem Content="2"/>'
    _,←L,'        <ComboBoxItem Content="3"/>'
    _,←L,'        <ComboBoxItem Content="4"/>'
    _,←L,'        <ComboBoxItem Content="5"/>'
    _,←L,'        <ComboBoxItem Content="One Line"/>'
    _,←L,'        <ComboBoxItem/>'
    _,←L,'      </ComboBox>'
    _,←L,'    </StackPanel>'
    _,←L,'    <StackPanel>'
    _,←L,'      <CheckBox'
    _,←L,'        Name="ckReorderAttributes"'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="10,5,5,5"'
    _,←L,'        Content="Reorder  Attributes  by  Importance">'
    _,←L,'        <CheckBox.ToolTip>'
    _,←L,'          <TextBlock'
    _,←L,'            Width="320"'
    _,←L,'            TextWrapping="Wrap">Scrubber Will Reorder Attributes on a Given Element According to a Predetermined Set of Rules About Which Attributes Are Most Important.'
    _,←L,'          </TextBlock>'
    _,←L,'        </CheckBox.ToolTip>'
    _,←L,'      </CheckBox>'
    _,←L,'      <CheckBox'
    _,←L,'        Name="ckRemoveCommonDefaultValues"'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="10,5,5,5"'
    _,←L,'        Content="Remove  Common  Default  Values">'
    _,←L,'        <CheckBox.ToolTip>'
    _,←L,'          <TextBlock'
    _,←L,'            Width="320"'
    _,←L,'            TextWrapping="Wrap">Scrubber Will Remove Common Properties Which Are Being Set to Their Default Values.'
    _,←L,'          </TextBlock>'
    _,←L,'        </CheckBox.ToolTip>'
    _,←L,'      </CheckBox>'
    _,←L,'      <!-- Indent Using Spaces -->'
    _,←L,'      <StackPanel Orientation="Horizontal">'
    _,←L,'        <StackPanel.ToolTip>'
    _,←L,'          <TextBlock'
    _,←L,'            Width="240"'
    _,←L,'            TextWrapping="Wrap">If This is Checked, Scrubber Will Indent Using Spaces Instead of Tabs.'
    _,←L,'          </TextBlock>'
    _,←L,'        </StackPanel.ToolTip>'
    _,←L,'        <CheckBox'
    _,←L,'          Name="ckConvertTabsToSpaces"'
    _,←L,'          HorizontalAlignment="Left"'
    _,←L,'          Margin="10,5,2,5"'
    _,←L,'          Content="Indent Using Spaces"/>'
    _,←L,'        <ComboBox'
    _,←L,'          Name="cbSpaceCount"'
    _,←L,'          Height="20"'
    _,←L,'          Width="72"'
    _,←L,'          HorizontalAlignment="Right"'
    _,←L,'          Margin="5"'
    _,←L,'          Padding="4,1">'
    _,←L,'          <ComboBoxItem Content="1 Space"/>'
    _,←L,'          <ComboBoxItem Content="2 Spaces"/>'
    _,←L,'          <ComboBoxItem Content="3 Spaces"/>'
    _,←L,'          <ComboBoxItem Content="4 Spaces"/>'
    _,←L,'          <ComboBoxItem Content="5 Spaces"/>'
    _,←L,'        </ComboBox>'
    _,←L,'      </StackPanel>'
    _,←L,'    </StackPanel>'
    _,←L,'    <StackPanel'
    _,←L,'      Name="StackPanelLine"'
    _,←L,'      Margin="5,5,5,0">'
    _,←L,'      <Line'
    _,←L,'        Stroke="Black"'
    _,←L,'        StrokeThickness="2"'
    _,←L,'        X1="0"'
    _,←L,'        X2="{Binding ElementName=StackPanelLine, Path=ActualWidth}"'
    _,←L,'        Y1="0"'
    _,←L,'        Y2="0"/>'
    _,←L,'    </StackPanel>'
    _,←L,'    <!-- Buttons -->'
    _,←L,'    <StackPanel'
    _,←L,'      Height="47"'
    _,←L,'      Orientation="Horizontal">'
    _,←L,'      <Button'
    _,←L,'        Name="bnSave"'
    _,←L,'        Height="24"'
    _,←L,'        Width="76"'
    _,←L,'        HorizontalAlignment="Left"'
    _,←L,'        Margin="10,10,20,10"'
    _,←L,'        VerticalAlignment="Center"'
    _,←L,'        Content="Save"'
    _,←L,'        IsDefault="True"/>'
    _,←L,'      <Button'
    _,←L,'        Name="bnCancel"'
    _,←L,'        Height="24"'
    _,←L,'        Width="76"'
    _,←L,'        HorizontalAlignment="Right"'
    _,←L,'        Margin="30,0,20,0"'
    _,←L,'        VerticalAlignment="Center"'
    _,←L,'        Content="Cancel"'
    _,←L,'        IsCancel="True"/>'
    _,←L,'    </StackPanel>'
    _,←L,'  </StackPanel>'
    _,←L,'</Window>'
    fmSettings_xaml←_

    _←⍬
    _,←'<Window'
    _,←L,'  x:Name="fmXED"'
    _,←L,'  xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"'
    _,←L,'  xmlns:avalonEdit="clr-namespace:ICSharpCode.AvalonEdit;assembly=ICSharpCode.AvalonEdit"'
    _,←L,'  xmlns:d="http://schemas.microsoft.com/expression/blend/2008"'
    _,←L,'  xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"'
    _,←L,'  xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"'
    _,←L,'  Height="500"'
    _,←L,'  MinHeight="400"'
    _,←L,'  MinWidth="820"'
    _,←L,'  Width="820"'
    _,←L,'  Title="Xaml Editor"'
    _,←L,'  WindowStartupLocation="CenterScreen"'
    _,←L,'  mc:Ignorable="d">'
    _,←L,'  <Window.Resources>'
    _,←L,'    <Style TargetType="{x:Type Image}">'
    _,←L,'      <Setter Property="RenderOptions.BitmapScalingMode" Value="NearestNeIghbor"/>'
    _,←L,'    </Style>'
    _,←L,'  </Window.Resources>'
    _,←L,'  <DockPanel>'
    _,←L,'    <ToolBarPanel'
    _,←L,'      VerticalAlignment="Top"'
    _,←L,'      DockPanel.Dock="Top">'
    _,←L,'      <ToolBar>'
    _,←L,'        <ToolBar.Resources>'
    _,←L,'          <Style TargetType="{x:Type Image}">'
    _,←L,'            <Style.Triggers>'
    _,←L,'              <DataTrigger'
    _,←L,'                Binding="{Binding RelativeSource={RelativeSource AncestorType={x:Type ButtonBase}, AncestorLevel=1}, Path=IsEnabled}"'
    _,←L,'                Value="False">'
    _,←L,'                <Setter Property="Opacity" Value="0.30"/>'
    _,←L,'              </DataTrigger>'
    _,←L,'            </Style.Triggers>'
    _,←L,'          </Style>'
    _,←L,'        </ToolBar.Resources>'
    _,←L,'        <ToolBar.Background>'
    _,←L,'          <LinearGradientBrush EndPoint="0,1" StartPoint="0,0">'
    _,←L,'            <GradientStop Color="White" Offset="0"/>'
    _,←L,'            <GradientStop Color="#FFFFFBFF" Offset="0.5"/>'
    _,←L,'            <GradientStop Color="#FFE3E69C" Offset="1"/>'
    _,←L,'          </LinearGradientBrush>'
    _,←L,'        </ToolBar.Background>'
    _,←L,'        <Button x:Name="bnExec">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Run24"'
    _,←L,'            ToolTip="Run Xaml"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button x:Name="bnScrubber">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Scrubber24"'
    _,←L,'            ToolTip="Scrub Xaml"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button x:Name="bnOpenObject">'
    _,←L,'          <Image'
    _,←L,'            x:Name="OpenObject24"'
    _,←L,'            ToolTip="Load from APL Variable"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button x:Name="bnSaveObject">'
    _,←L,'          <Image'
    _,←L,'            x:Name="SaveObject24"'
    _,←L,'            ToolTip="Save to APL Variable"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button x:Name="bnOpenFile">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Open24"'
    _,←L,'            ToolTip="Load from File"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button x:Name="bnSaveFile">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Save24"'
    _,←L,'            ToolTip="Save to File"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button'
    _,←L,'          Command="Cut"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Cut24"'
    _,←L,'            ToolTip="Cut"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button'
    _,←L,'          Command="Copy"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Copy24"'
    _,←L,'            ToolTip="Copy"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button'
    _,←L,'          Command="Paste"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Paste24"'
    _,←L,'            ToolTip="Paste"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button'
    _,←L,'          Command="SelectAll"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="SelectAll24"'
    _,←L,'            ToolTip="Select All"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button'
    _,←L,'          Command="Delete"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Delete24"'
    _,←L,'            ToolTip="Delete"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button x:Name="bnClearClipboard">'
    _,←L,'          <Image'
    _,←L,'            x:Name="ClearClipboard24"'
    _,←L,'            ToolTip="Clear Clipboard"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button'
    _,←L,'          Command="Undo"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Undo24"'
    _,←L,'            ToolTip="Undo"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button'
    _,←L,'          Command="Redo"'
    _,←L,'          CommandTarget="{Binding ElementName=tbTextBox}">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Redo24"'
    _,←L,'            ToolTip="Redo"'
    _,←L,'            ToolTipService.ShowOnDisabled="True"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button x:Name="bnPrint">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Print24"'
    _,←L,'            ToolTip="Print"/>'
    _,←L,'        </Button>'
    _,←L,'        <Button x:Name="bnPrintPreview">'
    _,←L,'          <Image'
    _,←L,'            x:Name="PrintPreview24"'
    _,←L,'            ToolTip="Print Preview"/>'
    _,←L,'        </Button>'
    _,←L,'        <Separator Width="3"/>'
    _,←L,'        <Button x:Name="bnSettings">'
    _,←L,'          <Image'
    _,←L,'            x:Name="Options24"'
    _,←L,'            ToolTip="Change Settings"/>'
    _,←L,'        </Button>'
    _,←L,'      </ToolBar>'
    _,←L,'    </ToolBarPanel>'
    _,←L,'    <avalonEdit:TextEditor'
    _,←L,'      x:Name="textEditor"'
    _,←L,'      Background="#FFEBECCA"'
    _,←L,'      DockPanel.Dock="Bottom"'
    _,←L,'      FontFamily="APL385 Unicode"'
    _,←L,'      FontSize="14pt"/>'
    _,←L,'  </DockPanel>'
    _,←L,'</Window>'
    fmXED_xaml←_

    _←'ClearClipboard24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 32 0 0 0 32 8 6 0 0 0 115 122 122 244 0 0 5 108 73 68 65 84 120 94 173 150 109 108 20 101 30 192 127 51 179 187 45 203 110 187 125 179 45 28 150 2 149 150 154 174 71 226 97 164 109 106 148 182 22 48 136 68 131 177 13 132 180 209 227 98 184 139 212 15 222 65 64 32 66 137 1 67 130 128 36 10 21 99 2 2 10 74 17 69 155 70 131 98 239 138 81 73 93 14 10 74 145 210 151 133 109 187 47 51 243 184 204 135 205 110 183 117 167 224 47 121 50 31 255 191 231 255 246 140 132 9 214 53 189 221 33 73 82 9 9 240 221 242 30 217 178 238 197 127 0 55 128 0 38 176 16 197 246 101 164 217 44 108 23 48 91 211 241 1 232 178 3 111 56 120 102 70 42 89 25 46 198 162 167 119 0 96 209 166 218 137 69 118 121 112 80 81 8 1 168 42 191 28 57 203 43 95 254 200 101 192 15 136 49 5 146 44 28 205 202 123 160 108 246 130 181 36 217 93 12 120 7 249 226 76 39 4 160 168 32 143 191 76 202 34 197 105 103 36 189 125 55 185 120 185 155 27 189 94 146 75 26 103 86 206 190 74 234 68 149 160 223 79 251 233 207 231 212 104 221 197 97 129 165 64 39 48 52 166 128 44 83 54 239 133 195 56 146 126 198 63 240 53 205 45 185 248 131 25 76 207 159 76 118 86 26 147 115 51 73 78 178 49 18 251 132 100 0 6 110 14 114 225 34 28 251 225 94 234 231 255 70 174 67 37 55 63 159 125 27 55 20 2 15 1 253 64 23 81 72 68 177 187 1 81 191 227 18 162 191 153 111 126 114 114 242 172 139 146 89 211 12 1 179 156 255 249 50 63 117 118 81 91 121 157 188 236 0 0 111 253 231 223 52 236 102 11 208 12 124 31 159 129 120 200 78 15 2 112 238 199 255 27 103 60 36 219 116 178 211 66 35 203 109 3 20 0 51 2 134 253 210 71 189 248 67 78 198 67 32 8 185 233 189 36 89 85 116 77 67 146 100 179 83 16 207 212 156 97 134 131 54 204 33 64 0 146 78 178 101 24 93 11 129 208 65 182 220 185 128 34 135 112 36 223 32 154 3 159 149 51 30 158 41 63 105 86 192 60 197 21 101 36 66 13 248 240 94 185 202 251 173 149 192 134 63 87 64 34 49 193 144 160 160 104 10 86 37 196 149 138 86 216 93 126 247 2 66 232 0 200 50 9 113 166 56 65 237 167 200 61 139 128 191 157 230 3 173 171 158 123 182 188 5 112 0 190 136 128 233 192 225 35 116 29 0 69 194 20 25 105 46 44 50 20 20 23 210 55 208 17 10 75 124 18 150 40 5 58 0 95 2 1 129 16 194 8 138 208 12 9 67 4 80 100 18 34 75 96 65 197 166 88 200 73 183 243 112 197 223 172 95 157 254 230 182 68 91 180 132 101 212 105 18 154 17 24 35 168 113 243 24 1 89 34 33 193 225 33 78 125 127 158 174 174 75 120 7 250 201 206 202 96 202 212 2 235 245 174 62 128 185 192 111 128 39 78 64 32 208 53 21 34 65 181 136 4 194 124 9 84 77 197 145 226 226 222 169 249 12 15 101 19 12 4 232 238 254 21 133 9 0 57 128 99 244 18 8 129 208 66 81 55 143 206 132 22 41 65 206 226 76 116 160 175 233 51 212 233 110 70 146 234 116 146 100 181 144 158 145 110 108 68 93 215 81 85 149 206 255 122 98 6 41 62 3 194 200 64 228 246 145 76 140 104 66 9 144 125 3 164 173 126 12 239 214 83 104 35 37 20 9 199 68 59 178 47 72 202 182 6 212 105 110 124 117 107 110 11 36 24 67 161 71 214 104 116 96 67 38 170 9 7 118 157 69 172 170 38 165 231 2 19 94 172 160 237 233 215 241 102 205 32 26 87 143 135 57 45 155 241 3 106 227 94 20 217 204 30 48 26 80 13 159 168 206 143 107 66 129 200 205 67 218 251 53 218 63 231 97 245 116 80 126 184 145 224 202 173 104 213 117 0 88 219 142 226 216 179 26 109 134 155 91 235 15 34 57 82 145 16 137 5 132 80 141 12 68 53 94 100 42 196 200 38 116 186 8 108 251 20 235 142 151 144 78 236 195 186 185 30 89 2 197 115 14 219 193 55 8 62 94 139 255 229 189 200 227 122 11 68 200 16 24 117 4 245 216 61 240 241 145 54 12 102 45 227 175 87 174 51 249 135 22 148 215 194 18 8 46 21 87 211 94 180 28 62 108 3 160 102 81 169 57 1 33 66 136 72 19 234 145 82 32 226 51 176 240 201 82 34 222 247 59 209 150 183 32 73 160 35 145 243 196 18 22 214 196 4 53 39 128 30 66 15 159 152 17 140 218 9 163 189 5 161 227 251 9 109 95 141 4 88 107 234 208 142 239 67 223 84 143 36 73 88 230 215 142 79 64 24 37 80 163 87 240 152 139 232 131 131 109 76 63 189 147 41 237 135 144 129 243 213 141 92 45 172 226 30 203 12 138 79 108 65 223 88 79 251 119 157 116 23 87 177 120 73 169 233 12 196 45 34 33 12 153 152 61 160 223 242 242 200 231 175 18 106 111 69 7 156 107 246 48 55 114 219 82 180 167 170 185 185 178 138 162 19 77 184 149 107 40 146 73 129 225 33 63 2 5 93 11 196 141 96 68 64 134 158 191 87 18 236 60 135 236 76 37 245 95 77 76 92 16 155 106 165 208 141 107 103 11 222 245 13 248 195 37 2 112 173 221 243 199 2 254 32 223 182 190 183 238 193 178 37 47 96 209 47 162 7 123 65 220 22 17 32 4 2 1 215 34 143 145 17 60 103 215 73 108 247 149 48 26 182 153 110 50 222 108 225 198 243 85 12 29 107 70 72 18 204 172 27 91 224 221 54 86 44 10 156 61 212 237 89 81 192 24 216 43 86 0 48 233 192 25 204 32 59 93 100 134 37 250 214 53 48 248 209 254 56 1 137 88 236 64 17 48 7 200 3 108 196 98 252 217 112 151 132 255 7 182 1 239 0 255 147 32 78 104 2 144 5 164 0 10 241 228 0 110 227 123 231 120 128 22 192 35 49 126 28 145 247 252 206 241 1 215 0 223 239 134 165 234 84 84 116 40 194,(4/0),73 69 78 68 174 66 96 130)
    _,←'Copy24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 3 0 0 0 215 169 205 202 0 0 0 105 80 76 84 69 255 255 255 0 0 0,(12/182),255 255 255 254 254 254 182 182 182 253 253 253 244 244 244 251 251 251 196 196 196 174 174 174 135 135 135 248 248 248 252 252 252 249 249 249 204 204 204 158 158 158 191 205 219 168 168 168 165 165 165 205 205 205 177 177 177 238 238 238 139 139 139 228 228 228 218 218 218 232 232 232 250 250 250 230 230 230 140 140 140 166 166 166 242 242 242 79 3 217 73 0 0 0 6 116 82 78 83 0 0 207 16 239 48 163 145 133 62 0 0 0 138 73 68 65 84 120 94 133 204 217 14 194 48 12 68 209 186 128 147 238 236 251 254 255 31 137 137 48 99 87 65 220 215 163 153 130 136 162 182 95 77 103 148 42 82 145 181 221 178 156 100 225 80 139 56 112 111 14 252 27 96 244 150 131 27 191 37 3 39 230 186 45 51 48 60 47 231 118 227 0 5 7 235 111 204 71 133 32 96 235 255 66 229 174 60 4 19 96 254 11 26 25 225 170 2 44 4 144 129 171 188 153 0 44 111 184 106 20 186 123 47 111 38 5 34 138 60 152 18 208 7 92 128 238 177 181 9 188 0 221 203 17 149 42 232 200 151,(4/0),73 69 78 68 174 66 96 130)
    _,←'Cut24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 3 112 73 68 65 84 72 137 157 149 95 104 28 85 24 197 207 157 157 221 54 38 89 111 186 171 77 178 214 110 27 16 243 162 19 164 254 43 200 22 52 10 197 182 160 125 16 132 110 80 17 138 176 70 209 215 54 40 136 190 84 31 171 152 172 1 125 40 106 86 76 81 68 152 180 90 165 109 140 219 166 77 186 173 201 110 75 11 166 155 236 110 166 59 59 119 102 238 157 241 161 93 77 220 174 153 241 123 186 204 156 239 252 190 51 247 114 7 174 235 194 117 93 120 173 15 70 190 163 239 125 156 201 175 165 171 251 202 158 157 111 149 44 7 85 215 225 113 207 122 66 136 103 243 244 209 211 7 59 218 239 80 174 23 139 222 7 242 172 4 176 46 20 76 109 221 180 17 165 197 197 219 190 63 50 49 67 29 7 42 113 237 231 1 204 1 128 228 213 188 164 219 113 57 32 209 206 72 24 129 64 99 234 175 142 207 82 66 160 58 174 171 152 166 217 86 127 238 25 176 161 53 88 144 37 130 160 76 32 75 171 1 227 191 228 104 64 34 234 198 142 176 66 28 27 218 178 54 231 27 0 0 114 128 32 36 75 171 18 124 127 234 18 149 8 81 59 35 97 37 122 231 122 44 22 139 211 175 189 240 100 245 255 3 2 255 36 80 167 254 160 178 4 181 51 218 174 244 110 190 11 51 23 243 48 106 181 247 87 245 248 1 0 0 33 4 32 192 143 147 151 40 64 212 174 104 187 210 19 139 226 231 201 89 92 189 246 103 250 237 151 119 127 190 82 239 43 129 45 92 88 220 129 11 9 92 64 141 210 86 229 254 123 239 198 169 108 14 231 115 243 95 190 158 124 118 224 223 61 190 0 92 184 208 25 71 27 141 34 66 219 148 173 177 40 142 159 158 193 185 92 62 157 74 238 218 123 187 30 95 0 102 218 24 87 79 66 215 171 216 210 29 193 79 147 179 56 123 97 62 189 255 197 157 13 147 251 6 108 219 63 149 88 40 46 33 40 203 88 90 42 225 196 111 179 152 190 144 79 167 246 53 126 150 149 213 176 201 3 223 88 9 71 56 41 118 131 81 99 217 56 166 151 244 12 211 216 33 83 55 19 92 138 160 198 170 8 134 214 99 238 242 181 204 91 47 237 250 79 243 6 192 222 225 82 210 208 140 17 135 59 89 83 55 169 169 155 9 110 242 3 220 230 16 182 168 124 173 242 244 3 79 116 237 185 146 61 31 191 56 189 112 204 75 114 169 126 173 246 191 155 167 181 114 237 144 94 210 7 143 12 116 244 25 203 70 31 211 24 172 154 5 97 138 44 128 45 191 126 242 248 160 21 222 60 36 117 245 2 192 110 79 128 250 130 85 89 82 47 235 149 111 83 221 31 2 128 182 160 81 86 101 176 13 59 235 8 103 199 239 195 15 87 0 64 112 49 17 108 105 65 32 20 82 122 250 15 239 241 14 208 24 53 202 6 0 224 161 87 38 169 205 236 49 187 102 103 5 23 59 166 62 221 86 169 235 70 159 107 41 172 107 13 163 247 153 167 105 236 193 190 177 251 118 14 231 123 250 15 199 215 4 88 53 171 96 234 102 92 73 158 204 115 139 231 109 102 67 216 226 239 201 87 22 183 56 0 169 18 238 142 13 133 187 55 81 0 7 214 4 100 211 143 164 185 197 211 220 228 5 97 137 143 206 142 62 214 119 102 244 209 6 243 58 192 17 206 192 248 96 236 160 235 98 16 64 162 25 96 213 41 58 247 197 246 53 143 29 0 112 147 227 232 155 247 100 0 192 21 110 1 0 109 154 128 16 2 63 191 77 224 102 130 167 222 153 143 223 132 177 125 0 38 154 2 124 57 223 42 97 137 2 55 249 216 246 55 206 140 48 173 156 4 240 89 51 173 239 235 26 0 184 197 135 204 27 218 72 249 114 78 17 22 27 154 251 225 213 76 51 237 95 45 33 158 80 19 223 48 202,(4/0),73 69 78 68 174 66 96 130)
    _,←'Delete24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 3 0 0 0 215 169 205 202 0 0 2 148 80 76 84 69,(9/0),241 179 170 66 44 43 110 47 47 238 163 157 225 100 92 179 55 52 57 15 14 147 126 123 235 145 131 173 87 83 124 103 102 237 153 142 146 82 76 13 7 7 55 23 22 227 112 107 239 147 134 198 84 75 75 12 12 3 0 0 142 99 95 214 106 93 47 20 19 54 23 23 170 81 77 238 132 119 132 40 38 16 1 1 70 42 41 243 143 129 235 90 70 96 35 32 110 47 46 228 93 81 206 78 66 74 13 13 0 0 0 8 4 4 102 47 44 218 111 100 121 35 28 165 68 63 225 56 32 175 41 31 31 4 4 20 6 6 137 48 42 131 38 31 214 77 67 140 20 14 14 2 2 48 10 10 183 37 23 112 17 13 3 0 1 6 1 1 101 15 11 201 29 4 197 109 100 126 21 15 0 0 0 146 58 53 199 37 13 199 36 15 50 15 14 55 25 25 219 91 82 173 35 24 194 38 20 21 6 6 55 23 23 219 86 76 128 28 24 1 0 0 127 28 23 155 35 23 57 28 28 219 91 80 152 34 27 7 0 1 118 28 22 83 22 18 166 69 61 176 36 24 18 2 3 131 32 25 205 45 26 18 5 5 167 76 71 197 41 22 41 6 7 1 0 0 158 40 31 127 27 19 67 31 30 228 79 58 98 17 15 13 3 3 199 46 29 220 39 14 32 7 7 218 85 73 172 35 23 8 1 1 47 11 10 226 39 13 133 25 15 107 31 30 69 12 11 117 26 21 215 42 30 13 2 2 218 65 61 20 2 3 7 2 2 195 37 33 83 14 14 172 60 54 199 47 31 7 0 1 80 23 23 155 25 23 131 45 40 150 34 28 0 0 0 100 21 20 17 4 4 199 70 59 199 58 44 38 6 6 20 2 2 90 17 16 36 5 6 193 25 0 227 58 34 216 44 18 232 130 116 237 149 136 208 57 32 206 45 20 210 46 20 225 35 7 213 31 3 202 27 0 223 57 32 199 28 3 195 26 1 190 24 2 212 64 41 207 49 25 208 47 22 238 144 130 237 99 77 195 31 6 197 32 6 218 67 43 209 54 30 209 48 23 195 27 2 240 150 136 240 152 137 230 51 23 238 142 127 199 39 15 229 96 76 219 72 49 211 56 32 208 48 25 202 44 20 208 53 32 237 131 115 221 51 25 207 33 7 231 101 80 221 76 54 212 58 34 205 47 28 236 110 91 234 136 121 228 83 64 232 70 45 202 37 11 232 107 87 223 81 59 215 61 37 205 45 21 216 65 42 229 51 24 224 37 8 219 37 9 236 112 92 225 87 66 219 66 42 229 59 34 211 35 8 230 117 101 232 127 111 202 32 9 230 94 74 205 45 20 206 47 23 233 91 69 231 69 45 235 58 32 227 110 92 231 125 108 242 138 121 223 48 21 240 132 115 195 25 0 236 114 96 208 51 27 207 47 22 205 40 15 214 54 29 207 28 2 235 144 130 119 212 46 246 0 0 0 136 116 82 78 83 0 3 2 185 55 28 172 194 205 53 90 248 123 49 208 145 11 30 163 250 242 138 25 151 219 46 6 126 246 196 50 102 253 247 97 51 211 249 135 11 18 140 246 121 121 250 232 78 41 182 137 187 210 45 83 227 187 26 17 148 254 231 195 14 131 254 247 52 4 164 235 229 22 7 183 197 25 165 178 5 183 218 29 149 95 159 236 50 163 236 20 116 254 89 7 192 149 49 250 161 22 227 251 39 203 235 28 62 252 160 66 122 137 247 20 138 43 10 221 111 152 254 28 95 184 126 209 5 150 19 225 251 81 31 147 83 246 120 136 143 0 0 1 75 73 68 65 84 120 94 141 143 83 151 35 1 20 6 111 122 108 219 182 109 219 182 109 207 218 202 216 182 109 219 54 214 127 102 55 233 164 131 57 115 206 212 83 221 250 158 46 60 14 23 3 55 51 35 77 97 226 100 97 101 3 118 142 171 235 5 89 30 94 44 243 241 11 8 10 9 3 136 28 30 253 45 111 189 17 21 67 179 184 132 100 153 148 52 2 0 56 25 171 222 190 129 137 121 57 121 66 87 80 108 83 82 86 1 20 85 53 245 187 239 231 23 151 26 0 154 90 253 218 58 56 192 208 213 91 92 218 220 218 214 55 56 254 99 104 4 52 24 155 84 215 212 214 55 52 154 154 1 29 230 22 150 247 120 188 181 13 60 192 214 174 189 163 199 222 129 62 59 58 13 142 141 59 79 78 185 184 210 118 55 247 229 245 13 15 79 175 253 3 111 154 238 227 123 250 235 183 159 63 32 1 37 165 129 212 67 80 69 101 85 112 8 193 66 235 194 194 41 61 162 169 185 37 50 138 168 209 49 183 177 88 143 139 255 241 243 44 129 244 111 98 82 114 10 121 72 237 236 234 78 75 39 95 25 153 89 36 203 30 26 30 25 205 1 140 220 188 124 84 10 166 103 102 231 10 1 163 168 248 25 42 207 87 86 215 94 188 4 10 175 94 163 242 102 103 119 239 237 59 160 226 61 66 148 15 31 79 62 125 6 106 16 146 124 249 250 13 7 79 231 31 61 50 94 172 9 118 14 119,(4/0),73 69 78 68 174 66 96 130)
    _,←'Open24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 0 7 116 73 77 69 7 218 9 9 21 38 53 128 233 158 142 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 4 170 73 68 65 84 120 218 229 85 107 136 85 85 20 254 206 185 231 206 75 103 196 116 198 208 17 33 66 44 50 8 235 87 98 20 209 63 37 162 126 132 68 18 21 82 65 80 73 6 18 97 16 22 84 164 148 61 40 178 126 84 246 203 130 194 30 12 13 147 137 17 190 198 28 199 7 227 52 211 157 215 29 207 189 247 188 247 171 111 159 123 198 98 160 96 126 119 96 113 247 217 103 175 111 125 235 91 107 175 11 252 239 159 251 159 124 221 161 185 255 246 221 89 40 224 158 15 127 116 46 143 142 150 174 248 245 23 165 20 15 16 98 157 49 230 180 86 98 95 181 50 114 176 255 171 247 234 59 95 61 128 61 207 63 108 236 121 111 33 224 219 118 236 119 6 207 94 132 235 136 205 4 221 94 191 50 181 53 139 195 20 48 232 94 117 125 223 178 149 215 221 185 168 243 154 71 206 92 172 62 202 227 239 208 204 130 2 56 14 220 36 106 116 105 173 150 94 60 125 100 253 240 241 62 25 135 53 201 79 106 211 150 199 111 233 89 189 182 255 238 7 119 236 133 235 110 229 222 71 180 204 93 72 128 21 203 58 52 127 130 201 209 115 159 15 159 232 171 19 60 228 187 181 232 212 47 223 158 205 82 241 74 185 220 250 24 169 116 112 111 145 85 200 249 226 192 107 55 115 113 82 73 1 99 174 50 133 84 6 34 19 191 251 126 253 161 103 119 189 121 210 178 156 87 59 107 214 195 80 243 210 208 136 255 18 253 119 210 215 179 56 135 222 125 186 151 223 102 157 79 63 120 217 220 116 227 181 178 119 205 18 175 189 165 19 229 178 195 64 13 216 104 99 151 103 49 112 116 2 65 109 250 158 167 94 120 187 143 14 242 63 18 108 161 149 105 29 197 58 163 213 157 79 222 223 109 54 111 185 21 237 109 89 243 152 38 81 25 193 81 49 57 186 152 153 14 209 255 243 36 188 146 134 209 2 81 162 145 166 41 100 146 33 8 99 190 11 248 126 140 183 14 124 221 94 100 52 151 157 149 83 228 233 180 182 177 214 38 161 8 108 8 205 13 153 193 216 181 104 96 185 55 131 123 239 104 129 75 66 142 230 190 108 122 42 225 65 200 50 2 63 193 174 253 53 148 61 119 137 235 58 181 13 107 123 242 207 71 6 43 185 164 158 177 128 92 186 58 102 120 3 93 106 197 192 192 101 212 27 17 146 40 70 150 41 40 21 64 74 201 181 128 20 138 192 92 167 25 45 69 20 39 232 234 242 176 237 190 219 39 226 152 123 89 130 202 120 195 102 210 158 103 160 181 102 62 58 103 231 24 133 161 161 113 68 97 3 119 109 92 129 82 137 153 58 1 76 70 249 100 76 245 152 1 207 65 211 195 8 56 210 32 101 64 87 41 68 145 100 83 40 92 26 171 227 141 143 207 79 17 188 203 214 128 18 25 6 80 116 84 72 120 248 248 137 49 108 218 184 26 109 45 204 39 157 133 67 150 198 68 4 76 225 146 185 161 148 74 164 4 215 204 42 134 67 198 146 204 101 144 98 124 58 198 190 207 34 84 253 240 9 130 151 114 137 180 110 6 208 60 124 233 130 143 149 43 186 208 179 188 21 58 245 161 227 42 180 8 242 218 229 23 38 175 17 193 181 164 164 18 229 52 230 86 198 51 182 54 33 70 43 148 171 85 255 249 205 185 169 223 120 58 182 93 231 218 26 168 44 206 83 29 26 158 197 13 235 22 195 205 106 205 0 116 114 9 238 81 81 199 202 195 206 50 100 12 130 90 51 204 196 250 202 36 193 84 67 227 240 49 3 233 100 187 139 22 77 45 37 79 43 77 109 5 254 24 173 161 187 187 29 75 59 37 84 66 105 68 68 150 34 103 171 8 218 12 32 200 150 237 106 187 140 89 43 182 170 136 88 216 56 194 248 132 68 119 167 241 247 30 170 28 46 216 139 171 18 9 22 235 212 217 42 110 219 208 3 36 53 118 108 68 118 97 126 23 44 67 75 192 40 246 167 104 222 5 187 150 25 139 154 196 8 163 8 147 129 192 119 191 150 16 200 108 123 193 60 45 238 1 242 34 87 167 167 120 209 90 209 221 37 154 186 103 212 61 75 115 224 188 187 50 149 119 143 102 183 200 36 165 58 22 60 69 64 240 48 142 49 225 151 80 118 202 103 14 254 48 117 212 206 165 66 34 83 100 160 49 124 97 26 107 122 61 56 81 5 130 133 147 86 2 22 93 178 120 174 96 80 118 146 102 16 41 210 60 128 228 94 204 189 152 18 205 164 30 190 28 232 192 180 223 120 134 120 201 156 246 115 243 35 191 104 42 180 163 96 4 35 19 117 100 28 19 150 169 49 154 237 168 33 132 204 47 89 34 220 124 62 197 116 143 83 118 29 107 23 37 101 204 134 158 29 29 223 255 116 106 118 232 159 218 255 29 128 137 156 31 235 192 241 96 21 130 96 49 47 91 146 183 158 189 145 246 168 148 236 20 164 57 123 157 105 36 236 255 122 104 167 173 131 48 49 124 207 6 43 213 240 185 2 60 153 55 117 243 161 180 190 152 128 182 225 235 197 190 89 192 223 132 41 52 15 139 0 122 126 128 82 49 94 75 115 243 125 129 1 80 176 150 243 193 237 243 23 55 105 53 22 92 39 193 114,(4/0),73 69 78 68 174 66 96 130)
    _,←'OpenObject24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 0 7 116 73 77 69 7 218 9 9 21 37 36 193 116 237 191 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 4 143 73 68 65 84 120 218 229 84 125 104 85 101 24 255 157 115 239 221 221 102 187 155 205 109 50 103 66 132 68 168 16 218 63 41 70 17 245 151 18 145 65 72 36 81 161 21 4 149 100 32 17 6 161 65 69 74 173 15 250 152 65 214 252 107 5 133 125 13 199 50 233 131 233 156 109 78 39 219 117 219 157 155 119 158 123 238 61 159 239 87 207 123 238 217 90 163 13 246 119 7 94 206 251 249 60 191 231 247 252 158 7 248 223 127 15 61 243 166 65 195 92 232 220 88 170 193 131 31 255 100 140 100 179 137 235 150 253 10 231 108 7 153 184 85 41 117 78 10 118 36 159 27 110 239 250 250 3 123 223 161 54 28 124 233 49 165 239 39 151 98 124 215 222 86 163 175 127 8 166 193 182 145 209 221 246 245 201 157 161 231 4 128 66 195 170 91 58 235 155 111 190 123 89 205 141 143 159 31 202 63 65 215 223 163 161 150 228 192 48 96 250 110 49 35 165 88 62 116 238 212 250 193 158 78 238 57 5 78 71 98 235 246 167 110 111 92 189 182 235 222 71 246 30 134 105 238 164 189 79 104 132 230 82 28 52 213 87 75 250 149 174 102 47 124 57 120 166 211 38 227 14 173 245 112 123 127 253 174 63 12 216 235 169 84 250 73 130 82 77 123 203 52 67 198 87 109 111 108 160 201 89 193 25 148 154 69 10 46 20 88 200 254 178 44 251 209 23 246 191 125 86 163 156 151 59 61 244 11 69 156 39 6 134 173 87 233 253 62 122 155 212 118 58 222 127 174 133 206 166 141 207 63 122 77 173 187 109 37 111 89 83 155 172 170 168 65 42 101 144 163 34 180 183 209 145 105 116 159 158 64 169 48 117 223 179 47 191 219 73 15 248 34 1 86 208 72 209 168 142 231 33 13 219 56 250 225 1 181 109 251 38 84 85 134 229 107 146 128 114 23 134 240 8 163 137 107 83 14 186 126 185 138 100 66 66 73 6 215 151 8 130 0 220 15 81 114 60 90 51 88 150 135 119 218 190 169 138 35 154 137 78 211 201 162 112 210 149 148 107 229 19 9 36 8 73 27 60 132 210 115 86 196 138 228 53 60 112 87 5 76 2 100 72 218 231 229 151 130 37 193 120 10 37 203 199 254 214 2 82 73 179 214 52 141 194 198 181 141 209 241 169 190 92 68 105 82 105 131 52 53 165 71 238 21 100 34 141 238 238 17 216 69 23 190 235 33 12 5 132 40 129 115 78 115 6 206 4 25 166 121 16 210 8 224 122 62 50 153 36 118 61 184 121 194 243 104 47 244 145 27 43 234 72 170 162 8 164 148 20 143 140 208 25 74 96 96 96 12 174 83 196 61 91 154 80 119 232 232 130 132 231 247 108 130 193 21 2 114 104 10 1 215 229 36 10 129 203 163 54 222 250 236 226 36 93 201 232 28 16 69 138 28 8 226 87 192 167 203 61 103 70 177 117 203 106 84 86 168 69 37 91 223 250 71 244 159 216 145 1 39 228 188 20 96 108 202 195 145 99 46 242 150 243 52 29 37 34 138 164 44 59 144 220 195 229 75 22 154 155 50 104 92 145 134 12 172 89 99 246 238 117 136 10 134 242 162 100 0 41 57 106 63 29 143 206 86 30 183 49 126 63 41 143 59 200 230 136 174 180 28 255 246 194 228 159 116 228 105 213 69 57 16 161 7 70 232 7 6 167 177 249 206 70 152 97 129 108 89 176 247 108 128 169 36 37 138 52 64 20 74 82 150 98 164 54 33 97 61 92 135 186 246 50 8 238 251 152 44 74 156 248 77 129 27 225 129 88 162 129 134 148 148 116 153 26 21 174 100 11 104 104 168 194 242 26 14 225 79 195 96 46 37 158 81 110 104 77 137 51 132 86 22 131 100 36 87 173 178 57 117 23 58 14 198 38 56 26 106 148 117 184 35 119 34 70 207 102 41 98 148 172 222 254 60 238 216 216 8 248 5 82 172 75 72 157 168 22 116 116 26 128 18 164 79 86 174 5 61 231 161 174 185 114 167 25 47 250 248 254 247 4 74 60 220 29 35 15 226 58 64 148 228 252 212 36 21 90 26 13 25 66 232 229 33 195 18 193 10 34 195 145 186 66 17 83 36 136 142 0 1 25 191 233 100 122 54 130 137 60 71 202 72 157 111 255 113 242 180 238 75 49 69 42 142 64 98 240 210 20 214 180 36 97 184 57 176 192 35 205 235 130 242 208 116 44 255 31 250 73 204 8 36 250 58 154 179 56 254 115 53 166 172 226 243 180 244 103 184 159 57 47 39 217 209 173 96 24 195 19 54 66 106 19 26 169 162 228 54 161 118 65 153 182 213 101 169 77 24 152 190 146 210 173 227 135 147 189 211 3 115 185 255 199 1 5 114 113 180 26 61 165 85 40 149 110 160 98 243 73 65 229 138 252 130 174 114 30 16 28 234 61 140 40 11 37 124 46 97 59 68 149 72 194 241 21 173 195 190 92 222 121 49 54 238 207 235 186 81 83 90 31 119 192 146 150 124 188 191 120 149 253 251 83 49 231 78 236 64 206 119 160 9 173 136 255 106 206 88 202 167 81 243 249 198 245 247 55 184 3 200 193 75 93 12 60,(4/0),73 69 78 68 174 66 96 130)
    _,←'Options24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 32 0 0 0 32 8 6 0 0 0 115 122 122 244 0 0 1 79 73 68 65 84 120 218 237 150 209 146 195 32 8 0 225 203 131 95 238 197 88 175 8 66 48 167 157 123 40 47 237 196 150 93 81 137 8 0 25 54 68 206 118 90 68 246 189 254 118 173 3 158 4 43 39 135 247 2 114 164 159 206 180 0 17 93 223 143 227 48 225 165 246 189 192 8 116 62 79 103 50 158 232 137 192 8 94 232 31 17 176 224 122 9 54 8 16 137 255 49 248 118 129 59 248 180 0 138 90 218 59 29 65 157 238 1 92 111 66 35 210 53 27 10 193 95 206 33 120 191 9 11 40 165 97 194 93 240 126 9 156 224 165 95 9 239 5 140 37 224 79 87 195 137 82 47 208 198 56 168 205 126 7 92 85 64 10 68 75 95 210 32 230 105 184 41 32 131 110 250 64 74 8 101 24 219 185 10 194 149 64 233 3 234 172 3 184 141 168 193 203 1 186 36 48 135 225 87 111 241 4 218 115 191 19 226 47 156 127 210 145 111 225 186 2 172 44 109 2 224 86 96 12 151 149 176 224 189 0 188 27 17 13 58 223 72 160 94 60 192 149 240 224 74 96 54 234 75 103 12 175 115 9 76 2 30 222 136 248 94 145 18 101 40 82 65 45 48 241 58 150 167 165 74 188 251 7 23 240 142 240 18 1 185 132 214 157 112 169 128 155 244 43 240 88 192 136 207 8 128 125 35 138 36 90 34 176 42 166 5 150 210 95 17 22 104 211 247 202 255 151 248 247 2 63 54 177 133 208 40 45 121 35,(4/0),73 69 78 68 174 66 96 130)
    _,←'Paste24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 3 0 0 0 215 169 205 202 0 0 0 120 80 76 84 69 0 0 0 123 123 123 156 156 156 181 181 181 206 165 74 206 165 82 206 173 90 231 206 156 247 247 247 255 255 255 206 173 99 231 231 231 206 206 206 214 181 107 214 189 123 132 132 132 231 206 165 231 214 165 148 148 148 90 90 90 181 156 99 189 206 222 140 140 140 165 123 66 165 132 57 165 140 90 165 165 165 214 173 99 173 132 66 173 148 90 214 214 214 222 198 132 231 206 148 181 132 49 181 148 99 148 115 66 140 99 24 239 239 239 189 189 189 198 173 115 190 189 26 212 0 0 0 1 116 82 78 83 0 64 230 216 102 0 0 0 163 73 68 65 84 120 94 141 205 87 14 195 32 16 69 81 211 139 123 75 239 117 255 59 12 227 120 132 177 137 148 251 133 116 244 134 228 91 9 37 97 87 136 80 74 201 3 94 19 168 235 170 106 9 33 109 167 148 10 224 92 20 13 116 200 139 0 170 60 183 239 238 118 220 110 164 156 129 117 49 140 50 134 96 7 48 99 105 6 18 1 157 25 144 37 152 147 6 65 144 30 82 150 189 16 164 116 224 123 166 187 1 164 11 160 199 140 89 1 40 0 225 96 82 57 130 16 81 16 80 112 202 3 231 145 197 101 207 121 28 134 34 167 214 119 104 177 192 152 209 190 25 244 152 254 107 241 251 143 166 156 150 124 0 54 97 22 53 163 234 55 133,(4/0),73 69 78 68 174 66 96 130)
    _,←'Print24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 32 0 0 0 32 8 6 0 0 0 115 122 122 244 0 0 7 22 73 68 65 84 120 218 165 87 105 76 148 103 16 30 64 64 20 202 37 85 20 16 80 139 21 176 96 66 53 30 9 72 75 209 128 24 90 21 21 228 208 160 184 197 26 211 104 172 54 13 82 16 77 181 2 233 15 126 212 160 38 198 11 43 98 61 90 34 162 212 162 130 114 168 171 171 120 112 46 2 11 44 247 177 208 121 94 249 40 114 232 42 95 50 217 239 156 231 153 153 103 230 125 87 135 6 28 145 145 41 246 61 61 154 197 189 189 61 142 108 186 75 150 184 83 109 109 203 194 158 158 30 93 99 99 35 178 183 255 248 35 181 186 205 85 122 255 213 171 250 7 201 201 7 150 21 20 92 126 206 151 189 244 1 135 206 160 107 189 136 136 228 140 31 127 90 189 196 198 198 82 43 7 251 18 142 43 119 239 90 235 195 167 247 63 132 196 96 2 228 230 230 235 232 243 213 218 226 93 49 193 227 180 113 160 82 53 209 234 192 136 115 185 185 103 162 249 178 124 212 4 112 111 213 170 216 223 194 55 173 140 154 243 249 39 90 57 73 220 159 218 16 23 19 30 198 167 127 177 181 141 150 0 205 159 31 100 227 237 243 77 217 250 173 95 107 229 164 40 79 65 209 235 87 30 126 249 178 112 23 95 42 71 77 0 247 67 67 15 254 45 219 189 193 219 212 220 228 157 78 26 235 155 104 75 72 248 189 220 220 180 72 190 44 96 235 30 45 1 90 183 238 151 31 150 174 246 143 115 114 211 178 12 187 247 215 165 254 190 227 91 62 61 207 214 58 106 2 203 151 239 248 116 193 151 190 15 23 5 120 106 229 232 122 250 53 218 46 243 138 229 211 67 108 170 81 19 224 67 111 235 247 199 74 151 201 130 39 107 227 232 254 141 124 218 183 51 228 66 69 133 60 138 222 163 27 222 70 128 34 34 146 46 45 223 182 209 87 223 208 96 196 119 122 184 243 187 52 68 53 101 85 116 240 187 160 127 229 242 235 50 190 93 76 90 234 224 173 4 194 194 14 37 44 88 179 114 135 133 141 53 105 58 25 140 93 106 96 154 215 214 201 247 44 63 34 50 49 98 211 105 34 131 246 114 202 200 56 217 18 19 19 51 139 63 47 29 21 129 189 123 247 62 119 113 153 99 223 169 231 74 29 26 107 113 207 210 130 205 156 205 140 71 158 70 77 198 134 29 212 218 218 74 45 45 45 196 227 154 116 117 117 233 206 157 59 116 236 216 177 127 38 76 152 144 213 209 209 113 252 220 185 115 10 254 84 243 222 4 14 28 56 80 186 121 243 102 91 29 29 29 225 248 197 139 23 84 87 87 199 81 119 82 99 99 35 25 26 26 146 129 129 129 248 149 206 113 220 187 119 143 140 140 140 196 53 222 187 121 243 102 222 163 71 143 54 94 187 118 173 112 56 34 67 8 132 135 135 155 113 52 238 46 46 46 201 161 161 161 206 210 253 188 188 60 17 233 64 64 149 74 69 106 181 154 234 235 235 201 212 212 148 198 140 25 195 11 150 61 57 56 56 80 119 119 55 49 176 184 159 145 145 209 158 153 153 185 158 125 164 177 171 142 17 9 48 96 216 164 73 147 14 77 157 58 213 20 145 91 90 90 138 40 172 172 172 168 183 183 151 42 42 42 68 202 141 141 141 201 196 196 132 38 79 158 76 252 62 89 91 91 139 231 40 195 192 95 24 50 2 63 105 105 105 29 87 174 92 145 21 21 21 29 167 1 227 186 159 64 112 112 176 167 187 187 123 150 191 191 191 96 223 213 213 69 250 250 250 34 157 133 133 133 188 244 190 162 185 115 231 210 180 105 211 134 5 26 238 30 50 86 86 86 38 74 55 99 198 12 218 179 103 79 205 217 179 103 3 25 46 87 234 146 126 2 65 65 65 233 172 222 101 0 5 1 137 4 28 113 70 168 188 188 156 148 74 165 32 49 18 56 0 107 106 106 120 15 81 43 190 69 150 204 205 205 133 225 157 83 167 78 209 137 19 39 206 23 20 20 160 85 203 223 32 176 98 197 138 27 137 137 137 11 37 240 129 36 240 49 106 91 90 90 74 205 205 205 52 107 214 44 1 8 65 34 51 0 5 56 178 133 116 91 88 88 144 158 158 158 248 174 173 173 77 60 67 233 88 144 116 250 244 105 57 107 97 27 67 102 163 20 253 4 2 3 3 115 226 226 226 22 64 52 195 145 128 67 91 91 91 186 123 247 174 0 134 33 91 168 63 0 33 64 144 194 187 16 39 136 54 53 53 9 223 32 130 50 112 253 233 234 213 171 10 214 197 207 124 251 15 182 230 126 2 1 1 1 57 51 103 206 92 224 229 229 69 179 103 207 30 150 4 82 106 102 102 38 156 35 90 41 253 232 4 0 0 184 189 189 189 63 250 234 234 106 97 208 1 90 153 103 3 93 186 116 73 193 37 216 203 144 103 222 32 224 231 231 247 167 76 38 91 138 52 63 120 240 64 212 218 195 195 67 68 38 145 208 240 248 131 234 145 78 0 34 253 13 13 13 130 8 64 145 21 148 3 221 130 247 167 76 153 34 90 178 72 125 157 186 28 148 84 123 220 148 110 221 186 165 224 76 12 37 224 235 235 155 16 27 27 187 195 201 201 73 164 46 39 39 135 242 243 243 201 206 206 142 55 40 243 69 235 193 105 73 73 137 0 1 49 68 5 50 136 18 68 208 186 136 18 162 149 116 0 203 126 153 78 173 166 213 164 147 107 71 23 47 94 84 20 23 23 15 37 224 227 227 19 159 144 144 176 147 203 208 31 49 136 64 56 232 101 27 27 27 226 54 21 81 35 75 82 218 161 3 60 67 102 64 82 2 149 12 68 159 61 123 70 8 44 59 59 155 142 30 61 170 224 12 15 37 224 237 237 29 207 227 119 167 179 179 179 168 247 96 13 64 124 112 128 193 132 40 209 90 19 39 78 164 241 227 199 15 1 69 253 145 37 104 130 69 39 74 18 31 31 79 89 89 89 148 158 158 174 144 203 229 67 9 120 122 122 198 243 24 222 25 18 18 34 106 61 28 9 24 215 80 252 14 6 133 65 108 0 6 73 136 20 130 69 87 141 27 55 142 42 43 43 197 44 225 32 202 56 152 24 134 60 249 6 1 78 209 20 126 241 87 158 88 1 209 209 209 6 243 230 205 27 150 4 43 88 116 129 4 250 228 201 19 81 18 148 102 236 216 177 253 160 210 47 180 128 76 97 142 32 115 145 145 145 170 212 212 84 180 225 97 182 198 129 107 1 206 109 57 181 43 89 68 178 69 139 22 217 71 69 69 137 15 7 19 128 54 0 14 130 73 73 73 98 248 12 4 149 126 161 39 140 96 172 37 16 244 145 35 71 26 120 169 150 179 96 207 246 17 80 13 94 13 113 205 91 12 154 206 3 102 3 47 54 171 120 62 152 111 218 180 73 68 7 2 16 228 255 4 198 242 138 167 166 170 170 124 238 136 22 1 138 97 133 22 70 251 161 61 47 92 184 64 41 41 41 74 38 80 194 153 195 38 69 206 150 197 118 155 173 115 164 253 128 30 27 254 155 185 58 58 58 110 231 140 120 241 168 214 231 5 75 136 17 181 198 198 227 249 243 82 158 132 110 228 234 186 152 193 90 105 241 98 71 114 118 158 193 132 170 0 218 205 155 145 42 133 66 241 148 199 113 25 182 10 108 183 216 74 216 26 0 254 214 13 73 223 129 93 198 36 22 212 23 220 106 91 56 165 159 33 202 199 143 31 163 239 149 44 182 74 118 222 56 125 186 147 139 135 135 159 149 167 167 27 101 102 158 111 191 124 249 114 41 139 241 41 151 72 2 70 180 47 217 212 52 104 175 248 46 2 210 97 204 102 199 181 94 195 100 124 120 85 172 227 114 84 211 235 205 231 67 68 195 117 94 195 101 113 230 103 21 125 96 183 251 192 241 79 169 153 70 216 150 105 75 64 122 215 28 13 195 134 77 226 35 182 74 122 189 185 128 243 9 125 207 140 251 136 213 246 61 123 235 63 230 255 0 23 154 58 93 23 114 7 135,(4/0),73 69 78 68 174 66 96 130)
    _,←'PrintPreview24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 32 0 0 0 32 8 3 0 0 0 68 164 138 198 0 0 0 7 116 73 77 69 7 218 6 3 1 10 48 63 112 125 182 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 3 0 80 76 84 69 66 90 123 74 99 173 74 123 206 74 132 206 82 115 198 82 123 198 82 123 206 82 132 206 82 132 214 90 123 189 90 123 206 90 132 206 90 132 214 90 140 214 99 107 123 99 115 132 99 132 206 99 140 198 99 140 214 107 115 123 107 123 140 107 132 156 107 156 222 115 123 140 115 132 148 115 132 165 115 140 156 115 140 206 115 148 206 115 156 222 123 132 148 123 132 156 123 140 156 123 140 165 123 140 173 123 148 189 123 156 206 123 165 222 132 132 148 132 140 140 132 140 148 132 156 189 132 156 198 132 165 222 132 165 231 132 173 231 132 181 231 140 148 173 140 156 165 140 156 173 140 165 189 140 165 214 140 165 222 140 173 206 140 173 214 140 173 231 140 181 231 140 189 239 140 189 247 148 140 140,(4/148),156 173 148 165 181 148 173 198 148 173 214 148 173 222 148 173 231 148 181 222 148 181 231 148 181 239 148 189 247,(5/156),173 156 165 173 156 165 198 156 173 222 156 181 222 156 181 231 156 189 231 156 189 239 156 198 247,(5/165),173 165 173 181 165 173 189 165 173 198 165 181 206 165 181 231 165 189 222 165 189 231 165 189 239 165 198 239 165 198 247 165 206 247 165 206 255 173 165,(6/173),181 173 181 198 173 189 206 173 189 214 173 189 222 173 189 231 173 189 239 173 198 222 173 198 231 173 198 239 173 198 247 173 206 239 173 206 247 173 206 255 181 173 173,(5/181),189 181 181 198 181 189 222 181 198 214 181 198 222 181 198 239 181 206 231 181 206 239 181 206 247 181 206 255 181 214 247 181 214 255,(5/189),198 189 189 214 189 198 214 189 206 231 189 206 239 189 206 247 189 214 239 189 214 247 189 214 255 189 222 247 189 222 255 198 189 198 198 198 206 198 206 231 198 214 231 198 214 239 198 214 247 198 222 247 198 222 255 198 231 255,(5/206),214 206 214 222 206 214 239 206 214 247 206 222 247 206 222 255 206 231 255 214 206 214 214 214 222 214 222 231 214 222 239 214 222 247 214 231 247 214 231 255 214 239 255,(5/222),231 222 222 239 222 231 247 222 231 255 222 239 255 231 222,(6/231),239 231 239 247 231 239 255 231 247 255,(5/239),247 239 239 255 239 247 255,(5/247),255 247,(227/255),69 128 30 191 0 0 0 182 116 82 78 83,(181/255),0 138 230 189 112 0 0 2 164 73 68 65 84 120 218 109 211 253 79 211 64 24 7 112 36 190 68 150 32 34 138 47 168 248 130 162 132 248 130 58 141 16 71 48 46 202 219 6 130 142 182 9 32 78 86 215 136 111 179 75 131 194 194 13 132 172 215 72 150 163 32 180 98 74 179 172 148 189 192 255 103 175 215 105 151 248 253 161 63 244 62 121 238 201 221 61 21 123 36 187 186 129 147 203 23 247 202 83 97 127 139 134 174 129 4 207 43 91 186 166 105 217 172 150 205 187 65 94 143 199 222 176 27 138 162 164 18 137 4 203 82 172 162 41 154 246 23 228 243 249 109 67 69 192 90 25 65 10 14 140 246 178 60 210 74 32 37 35 17 234 134 102 181 144 77 179 35 161 190 158 153 193 155 29 183 134 65 9 36 100 4 32 192 17 192 130 174 27 203 79 59 134 39 39 38 158 159 55 29 192 227 10 128 8 65 136 127 186 240 42 165 170 72 149 191 12 108 186 0 132 243 132 204 158 158 252 5 101 25 42 42 66 193 2 1 170 12 37 156 100 210 18 131 119 101 149 128 148 254 115 186 28 72 18 0 179 141 81 73 162 4 129 2 144 55 140 160 3 84 105 133 148 0 96 252 17 0 243 20 16 162 34 228 245 173 177 140 3 86 228 21 73 224 112 110 76 8 80 157 131 144 71 104 110 219 120 191 102 3 93 149 165 56 71 211 12 195 49 22 152 151 121 40 70 33 74 236 228 70 55 49 136 233 91 192 94 102 172 10 183 199 1 144 163 16 176 34 138 236 228 218 77 2 32 109 1 44 24 174 175 67 180 218 19 69 54 133 34 185 223 15 237 38 99 50 176 1 201 197 25 8 35 80 140 165 55 194 197 246 53 27 176 11 110 64 95 89 66 33 171 201 244 198 172 63 72 78 50 86 6 168 175 239 198 62 35 148 90 95 174 30 45 16 192 74 46 64 125 155 58 118 188 162 190 185 174 250 200 190 210 109 186 183 160 62 124 191 122 137 243 180 93 62 236 11 31 40 129 72 18 208 111 29 48 180 232 175 163 57 207 99 111 109 79 232 31 0 160 180 193 208 234 64 125 39 205 29 242 122 107 187 67 251 255 110 1 146 206 250 240 218 84 243 53 154 0 95 168 170 96 133 84 112 192 203 213 105 255 41 218 6 45 53 221 47 170 76 211 44 20 44 16 118 64 104 241 135 191 174 31 31 41 6 157 189 53 102 38 67 142 26 8 118 143 207 54 131 231 238 217 180 210 219 226 241 245 212 100 50 228 69 173 75 2 190 75 186 171 241 228 117 188 28 160 43 239 52 121 124 189 103 204 66 105 112 52 158 225 226 28 75 245 147 179 160 43 91 154 60 71 219 210 174 209 219 213 248 190 39 12 54 84 32 64 209 7 59 155 26 94 175 151 38 107 215 25 194 108 170 235 62 126 17 20 174 112 98 169 124 120 157 20 195 173 15 2 12 39 52 124 76 239 253 23 224 126 210 161 214 179 75 174 31 127 0 5 196 1 113 19 164 228 220,(4/0),73 69 78 68 174 66 96 130)
    _,←'Redo24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 3 118 73 68 65 84 72 137 181 149 93 111 84 69 28 135 159 223 156 179 123 118 219 221 237 169 136 54 52 74 185 106 137 21 54 126 130 106 148 104 76 149 151 196 27 9 82 189 55 248 1 12 233 71 48 122 47 17 13 168 224 53 196 151 93 189 54 161 198 82 111 124 105 77 68 160 182 93 186 123 78 123 222 102 188 40 69 72 160 52 177 60 87 147 201 204 239 249 207 75 102 196 22 140 29 187 216 4 38 144 66 73 128 64 154 1 181 127 249 226 181 206 86 115 55 209 253 58 199 223 248 234 109 208 251 131 3 253 35 245 70 157 114 80 197 90 40 172 99 181 219 163 219 139 72 179 188 141 52 61 119 238 213 246 131 194 159 121 243 82 243 30 193 232 145 47 67 164 214 211 195 187 154 195 123 134 144 87 101 53 46 136 147 2 235 192 24 131 239 121 224 32 142 35 86 86 254 33 73 226 54 210 212 213 207 94 153 191 39 252 248 229 38 208 210 221 225 165 146 215 58 176 127 164 89 173 13 114 99 37 165 187 86 16 148 60 144 40 172 35 201 44 105 86 224 121 134 160 28 32 121 196 221 21 226 104 169 227 176 71 102 207 30 106 3 140 159 248 186 233 25 191 229 156 13 239 8 198 142 94 104 29 216 191 119 162 62 248 4 55 150 215 105 212 2 140 28 73 146 208 139 19 140 87 194 248 21 178 194 208 91 75 201 108 142 145 193 247 3 92 158 177 218 249 11 107 139 41 196 140 103 252 86 99 112 56 92 237 92 219 56 131 177 163 23 78 62 53 188 235 227 125 251 246 177 26 21 4 101 195 226 226 34 215 111 46 179 158 230 109 164 239 145 14 250 190 63 209 215 55 16 86 250 31 167 192 144 219 4 36 74 155 146 149 107 29 36 6 6 247 132 126 57 96 249 230 60 62 128 239 123 167 135 247 12 145 230 16 148 197 175 191 253 206 242 173 232 140 164 247 230 206 79 118 254 219 215 75 97 183 215 57 21 175 199 167 107 141 33 202 149 62 50 187 78 225 82 74 229 10 181 129 221 161 239 7 120 165 50 146 1 25 52 118 236 98 115 247 99 245 43 163 163 163 228 22 22 254 92 96 113 233 214 135 179 231 38 223 125 208 237 24 63 241 77 83 198 180 234 141 39 195 160 175 78 230 18 36 67 217 11 0 71 234 82 2 175 194 210 141 5 12 112 184 222 168 147 59 209 235 117 185 190 216 105 111 21 14 48 251 201 139 51 192 243 81 111 169 147 103 25 214 150 88 203 10 162 44 37 202 19 114 235 0 33 25 12 210 193 106 181 159 162 128 107 127 223 68 210 244 86 225 155 252 124 230 133 25 135 155 142 163 101 140 60 36 159 212 90 50 11 153 21 153 5 36 124 80 88 169 84 73 179 148 110 180 54 63 119 126 178 189 29 193 129 119 126 104 26 227 159 174 212 118 129 1 112 120 18 146 64 96 204 198 10 124 73 228 78 68 81 76 97 221 246 195 61 191 213 223 24 10 77 41 32 39 195 247 61 100 54 4 18 120 70 183 87 32 145 164 57 105 238 64 90 216 94 229 94 171 191 182 59 244 60 31 97 9 228 35 195 70 245 220 126 127 156 69 198 224 131 102 162 120 125 34 77 179 237 20 79 185 90 187 2 144 164 93 72 187 15 29 239 35 126 138 163 53 156 41 221 169 96 43 126 252 232 185 135 15 186 11 35 212 142 215 34 10 235 241 128 199 245 127 97 230 62 159 156 79 210 100 198 218 12 164 189 59 46 0 64 250 32 207 214 65 52 119 90 112 103 79 198 143 95 254 67 198 27 113 206 14 206 158 61 180 173 223 106 59 152 205 134 195 77 91 91 0 58 188 83 225 247 8 174 126 250 242 25 73 109 196 235 143 68 176 129 166 64 19 227 111 125 27 62 18 193 236 217 151 230 145 142 72 58 181 83 130 251 242 236 201 239 70 118 42 235 95 227 150 105 161 245 191 89 3,(4/0),73 69 78 68 174 66 96 130)
    _,←'Run24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 4 3 0 0 0 18 89 32 203 0 0 0 7 116 73 77 69 7 218 9 17 23 22 24 140 89 95 22 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 0 42 80 76 84 69 0 0 0 214 231 198 214 222 198 239 239 222 41 123 33 74 148 57 123 181 115 90 165 74 123 198 107 148 222 140 49 132 41 82 148 74 206 255 198 107 173 99 166 53 48 24 0 0 0 1 116 82 78 83 0 64 230 216 102 0 0 0 116 73 68 65 84 120 218 99,(4/96),20 96 64 0 70 97 36 14 183 160 1 130 195 187 17 73 138 109 55 146 20 219 105 36 41 182 51 72 82 108 103 144 164 216 102 34 73 177 205 156 57 27 46 5 226 108 129 73 177 117 204 44 141 114 18 134 113 202 67 92 16 50 229 33 78 112 61 229 161 46 138 6 8 142 19 194 30 36 9 6 86 23 132 4 144 131 144 96 96 65 146 96 96 65 246 41 19 114 24 160 132 14 28 0 0 168 128 27 32 131 226 167 92,(4/0),73 69 78 68 174 66 96 130)
    _,←'Save24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 0 7 116 73 77 69 7 218 9 9 21 38 34 3 58 27 73 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 4 22 73 68 65 84 120 218 165 85 77 139 28 85 20 61 85 253 53 221 19 103 180 13 100 52 81 71 113 33 66 220 69 37 136 224 214 189 33 91 119 186 242 31 248 27 4 93 10 174 199 157 8 186 73 54 186 49 27 91 4 65 29 7 68 35 168 147 238 233 238 234 170 122 223 229 185 175 170 123 122 58 163 17 82 112 121 85 197 123 231 220 123 238 199 75 190 252 34 175 192 39 132 42 154 247 129 86 193 185 16 223 101 93 55 107 239 255 183 220 39 231 228 93 107 133 15 63 186 182 69 88 219 22 208 171 47 13 192 127 248 118 164 113 227 134 252 127 240 19 170 42 2 90 231 161 180 197 59 239 143 240 202 51 123 241 251 151 159 142 101 203 14 109 222 150 77 2 174 84 133 92 249 255 7 78 167 156 120 221 216 34 87 24 207 21 110 221 58 64 240 14 195 199 174 202 182 30 45 77 67 8 244 160 34 1 80 150 15 38 240 13 184 165 36 134 114 105 227 49 47 12 166 153 194 189 227 187 232 111 15 81 85 97 181 63 117 110 45 130 210 253 55 120 163 181 137 230 73 224 35 73 182 112 200 114 131 235 175 189 141 225 240 5 18 156 58 218 150 67 2 174 25 69 249 47 18 81 110 122 222 72 226 26 239 99 4 30 165 113 152 100 22 214 123 20 202 64 25 11 81 101 69 32 7 180 14 181 68 218 157 3 94 69 89 162 238 118 221 251 16 19 234 169 128 144 56 130 26 174 90 27 36 149 59 75 96 12 235 201 86 209 139 205 74 17 96 33 17 175 21 61 182 141 231 2 174 13 229 41 45 10 195 125 60 235 104 74 89 244 210 83 156 212 202 102 91 3 49 229 152 76 221 42 153 203 222 16 111 29 215 36 1 52 247 43 38 86 242 117 178 208 52 19 37 76 219 9 229 99 68 172 162 16 54 34 88 2 13 182 218 184 253 245 4 215 95 222 65 175 67 205 170 148 94 211 67 122 153 83 198 82 185 152 171 76 8 10 202 66 28 203 50 255 241 232 4 131 36 229 183 163 26 6 221 158 63 143 0 184 52 236 99 244 253 28 159 223 254 51 150 158 165 55 62 52 201 35 145 72 160 157 141 57 176 4 147 239 42 169 208 69 130 103 31 25 68 2 41 154 51 17 72 235 11 184 100 254 2 35 120 238 210 14 158 126 124 0 145 78 51 233 2 164 217 169 134 4 214 186 152 72 121 95 126 175 76 8 157 172 246 12 65 90 71 16 34 160 28 142 155 4 220 52 224 70 192 220 25 240 77 224 37 184 243 231 70 224 235 230 49 117 200 18 209 162 152 99 174 190 67 233 238 178 42 116 13 238 78 65 165 216 212 9 243 212 221 103 226 119 227 185 104 148 78 156 189 47 201 82 118 53 120 29 197 56 27 225 189 119 223 192 147 123 23 249 191 46 95 169 28 169 119 73 116 89 106 28 30 30 226 147 143 239 80 253 109 246 194 242 188 137 57 59 135 192 214 122 219 58 121 243 242 136 224 111 225 224 211 63 48 155 105 92 123 245 81 236 93 238 49 26 23 235 191 211 233 224 202 149 203 8 173 95 89 70 251 177 195 125 140 34 52 189 179 49 42 172 241 81 103 9 209 152 58 129 226 249 207 135 51 44 88 231 251 207 247 176 123 177 213 68 42 29 109 217 47 11 18 142 209 174 252 138 64 228 169 98 69 174 69 144 176 123 4 180 6 175 19 89 114 110 88 185 56 248 46 179 37 103 201 202 56 145 114 182 66 46 206 112 149 81 159 64 113 109 69 2 233 184 205 8 82 165 138 24 118 212 95 42 132 18 157 76 198 4 118 108 48 67 189 13 231 189 141 218 174 74 81 180 150 126 224 26 130 138 128 66 224 57 131 100 84 251 245 97 87 20 11 228 249 133 56 26 164 66 20 103 119 190 200 112 60 89 160 219 229 134 118 133 173 126 69 210 12 54 208 1 137 214 104 76 167 51 168 178 68 103 192 104 109 139 196 22 162 70 144 187 96 125 216 149 101 254 250 189 191 255 250 170 219 223 166 12 26 69 81 96 58 57 198 232 135 223 177 255 162 71 171 213 134 14 115 28 253 198 67 60 204 129 32 99 16 243 44 139 154 47 50 77 89 69 18 202 149 38 49 130 245 251 32 17 146 155 55 15 222 108 165 253 207 66 44 177 165 185 184 81 86 49 239 237 234 93 204 121 145 134 131 46 36 107 251 252 234 54 251 230 206 7 79 113 25 39 13 81 135 246 68 175 183 235 181 158 165 209 157 135 127 52 109 190 36 144 181 37 209 72 226 151 119 205 67 18 72 40 246 31 65 101 115 230 107 89 99 3,(4/0),73 69 78 68 174 66 96 130)
    _,←'SaveObject24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 0 7 116 73 77 69 7 218 9 9 21 37 57 162 114 129 102 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 4 54 73 68 65 84 120 218 165 85 59 108 28 85 20 61 51 59 235 181 215 193 14 38 82 12 78 192 32 10 20 41 84 88 124 132 64 180 148 72 68 105 169 32 85 58 10 10 106 10 138 68 80 36 18 18 29 200 116 8 9 132 148 52 161 8 145 16 6 36 23 193 68 32 112 4 137 217 245 238 204 236 204 251 15 247 222 253 205 26 71 32 229 73 87 111 118 246 189 115 238 57 239 190 59 209 87 95 14 42 208 8 161 146 240 62 80 84 112 46 200 51 207 245 176 246 223 239 198 235 120 31 63 107 173 240 193 135 27 243 4 107 19 6 61 253 116 27 244 14 223 111 105 156 57 195 239 105 188 249 12 238 57 46 127 135 80 85 2 104 157 135 210 22 111 189 187 133 103 31 91 149 223 191 220 220 227 85 75 20 105 194 139 24 92 169 10 3 229 241 191 6 145 199 52 169 139 215 225 40 227 124 160 208 73 21 174 92 217 68 240 14 43 15 158 230 85 45 138 152 20 4 202 160 34 2 160 44 253 161 217 214 135 39 197 141 115 27 242 220 62 255 60 246 222 187 134 180 48 232 101 10 241 222 46 214 78 158 66 85 133 201 250 216 185 154 130 210 205 2 31 4 31 121 157 95 184 62 121 103 232 76 178 220 33 27 24 188 240 226 27 88 89 121 138 8 166 137 38 188 137 193 53 169 40 239 97 17 217 77 153 7 177 67 14 218 77 51 28 144 255 221 204 194 122 143 66 25 40 99 193 174 76 8 120 131 214 97 104 145 118 135 128 87 98 11 23 131 163 108 141 227 152 38 226 201 129 210 56 56 2 53 52 107 109 16 85 83 156 152 9 140 161 122 178 149 100 81 31 97 4 206 36 156 181 178 30 154 98 245 157 151 229 255 237 183 191 70 159 172 41 12 37 64 123 29 133 82 172 160 102 145 165 13 12 206 25 198 113 64 183 71 85 112 52 153 84 74 60 90 216 164 88 60 160 110 63 215 200 232 220 216 194 56 137 200 62 79 231 228 64 70 206 42 24 95 178 246 124 130 171 223 116 113 251 142 250 207 74 253 244 181 79 240 235 110 142 63 239 150 216 190 217 69 59 138 137 192 145 27 102 86 193 148 0 56 190 178 128 173 159 82 124 113 245 47 164 213 37 178 204 145 69 163 195 179 86 44 208 206 202 25 216 247 127 148 223 85 84 97 14 17 30 127 160 45 4 92 52 33 76 21 36 124 245 25 156 79 254 8 41 120 226 248 18 30 125 168 13 182 78 211 161 91 199 7 71 160 68 96 173 147 131 228 231 241 239 73 208 58 38 176 206 206 16 140 44 10 2 200 155 101 17 131 155 17 184 97 48 55 3 126 16 120 12 238 252 161 10 60 134 149 228 68 50 43 202 139 20 169 250 1 165 219 165 170 208 67 112 55 5 229 98 83 251 75 104 205 173 35 138 150 101 159 4 89 199 201 206 16 8 184 245 35 240 161 138 78 182 133 243 231 94 193 35 171 199 232 253 176 124 149 241 82 239 165 162 40 53 118 118 118 240 241 71 55 200 253 69 186 11 227 253 70 206 236 16 2 59 244 155 75 150 50 77 203 91 4 254 58 54 63 187 141 126 95 99 227 185 163 88 93 107 145 26 39 221 178 217 108 226 196 137 53 132 198 111 116 129 214 229 134 123 81 17 228 238 204 84 17 123 102 41 59 246 153 37 26 51 60 64 206 252 231 157 62 242 220 96 253 201 22 150 143 53 70 74 249 70 83 123 232 229 68 216 65 82 249 9 1 219 83 73 69 214 20 68 81 36 160 67 240 225 65 150 212 55 44 127 56 232 153 123 203 128 186 37 183 19 46 103 203 228 156 12 205 220 234 35 40 154 27 66 192 55 238 160 130 88 169 66 100 139 255 92 33 100 209 126 183 67 192 14 69 105 200 111 67 253 222 138 183 147 82 100 175 249 62 88 190 84 74 0 153 192 87 78 90 181 175 55 187 162 200 49 24 28 145 158 195 21 162 168 119 15 242 12 123 221 28 115 115 180 32 169 48 191 80 17 105 6 27 40 1 86 107 52 122 189 62 84 89 162 217 38 181 182 65 196 22 236 70 224 111 65 173 217 37 101 57 120 233 239 187 119 174 205 45 44 146 13 26 69 81 160 215 221 195 214 246 31 88 63 229 209 104 36 208 33 197 173 223 105 19 109 142 165 207 4 164 89 38 158 231 153 38 91 217 18 178 43 142 68 65 253 123 16 49 201 217 179 155 175 54 226 133 207 131 148 216 56 156 44 228 153 195 123 59 121 230 112 158 173 49 196 25 213 214 249 201 215 236 219 27 23 78 210 212 137 106 205 242 225 86 107 217 107 221 143 37 157 251 31 154 34 29 19 240 220 96 53 181 14 125 191 36 44 197 254 3 131 218 43 220 156 62 233 202,(4/0),73 69 78 68 174 66 96 130)
    _,←'Scrubber24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 0 7 116 73 77 69 7 218 9 17 23 19 44 208 154 95 230 0 0 0 9 112 72 89 115 0 0 11 18 0 0 11 18 1 210 221 126 252 0 0 5 136 73 68 65 84 120 218 133 85 223 111 20 85 20 254 238 204 157 217 217 221 238 78 187 5 10 45 180 229 71 161 145 6 140 13 37 10 1 129 22 9 9 49 18 52 208 55 245 31 32 34 15 134 40 60 25 95 64 141 47 234 131 36 132 24 94 48 68 31 136 6 99 53 16 68 30 64 193 148 132 150 212 166 180 221 82 218 238 110 187 63 102 238 15 207 204 118 237 182 193 48 201 201 189 119 247 222 243 157 251 157 239 156 203 176 240 213 239 59 213 183 107 60 47 118 59 49 171 155 51 221 64 191 197 200 24 153 14 119 176 202 86 22 252 98 66 43 9 67 103 132 199 238 169 60 110 14 13 140 158 125 255 72 91 238 212 187 93 149 221 154 87 1 212 25 156 117 220 249 100 231 6 154 39 201 34 100 124 17 192 226 207 152 255 47 8 162 249 240 185 251 130 171 85 39 134 39 229 197 159 110 13 15 236 223 222 28 156 97 70 213 129 85 166 101 244 208 216 101 217 246 114 26 163 100 54 153 53 111 246 18 227 243 163 75 214 254 221 123 29 7 183 111 91 127 122 112 82 159 57 243 205 93 155 64 176 244 6 53 2 118 43 141 228 156 233 127 10 17 40 90 152 198 2 57 154 98 82 243 163 231 209 143 82 96 67 189 1 83 251 225 166 115 199 26 112 244 179 236 155 74 129 93 238 27 253 240 227 243 183 31 85 3 68 82 209 92 93 48 241 51 55 216 244 99 1 165 88 0 160 117 217 59 211 228 93 5 43 89 64 198 175 135 103 175 197 166 84 18 218 138 104 230 151 130 93 242 210 241 54 251 196 183 233 222 219 119 6 134 174 220 24 251 180 26 192 152 42 36 248 175 125 39 241 106 235 8 154 138 17 77 225 146 83 205 20 133 20 192 104 233 67 73 5 67 228 96 91 221 152 113 122 195 116 145 243 74 250 67 127 103 123 233 38 19 99 187 192 150 215 87 231 64 127 222 121 128 55 170 139 48 237 26 36 221 38 36 18 203 117 34 230 34 25 171 67 50 190 18 9 58 30 151 227 208 172 22 158 179 145 8 174 193 208 76 181 11 232 124 25 3 151 142 191 184 165 168 217 235 255 221 224 239 171 117 206 83 243 160 209 210 184 6 58 154 4 247 36 165 48 202 160 40 143 162 4 85 156 64 145 37 80 116 58 145 139 237 133 23 219 15 198 28 60 201 9 248 158 128 99 27 88 83 203 89 12 162 162 184 164 167 145 12 225 179 131 31 33 233 36 86 180 196 199 153 27 117 32 148 9 49 247 144 41 63 75 156 27 240 74 25 148 178 3 152 245 109 100 98 123 80 76 236 132 233 184 160 90 65 145 242 59 156 97 232 159 80 184 114 191 88 17 68 96 158 193 141 60 207 12 158 134 151 253 179 57 18 93 217 193 99 117 204 142 184 80 34 167 117 105 130 196 217 196 68 33 141 185 201 191 224 179 122 20 234 143 64 186 123 136 26 7 70 32 39 186 1 233 128 18 85 8 146 69 212 45 169 20 198 192 243 163 23 106 180 228 135 35 201 142 131 142 187 89 155 118 45 147 179 147 180 159 118 19 75 254 92 26 249 233 1 120 238 58 40 190 26 134 63 141 72 81 130 211 173 36 121 52 185 11 211 170 163 173 38 12 102 46 69 160 91 90 141 47 40 149 221 1 154 48 94 75 202 33 165 40 162 209 92 65 178 177 41 6 11 6 57 224 20 165 154 254 13 90 20 8 152 114 34 124 8 163 150 202 172 19 50 249 82 217 153 185 184 232 3 89 115 34 249 132 149 216 212 19 77 52 193 17 163 172 48 91 84 66 216 204 138 183 209 17 7 156 212 154 72 109 129 79 92 203 252 61 98 150 242 145 190 137 18 82 96 77 71 9 60 14 110 69 32 73 190 92 47 1 8 116 43 139 211 235 184 25 77 79 77 69 89 174 96 214 196 44 195 136 219 20 93 73 107 203 174 5 183 235 168 233 216 48 158 222 163 68 143 64 48 23 210 110 166 168 183 67 45 163 139 199 215 82 49 146 24 233 230 166 177 152 32 73 20 243 166 189 253 151 128 254 96 189 231 171 243 39 247 111 93 223 108 180 217 15 145 25 253 3 53 169 118 106 32 173 240 102 199 225 23 179 200 171 24 74 209 118 168 213 175 81 67 217 69 172 4 42 39 222 149 32 254 171 154 109 5 64 168 176 42 46 147 61 38 43 245 141 119 237 123 187 247 144 37 39 126 192 244 208 47 208 185 33 74 38 169 41 59 12 51 178 22 70 203 33 68 107 59 161 107 218 168 61 144 130 100 208 58 16 182 148 96 244 213 82 0 25 2 12 205 175 251 251 51 237 33 129 142 156 64 212 208 200 103 210 120 160 214 227 229 55 126 127 70 183 46 225 121 159 239 73 84 247 162 164 107 206 132 19 49 114 153 120 94 129 73 97 235 91 236 0 251 250 199 25 204 26 203 232 109 161 230 20 132 29 188 55 84 100 149 124 86 186 236 120 30 184 126 152 10 83 155 164 63 9 233 235 69 0 172 53 250 200 248 254 203 15 176 173 117 14 119 173 183 32 26 183 34 21 239 66 7 168 115 6 254 36 61 98 84 217 122 94 33 122 94 42 1 255 68 55 196 19 169 67 104 21 20 8 117 115 79 177 69 0 9 54 109 184 110 11 70 146 155 113 173 120 140 113 171 69 55 68 168 112 189 64 207 212 85 45 226 154 78 134 188 235 224 34 58 4 97 84 199 38 77 34 156 202 6 62 21 112 57 221 190 191 152 34 149 158 75 9 175 225 152 181 251 149 110 92 184 193 117 38 11 204 81 51 83 44 116 67 76 148 117 94 137 60 156 135 49 179 176 77 148 68 153 49 211 100 212 55 32 164 82 126 53 0 198 114 41 214 179 183 59 156 127 177 67 4 23 127 110 34 151 126 150 29 97 244 190 106 234 176 178 88 146 162 26 96 44 91 138 95 163 113 35 89 61 89 28 229 183 248 255 30 253 103 125 204 247 74 115 52 146 182 49 232 198 120 186 26 224 177 205 197 213 246 119 126 158 74 68 173 46 86 110 186 113 93 238 145 164 244 128 135 5 213 4 42 42 47 137 184 133 54 106 104 173 51 244 172 14 209 228 250 170 148 243 224 95 93 65 115 89 8 40 151 16,(4/0),73 69 78 68 174 66 96 130)
    _,←'SelectAll24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 3 147 73 68 65 84 120 94 181 149 61 79 36 71 16 134 223 174 158 207 5 223 234 238 176 207 18 39 11 19 56 32 57 91 58 217 137 115 66 71 206 73 44 33 34 19 89 247 11 46 176 8 65 72 252 3 140 68 194 31 112 132 68 96 201 150 65 14 184 21 150 96 145 191 216 131 131 93 118 102 103 202 221 165 238 153 93 86 227 204 181 106 117 175 74 211 207 188 111 213 116 227 255 14 53 158 216 223 223 255 106 110 110 238 99 165 84 233 1 204 140 178 44 253 90 134 141 162 40 112 121 121 25 12 6 131 223 162 40 250 113 101 101 229 238 63 1 59 59 59 241 210 210 210 47 11 11 11 31 153 135 153 148 2 148 242 16 183 84 21 128 136 112 126 126 78 102 252 125 125 125 253 131 25 223 173 175 175 103 15 1 1 234 160 52 77 159 22 37 39 253 251 12 68 36 124 163 198 113 220 236 214 33 24 237 118 27 113 28 207 27 21 223 30 31 31,(4/183),95 173 174 174 254 209 4 0 51 112 151 49 46 122 67 73 138 8 40 24 11 144 101 25 74 46 221 219 107 124 248 120 6 81 160 145 36 137 133 96 113 113 241 107 163 230 98 107 107 235 251 181 181 181 155 41 128 147 206 163 146 97 21 176 219 76 145 194 123 65 140 82 105 176 115 180 4 129 21 97 148 141 16 26 120 171 213 66 158 231 179 183 183 183 223 244 122 189 159 108 57 61 128 198 1 140 210 89 194 128 183 195 77 121 158 25 21 67 81 50 26 229 85 162 211 233 224 232 232 8 39 39 39 48 86 61 51 54 191 222 216 216 208 83 10 164 83 24 2 208 164 161 192 16 6 17 136 20 2 29 160 36 18 24 73 30 120 250 228 9 226 56 177 111 15 34 18 187 78 79 79 159 31 30 30 38 0 238 38 0 21 145 8 105 146 88 146 175 129 128 162 48 0 115 13 29 228 5 254 122 7 0 33 84 16 162 80 10 51 81 138 52 109 121 187 167 21 136 73 204 24 12 135 85 239 11 200 207 181 107 184 237 99 34 180 214 8 231 35 0 220 12 128 244 187 216 1 38 11 80 117 29 4 224 255 203 170 14 111 165 114 248 230 54 149 190 151 183 97 214 80 228 69 200 194 217 227 116 212 24 81 78 138 92 14 211 10 234 46 98 101 115 89 158 25 69 86 141 7 212 54 121 146 71 192 33 52 249 173 154 1 206 34 235 103 0 34 174 138 140 241 89 32 147 54 177 59 58 234 28 208 220 166 176 53 208 96 177 166 242 94 102 71 128 154 82 32 214 57 43 85 51 128 229 231 45 42 189 45 2 169 109 33 1 76 24 228 186 8 120 228 153 77 53 0 224 44 82 84 130 42 123 104 172 224 117 129 235 162 122 139 8 62 154 107 0 233 34 16 75 219 121 5 85 161 103 90 51 8 66 141 251 251 123 228 163 66 114 204 236 79 221 102 128 57 164 220 199 197 24 137 69 242 176 0 188 223 113 28 161 255 231 41 58 111 222 224 211 47 190 196 32 147 99 189 234 34 18 0 227 230 230 102 26 176 187 187 75 203 203 203 178 41 145 6 236 236 189 175 124 87 200 70 5 174 223 246 240 252 131 199 88 136 98 144 83 96 219 58 10 53 206 242 17 246 246 246 232 1 64 78 197 164 40 74 157 134 132 247 219 137 88 69 36 214 120 152 220 11 63 255 254 43 46 186 93 228 217 16 173 36 6 20 75 13 74 176 40 96 102 117 118 118 150 2 120 231 1 190 120 186 223 191 27 196 154 219 207 30 197 236 91 84 70 85 97 224 243 151 47 241 217 139 23 98 231 63 87 87 19 31 86 24 134 202 92 157 3 187 116 77 198 1 234 232 109 110 110 190 62 56 56 248 132 77 96 58 236 166 150 68 54 93 20 133 50 16 25 238 142 102 219 28 221 110 183 3 224 173 187 107 36 89 197 236 236 108 148 166 169 208 31 50 28 83 44 176 121 115 123 89 200 248 151 197 70 1 103 89 54 52 208 28 128 116 204 191 207 201 183 10 205 89 192 229,(4/0),73 69 78 68 174 66 96 130)
    _,←'Undo24.png' (⎕UCS 66 77 80 137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 24 0 0 0 24 8 6 0 0 0 224 119 61 248 0 0 3 166 73 68 65 84 72 137 181 147 93 136 85 85 20 199 127 107 157 125 206 253 152 185 122 28 199 175 64 180 15 199 132 116 174 160 153 132 52 55 146 10 138 140 80 130 160 18 170 183 162 135 168 71 153 199 124 202 222 162 135 38 41 233 33 154 12 223 231 130 224 131 70 94 73 163 102 52 70 68 209 28 241 220 153 185 31 231 99 159 221 131 140 244 229 168 112 251 63 254 217 235 255 219 107 173 189 133 123 212 166 125 71 67 96 4 145 42 128 32 32 18 1 245 95 190 121 161 113 167 58 217 240 194 215 213 169 99 175 221 241 192 166 189 223 143 32 114 160 88 8 70 250 251 251 41 20 251 201 114 33 181 144 196 9 73 210 38 203 186 211 56 55 122 238 200 243 99 255 2 12 189 120 228 38 170 181 201 163 175 54 254 17 188 30 248 98 96 89 101 100 229 202 85 168 41 211 234 58 58 137 37 201 28 206 129 122 6 227 121 56 27 211 158 191 65 183 59 219 0 169 157 251 234 217 232 54 96 120 223 183 46 8 76 52 219 74 107 191 141 239 109 44 220 186 16 152 241 117 107 215 132 165 254 1 102 154 41 215 155 49 153 205 241 125 131 167 30 158 42 162 160 234 240 141 135 239 249 196 157 57 154 209 213 134 195 213 206 30 222 29 1 152 82 209 103 199 182 141 225 137 83 147 19 143 190 242 93 77 84 215 151 75 193 248 198 161 135 232 164 1 83 151 91 116 98 75 169 232 83 233 235 195 218 28 107 19 200 83 60 175 128 111 10 168 39 56 177 244 45 9 65 92 53 138 254 24 7 106 0 242 196 27 63 184 93 79 110 165 213 106 115 242 244 249 8 17 54 14 61 28 118 179 128 11 151 231 177 206 177 172 82 70 128 214 252 13 226 184 21 57 103 235 136 156 17 228 41 245 252 145 82 121 41 149 37 3 120 6 60 81 102 174 95 161 221 158 219 255 243 216 211 99 6 132 60 135 140 2 195 143 61 18 206 206 119 73 242 128 233 171 243 32 48 184 180 31 155 118 137 162 107 216 60 27 5 249 100 161 125 128 205 251 39 194 78 39 58 148 101 221 215 7 87 60 128 250 66 56 48 72 55 238 28 0 198 100 199 155 199 220 206 199 171 52 166 154 172 93 221 135 67 185 112 121 142 52 115 132 149 50 73 183 197 205 232 90 228 156 171 157 61 188 251 142 175 173 250 246 241 79 11 133 190 119 87 173 94 131 241 132 75 151 46 17 199 157 173 234 16 108 14 57 112 241 106 155 43 51 109 108 14 75 250 75 228 54 35 106 94 143 156 99 209 112 128 198 231 187 222 235 198 157 122 171 213 38 206 28 197 82 25 81 221 163 32 164 57 136 8 0 113 154 83 42 250 24 99 104 54 103 200 93 62 122 246 240 51 139 134 223 150 234 232 236 92 147 78 2 104 0 162 195 234 156 144 101 14 85 80 79 240 60 33 119 57 169 117 4 165 65 60 47 56 176 121 127 189 122 47 249 103 62 219 89 207 108 54 173 146 83 40 149 16 213 80 29 183 126 101 224 27 10 129 193 247 13 214 41 55 102 187 100 18 80 170 172 9 61 207 159 216 242 214 241 123 130 32 82 183 54 65 85 65 20 205 17 58 221 148 114 96 48 106 40 26 195 178 178 207 218 229 5 86 86 28 203 43 134 82 121 69 168 234 77 12 191 115 226 238 16 213 139 162 2 226 16 85 140 226 248 253 252 228 93 235 12 132 192 105 64 22 109 64 61 172 181 220 130 104 195 252 248 229 115 139 22 220 183 68 241 140 33 77 83 68 245 140 246 52 28 110 141 197 247 72 226 20 68 235 255 7 96 93 150 100 164 185 109 252 116 104 203 116 207 1 136 86 147 212 34 170 135 224 46 11 187 95 109 255 104 42 84 184 233 156 155 62 121 112 232 65 0 211 75 128 136 236 201 29 0 163 11 94 143 71 36 47 137 72 253 212 193 13 99 11 78 207 58 216 254 225 100 232 92 62 34 34 91 255 234 247 12 224 156 123 95 224 229 147 31 111 152 238 85 230 223 180 237 131 95 215 255 151 255 39 236 42 134 163 149 151 79 67,(4/0),73 69 78 68 174 66 96 130)
    _,←'fmXED.ico' (⎕UCS 73 67 79 0 0 1 0 1 0 16 16 0 0 1 0 32 0 104 4 0 0 22 0 0 0 40 0 0 0 16 0 0 0 32 0 0 0 1 0 32,(29/0),123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255 123 123 123 255,(12/0),132 132 132,(45/255),132 132 132 255,(12/0),140 140 140,(45/255),140 140 140 255,(12/0),156 156 156,(45/255),156 156 156 255,(12/0),173 173 173,(9/255),247 247 247 255 247 239 239 255 247 247 247 255 247 247 247 255 247 247 247 255 247 247 247 255 247 247 247,(9/255),173 173 173 255,(12/0),173 173 173,(5/255),222 222 222 255 148 148 156 255 239 214 198 255 222 132 57 255 231 140 57 255 239 148 74 255 247 214 189 255 148 148 156 255 222 222 222,(5/255),173 173 173 255,(12/0),173 173 173 255 222 222 222 255 140 140 140 255 247 239 239 255 214 123 57 255 198 90 8 255 231 140 57 255 247 181 115 255 239 165 90 255 247 239 239 255 140 140 140 255 222 222 222 255 173 173 173 255,(12/0),181 181 181 255 148 148 148 255 247 255 255 255 214 181 156 255 198 82 8 255 189 90 8 255 214 115 33 255 231 156 82 255 239 156 82 255 214 181 156 255 247 255 255 255 148 148 148 255 181 181 181 255,(12/0),181 181 181 255 148 148 156 255 247 255 255 255 239 206 165 255 255 189 123 255 239 165 90 255 231 140 57 255 222 123 33 255 214 115 33 255 222 198 189 255 247 255 255 255 148 148 156 255 181 181 181 255,(12/0),181 181 181 255 222 222 222 255 148 148 156 255 247 247 247 255 247 206 156 255 255 214 156 255 247 181 115 255 231 156 82 255 231 156 82 255 247 247 247 255 148 148 156 255 222 222 222 255 181 181 181 255,(12/0),189 189 189,(5/255),222 222 222 255 148 148 156,(10/255),231 198,(9/255),148 148 156 255 222 222 222,(5/255),181 181 181 255,(12/0),189 189 189,(45/255),189 189 189 255,(12/0),189 189 189,(33/255),189 189 189 255 181 181 181 255 181 181 181 255 189 189 189 255,(12/0),189 189 189,(33/255),198 198 198,(5/255),189 189 189 255,(16/0),189 189 189,(33/255),198 198 198 255 189 189 189 255,(20/0),189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255 189 189 189 255,(20/0),128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 3 172 65 128 7 172 65 128 15 172 65 128 31 172 65)
    fmXED_toi←0,⍨18 2⍴_

    ⎕ex¨ 'L' '_'

:EndNamespace